USE <your_db_name_here>
GO

SET ANSI_NULLS ON
GO

SET QUOTED_IDENTIFIER ON
GO

-- Drop procedure if it already exist
IF OBJECT_ID ( 'OOGetErrorInfo', 'P' ) IS NOT NULL
    DROP PROCEDURE dbo.OOGetErrorInfo
GO

CREATE PROCEDURE [dbo].[OOGetErrorInfo]

/*
* Author: Harar Zafrir, zafrir@hpe.com
* Maintainer: Vlad Moldovan, moldovan@hpe.com
*
* Version: 4.2
*
* Last update: 2016-DEC-01
*
* This utility procedure gets error information.
* Verified on Microsoft SQL Server 2008R2 and 2012.
*
* -----------------------------------------------------------------------------
* Change List
* ------------
* 2014-AUG-13 [3.0] HP OO 10.20 - Replace the old usp_GetErrorInfo with this one.
* 2015-FEB-01 [3.1] No changes in this script,
*                   OOPurgeHistory.sql fixed - varchar to integer conversion overflow
* 2015-AUG-23 [3.5] No changes in this script, cross-db maintenance scripts version alignment
* 2015-NOV-04 [3.7] No changes in this script, HPE OO 10.60 adaptations
* 2016-DEC-01 [4.2] DB Maintenance Package alignment - no changes
* -----------------------------------------------------------------------------
*/

AS
SELECT
    ERROR_NUMBER() AS ErrorNumber
    ,ERROR_SEVERITY() AS ErrorSeverity
    ,ERROR_STATE() AS ErrorState
    ,ERROR_PROCEDURE() AS ErrorProcedure
    ,ERROR_LINE() AS ErrorLine
    ,ERROR_MESSAGE() AS ErrorMessage
GO

-- Verify that the error info procedure exist
IF OBJECT_ID ( 'OOGetErrorInfo', 'P' ) IS NULL
    RAISERROR( 'OOGetErrorInfo procedure is missing ...' , 0,1) WITH NOWAIT;
GO

-- Replace this stored procedure in case it already exists
IF OBJECT_ID ( 'OOPurgeHistory', 'P' ) IS NOT NULL
    DROP PROCEDURE [dbo].[OOPurgeHistory]
GO

CREATE PROCEDURE [dbo].[OOPurgeHistory]

@PurgeExecutionsOlderThan INTEGER = -1,           -- Retention in days
@PurgeExecutionsInBatchesOf INTEGER = 10000,      -- Default = 10000, minimum = 500, maximum = 100000.
@ShouldPurgeExecutionSummary INTEGER = 0,         -- Should OO_EXECUTION_SUMMARY table records be deleted as well
@Verbose INTEGER = 1,                             -- Verbosity level
@StopPurgingProcessAfter INTEGER = 4,             -- Run timeout for this procedure (at least 1 hour)
@DeepClean INTEGER = 0,                           -- Deeper data cleansing
@DisableIndexes INTEGER = 0                       -- Should OO_STEP_LOG_STARTED indexes be disabled for the duration of this procdure run

/*
* Author: Harar Zafrir, zafrir@hpe.com
* Maintainer: Vlad Moldovan, moldovan@hpe.com
*
* Version: 4.2
*
* Last update: 2016-DEC-01
*
* This procedure purges flow and step data from OO 10.22 to 10.51 (inclusive) database tables
* Compatible with Microsoft SQL Server 2008R2 and 2012
*
* Parameters:
* -----------------------------------------------------------------------------
*
* @PurgeExecutionsOlderThan:
*       How many days of history to keep from "now". This value has no default and must be specified.
*
* @PurgeExecutionsInBatchesOf:
*       The maximum number of flows to handle at once. Default is 10000, minimum value is 500, maximum is 100000.
*
* ShouldPurgeExecutionSummary:
*       Should OO_EXECUTION_SUMMARY table be purged. Default is 0 (do not purge this table)
*
* @Verbose:
*       Verbosity level. 0 means "quiet" output. 1 provides medium level details
*       and 2 provides extended details.
*
* @StopPurgingProcessAfter:
*       Maximum time for the procedure to run. Default is 4 hours.
*
* @DeepClean:
*       Perform deeper cleansing - search for orphan records.
*       Using this option usually extends procedure run time.
*
* @DisableIndexes:
*       Disable OO_STEP_LOG_STARTED indexes for the duration of this procedure run
*       Using this option may result in slower step drill-down operations
*       When the procedure ends, indexes are rebuilt
*       This feature is disabled by default (0)
*
* Return Values:
* -----------------------------------------------------------------------------
* The actual number of processed flows.
*
* -----------------------------------------------------------------------------
* Change List
* -----------------------------------------------------------------------------
* 2014-AUG-13 [3.0] HP OO 10.20 modifications
*                   Adding execution bount inputs / outputs purge
* 2015-FEB-01 [3.1] Fix for varchar to integer conversion overflow
* 2015-APR-13 [3.2] Fix for million-bindings flow
* 2015-AUG-23 [3.5] No changes in this script, cross-db maintenance scripts version alignment
* 2015-NOV-04 [3.7] HPE OO 10.60 adaptations
* 2016-MAY-16 [4.0] Enhance performance by reducing SELECT queries
* 2016-AUG-09 [4.1] HPE OO 10.70 adaptations
* 2016-DEC-01 [4.2] DB Maintenance Package alignment - no changes
* -----------------------------------------------------------------------------
*/

AS

BEGIN

  SET NOCOUNT ON;       --  Stops the message indicating the number of rows affected
                        --  by a Transact-SQL statement from being returned as part
                        --  of the results

  SET ARITHABORT ON;    --  Terminates a query when an overflow or divide-by-zero
                        --  error occurs during query execution

  SET XACT_ABORT ON;    --  SET XACT_ABORT ON will cause the transaction to be
                        --  uncommittable when the constraint violation occurs

  SET DEADLOCK_PRIORITY LOW;    --  Sets the deadlock priority for this session
                                --  to low such that in case we disturb OO operation
                                --  we're the losing party

  DECLARE
    -- General properties ---------------------------
     @purge_history_version            VARCHAR(10)
    ,@purge_history_compatibility      VARCHAR(15)
    ,@supported_versions               VARCHAR(255)
    ,@purged_flow_statuses             VARCHAR(255)
    ,@detected_oo_version              VARCHAR(30)
    -- Thresholds, batch size -----------------------
    ,@lots_of_steps_threshold          BIGINT = 100000 -- Flows with more than this number of steps are considered unusually big
    ,@delete_batch_size                BIGINT = 10000  -- An optimal number of rows to be deleted together, database efficiency
    ,@min_batch_size                   BIGINT = 500    -- Batch size lower limit
    ,@max_batch_size                   BIGINT = 10000  -- Batch size upper limit
    ,@flow_batch_size                  INTEGER = 0     -- An optimal number of flows to be handled together
    -- Validations and general purpose --------------
    ,@lockRetVal                       INTEGER = -10
    ,@msg                              VARCHAR(1000)
    ,@gen_prps_countr                  INTEGER = 0
    ,@OutputRowLength                  INTEGER = 80
    -- Operations and flow control ------------------
    ,@max_exec_summary_id_to_delete    BIGINT
    ,@min_exec_summary_id_to_delete    BIGINT
    ,@curr_batch_min_id                BIGINT
    ,@curr_batch_max_id                BIGINT
    ,@curr_set_max_id                  BIGINT
    ,@curr_batch_flow_count            INTEGER = 0
    ,@deleted_rows_in_current_batch    INTEGER = 0
    ,@curr_set_accum_step_count        INTEGER = 0
    ,@curr_step_count                  INTEGER = 0
    ,@curr_set_size                    INTEGER = 0
    ,@curr_set_idx                     BIGINT = 0
    ,@idx_shift                        BIGINT = 0
    ,@max_idx_value                    BIGINT = 0
    ,@curr_set_min_idx                 BIGINT = 0
    ,@curr_set_max_idx                 BIGINT = 0
    ,@temp_idx                         BIGINT = 0
    ,@last_start_time                  DATETIME
    ,@are_constraints_disabled         BIT = 0
    -- Preformatted queries -------------------------
    ,@query                            NVARCHAR(2000)
    ,@query1                           NVARCHAR(2000)
    ,@query2                           NVARCHAR(2000)
    ,@query3                           NVARCHAR(2000)
    ,@query4                           NVARCHAR(2000)
    ,@query5                           NVARCHAR(2000)
    ,@query_parameters                 NVARCHAR(2000)
    -- Counters -------------------------------------
    ,@rows_del_frm_es                  INTEGER = 0
    ,@rows_del_frm_slb                 BIGINT = 0
    ,@rows_del_frm_sle                 BIGINT = 0
    ,@rows_del_frm_slstr               BIGINT = 0
    ,@rows_del_frm_bo                  BIGINT = 0
    ,@rows_del_frm_bi                  BIGINT = 0
    ,@total_rows_deleted               BIGINT = 0
    ,@total_num_of_flows_handled       INTEGER = 0
    -- Time, duration -------------------------------
    ,@purge_start_time                 DATETIME
    ,@end_purge_by                     DATETIME
    ,@n_o_w                            DATETIME
    ,@temp_timestamp                   DATETIME
    ,@elapsed_seconds                  INTEGER = 0
    ;

  -- For iterating over OO_STEP_LOG_STARTED step counts
  DECLARE @sl_strtd_stps TABLE (
    v_idx BIGINT PRIMARY KEY IDENTITY(1,1),
    v_execution_id VARCHAR(64),
    v_step_count INTEGER,
    UNIQUE CLUSTERED (v_execution_id,v_idx)
  );

  -- A buffer for specific steps HASH IDs
  DECLARE @stps_hash_tab TABLE (v_step_hash_id VARCHAR(42));

  -- A buffer for specific steps ENDED IDs
  DECLARE @stps_ended_tab TABLE (v_ended_id BIGINT);

  -- validate input params
  IF (@PurgeExecutionsInBatchesOf < @min_batch_size) BEGIN
    SET @msg = REPLICATE('=', @OutputRowLength);
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    RAISERROR('**** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR ***', 0,1) WITH NOWAIT;
    SET @msg = '**** Purging batch size is too small. (provided ' + CAST(@PurgeExecutionsInBatchesOf AS VARCHAR) + '). Minimum value is ' + CAST(@min_batch_size AS VARCHAR) + '.';
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    SET @msg = REPLICATE('=', @OutputRowLength);
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    RETURN
    END;

  IF (@PurgeExecutionsInBatchesOf > @max_batch_size) BEGIN
    SET @msg = REPLICATE('=', @OutputRowLength);
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    RAISERROR('**** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR ***', 0,1) WITH NOWAIT;
    SET @msg = '**** Purging batch size is too big. (provided ' + CAST(@PurgeExecutionsInBatchesOf AS VARCHAR) + '). Maximum value is ' + CAST(@max_batch_size AS VARCHAR) + '.';
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    SET @msg = REPLICATE('=', @OutputRowLength);
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    RETURN
    END;

  IF (@PurgeExecutionsOlderThan < 1) BEGIN
    SET @msg = REPLICATE('=', @OutputRowLength);
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    RAISERROR('**** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR ***', 0,1) WITH NOWAIT;
    SET @msg = '**** Cannot keep less than 1 day. (provided ' + CAST(@PurgeExecutionsOlderThan AS VARCHAR) + '). Set @PurgeExecutionsOlderThan parameter accordingly.';
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    SET @msg = REPLICATE('=', @OutputRowLength);
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    RETURN
    END;

  IF (@PurgeExecutionsOlderThan > 1000) BEGIN
    SET @msg = REPLICATE('=', @OutputRowLength);
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    RAISERROR('**** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR ***', 0,1) WITH NOWAIT;
    SET @msg = '**** Retention is more than 1000 days? (provided ' + CAST(@PurgeExecutionsOlderThan AS VARCHAR) + '). Please check the value provided for @PurgeExecutionsOlderThan.';
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    SET @msg = REPLICATE('=', @OutputRowLength);
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    RETURN
    END;

  IF (@StopPurgingProcessAfter > 1000) BEGIN
    SET @msg = REPLICATE('=', @OutputRowLength);
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    RAISERROR('**** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR ***', 0,1) WITH NOWAIT;
    SET @msg = '**** Unlimited run time? (provided pStopPurgingProcessAfter: ' + CAST(@StopPurgingProcessAfter AS VARCHAR) + '). Please check the value provided for @StopPurgingProcessAfter.';
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    SET @msg = REPLICATE('=', @OutputRowLength);
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    RETURN
    END;

  IF (@ShouldPurgeExecutionSummary NOT IN (0,1)) BEGIN
    SET @msg = REPLICATE('=', @OutputRowLength);
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    RAISERROR('**** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR ***', 0,1) WITH NOWAIT;
    SET @msg = '**** @ShouldPurgeExecutionSummary must be either 0 or 1. (provided ' + CAST(@ShouldPurgeExecutionSummary AS VARCHAR) + ').';
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    SET @msg = REPLICATE('=', @OutputRowLength);
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    RETURN
    END;

  IF (@DeepClean NOT IN (0,1)) BEGIN
    RAISERROR('INFO: @DeepClean must be either 0 or 1. Deep clean is deactivated during this run.',0,1) WITH NOWAIT;
    SET @DeepClean = 0;
    END;

  IF (@DisableIndexes NOT IN (0,1)) BEGIN
    RAISERROR('INFO: @DisableIndexes must be either 0 or 1. Constraints will not be modified.',0,1) WITH NOWAIT;
    SET @DisableIndexes = 0;
    END;

  IF (@Verbose NOT IN (0,1,2,3,4)) BEGIN
    RAISERROR('INFO: @Verbose must be one of: 0,1 or 2. Defaulting to 1.',0,1) WITH NOWAIT;
    SET @Verbose = 1;
    END;

  SET @purge_history_version = '4.2';
  SET @purge_history_compatibility = '10.22 to 10.51 inclusive';
  SET @supported_versions = '''10.22'',''10.22.0001'',''10.22.0002'',''10.50'',''10.51''';
  SET @purged_flow_statuses = '''COMPLETED'',''CANCELED'',''SYSTEM_FAILURE''';
  SET @purge_start_time = GETDATE();
  SET @end_purge_by = DATEADD(hour,@StopPurgingProcessAfter,@purge_start_time);

  ---------------------------------------------------------------------------------------------------------
  -- Verify that OO release is supported by this stored procedure                                        --
  ---------------------------------------------------------------------------------------------------------
  SET @query = N'SELECT @detected_oo_versionOUT = VALUE, @gen_prps_countrOUT = CASE WHEN VALUE IN (' + @supported_versions + ') THEN 1 ELSE 0 END FROM OO_SYSINFO WHERE UPPER(KEY_)=''VERSION''';
  SET @query_parameters = N'@detected_oo_versionOUT VARCHAR(30) OUTPUT, @gen_prps_countrOUT INTEGER OUTPUT'
  EXECUTE sp_executesql @query, @query_parameters, @detected_oo_versionOUT=@detected_oo_version OUTPUT, @gen_prps_countrOUT = @gen_prps_countr OUTPUT;
  IF @gen_prps_countr <> 1 BEGIN
    SET @msg = REPLICATE('=', @OutputRowLength);
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    RAISERROR('**** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR ***', 0,1) WITH NOWAIT;
    SET @msg = '**** Unsupported OO version: ' + @detected_oo_version + '. This procedure only supports: ';
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    SET @msg = '**** ' + REPLACE(@supported_versions,'''''','''');
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    SET @msg = REPLICATE('=', @OutputRowLength);
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    RETURN
  END;

  -- Output header --------------------------------------------------------------------------------------------------------------------
  SET @msg = REPLICATE('-', @OutputRowLength);
  RAISERROR(@msg, 0,1) WITH NOWAIT;
  SET @msg = CAST(@purge_start_time AS VARCHAR) + ' Beginning execution of OOPurgeHistory procedure version ' + @purge_history_version;
  RAISERROR( @msg , 0,1) WITH NOWAIT;
  SET @msg = CAST(@purge_start_time AS VARCHAR) + ' This procedure is compatible with OO ' + @purge_history_compatibility;
  RAISERROR( @msg , 0,1) WITH NOWAIT;
  SET @msg = REPLICATE('-', @OutputRowLength);
  RAISERROR(@msg, 0,1) WITH NOWAIT;
  -- End of output header -------------------------------------------------------------------------------------------------------------

  -- Count rows in OO_EXECUTION_SUMMARY table. This could take a while, but it is worth it
  SELECT @gen_prps_countr = COUNT(1) FROM [dbo].[OO_EXECUTION_SUMMARY] WITH (NOLOCK);

  IF (@gen_prps_countr = 0)
  BEGIN
    SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' OO_EXECUTION_SUMMARY table is empty.';
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    IF (@DeepClean = 1)
      BEGIN
        GOTO Check_For_Orphans;
      END;
      ELSE BEGIN
        GOTO Summary;
      END;
    END;

  -- Retention period calculation. Get the newest run start time we need to KEEP
  SET @last_start_time = DATEADD(day, (-1 * @PurgeExecutionsOlderThan), @purge_start_time);

  -- Flow batch size is relative to the steps batch size, but at least 100
  SET @flow_batch_size = ROUND((@PurgeExecutionsInBatchesOf/10),0);
  IF (@flow_batch_size < 100) BEGIN
    SET @flow_batch_size = 100;
  END;

  ---------------------------------------------------------------------------------------------------------
  -- Get the minimal and maximal EXECUTION_ID values to handle                                           --
  -- As EXECUTION_ID allocation is asynchroneous, this range might also include EXECUTION_IDs            --
  -- which have started after the threshold time and should not be handled                               --
  ---------------------------------------------------------------------------------------------------------
  SET @query1 = N'SELECT' +
      ' @min_exec_summary_id_to_deleteOUT = MIN(CAST(EXECUTION_ID AS BIGINT)),' +
      ' @max_exec_summary_id_to_deleteOUT = MAX(CAST(EXECUTION_ID AS BIGINT))' +
      ' FROM [dbo].[OO_EXECUTION_SUMMARY] WITH (NOLOCK)' +
      ' WHERE START_TIME < @last_start_timeIN' +
      ' AND STATUS IN (' + @purged_flow_statuses + ')';

  SET @query2 = N'SELECT' +
      ' @min_exec_summary_id_to_deleteOUT = MIN(CAST(ES.EXECUTION_ID AS BIGINT)),' +
      ' @max_exec_summary_id_to_deleteOUT = MAX(CAST(ES.EXECUTION_ID AS BIGINT))' +
      ' FROM [dbo].[OO_EXECUTION_SUMMARY] ES WITH (NOLOCK)' +
      ' JOIN (SELECT MIN(CAST(EXECUTION_ID AS BIGINT)) MIN_EXEC_ID, MAX(CAST(EXECUTION_ID AS BIGINT)) MAX_EXEC_ID FROM [dbo].[OO_STEP_LOG_STARTED] WITH (NOLOCK)) SLSTR' +
      ' ON ( CAST(ES.EXECUTION_ID AS BIGINT) BETWEEN SLSTR.MIN_EXEC_ID AND SLSTR.MAX_EXEC_ID )' +
      ' WHERE ES.START_TIME < @last_start_timeIN' +
      ' AND ES.STATUS IN (' + @purged_flow_statuses + ')';

  IF (@ShouldPurgeExecutionSummary = 1) BEGIN
    SET @query_parameters = N'@last_start_timeIN DATETIME, @min_exec_summary_id_to_deleteOUT BIGINT OUTPUT, @max_exec_summary_id_to_deleteOUT BIGINT OUTPUT'
    EXECUTE sp_executesql @query1, @query_parameters, @last_start_timeIN = @last_start_time, @min_exec_summary_id_to_deleteOUT = @min_exec_summary_id_to_delete OUTPUT, @max_exec_summary_id_to_deleteOUT = @max_exec_summary_id_to_delete OUTPUT;
  END;
  ELSE BEGIN
    -- In case OO_EXECUTION_SUMMARY should not be purged, limit scope using OO_STEP_LOG_STARTED
    SET @query_parameters = N'@last_start_timeIN DATETIME, @min_exec_summary_id_to_deleteOUT BIGINT OUTPUT, @max_exec_summary_id_to_deleteOUT BIGINT OUTPUT'
    EXECUTE sp_executesql @query2, @query_parameters, @last_start_timeIN = @last_start_time, @min_exec_summary_id_to_deleteOUT = @min_exec_summary_id_to_delete OUTPUT, @max_exec_summary_id_to_deleteOUT = @max_exec_summary_id_to_delete OUTPUT;
  END;

  IF @min_exec_summary_id_to_delete IS NULL
  BEGIN
    SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' Could not find purge-eligible flows.';
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    IF (@DeepClean = 1)
      BEGIN
        GOTO Check_For_Orphans;
      END;
      ELSE BEGIN
        GOTO Summary;
      END;
  END;

  -- Acquire a lock to prevent multiple instances of this procedure to run in parallel
  EXECUTE @lockRetVal = sp_getapplock 'OOPurgeHistoryLock','Exclusive','Session',5000;

  IF (@lockRetVal < 0) BEGIN
    SET @msg = REPLICATE('=', @OutputRowLength);
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    RAISERROR('**** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR ***', 0,1) WITH NOWAIT;
    RAISERROR( '**** Could not aquire lock. There may be another instance of this procedure already running' , 0,1) WITH NOWAIT;
    RAISERROR( '**** Verify purge jobs do not overlap and forecfully  kill zombie processes.' , 0,1) WITH NOWAIT;
    RAISERROR( '**** As a last resort, repeatedly execute the following in order to release pending locks:' , 0,1) WITH NOWAIT;
    RAISERROR( '**** EXEC sp_releaseapplock ''OOPurgeHistoryLock'',''Session''' , 0,1) WITH NOWAIT;
    SET @msg = REPLICATE('=', @OutputRowLength);
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    RETURN 0;
    END;

  -- Beginning of operation ---------------------------------------------------------------------------------
  SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' This procedure will handle flows records older than: ' +
      CAST(DATEADD(day, (-1 * @PurgeExecutionsOlderThan), @purge_start_time) AS VARCHAR);
  RAISERROR( @msg , 0,1) WITH NOWAIT;
  SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' This procedure will stop running around ' + CAST(@end_purge_by AS VARCHAR) + ' even if purge is incomplete';
  RAISERROR( @msg , 0,1) WITH NOWAIT;

  -- Prepare for the the first batch run
  SET @curr_batch_min_id = @min_exec_summary_id_to_delete; -- Starting at the oldest record we need to delete
  SET @curr_batch_max_id = @min_exec_summary_id_to_delete;

  -- Define queries -----------------------------------------------------------------------------------------
  SET @query3 = 'SELECT TOP (@flow_batch_sizeIN) SLS.EXECUTION_ID, SLS.STEP_COUNT' +
               ' FROM ( SELECT EXECUTION_ID, COUNT(EXECUTION_ID) STEP_COUNT FROM [dbo].[OO_STEP_LOG_STARTED] WITH (NOLOCK)' +
               ' WHERE CAST(EXECUTION_ID AS BIGINT) BETWEEN @curr_batch_min_idIN AND @max_exec_summary_id_to_deleteIN GROUP BY EXECUTION_ID) SLS' +
               ' JOIN [dbo].[OO_EXECUTION_SUMMARY] ES WITH (NOLOCK) ON SLS.EXECUTION_ID = ES.EXECUTION_ID' +
               ' WHERE ES.STATUS IN (' + @purged_flow_statuses + ')' +
               ' AND ES.START_TIME < @last_start_timeIN ORDER BY SLS.EXECUTION_ID ASC';

  SET @query4 = N'DELETE [dbo].[OO_EXECUTION_SUMMARY] WITH (ROWLOCK,READPAST)' +
                ' WHERE CAST(EXECUTION_ID AS BIGINT) BETWEEN @curr_batch_min_idIN AND @curr_batch_max_idIN' +
                ' AND STATUS IN (' + @purged_flow_statuses + ') AND START_TIME < @last_start_timeIN';

  SET @query5 = N'DELETE TOP (@delete_batch_sizeIN) [dbo].[OO_EXECUTION_SUMMARY] WITH (ROWLOCK,READPAST)' +
                ' WHERE EXECUTION_ID NOT IN' +
                ' (SELECT DISTINCT(EXECUTION_ID) FROM [dbo].[OO_STEP_LOG_STARTED] WITH (NOLOCK))' +
                ' AND STATUS IN (' + @purged_flow_statuses + ') AND START_TIME < @last_start_timeIN';

  ---------------------------------------------------------------------------------------------------------
  --  Remove child-less OO_EXECUTION_SUMMARY records if applicable                                       --
  --  -------------------------------------------------------------                                      --
  --  If OO_EXECUTION_SUMMARY contains records that do not have corresponding OO_STEP_LOG_STARTED data   --
  --  repeatedly purge these records                                                                     --
  ---------------------------------------------------------------------------------------------------------
  IF (@ShouldPurgeExecutionSummary = 1) BEGIN
    SET @temp_timestamp = GETDATE(); -- Start measuring time for OO_EXECUTION_SUMMARY child-less records purge
    SET @deleted_rows_in_current_batch = 0;

    SET @gen_prps_countr = @delete_batch_size; -- Qualify 'while' condition for the first round
    WHILE (@gen_prps_countr >= @delete_batch_size)
    BEGIN

      BEGIN TRY
      BEGIN TRANSACTION;
        SET @query_parameters = N'@delete_batch_sizeIN INTEGER, @last_start_timeIN DATETIME';
        EXECUTE sp_executesql @query5, @query_parameters, @delete_batch_sizeIN = @delete_batch_size, @last_start_timeIN = @last_start_time;
        SET @gen_prps_countr = @@ROWCOUNT;
        SET @deleted_rows_in_current_batch = @deleted_rows_in_current_batch + @gen_prps_countr;
        SET @rows_del_frm_es = @rows_del_frm_es + @gen_prps_countr;
        SET @total_num_of_flows_handled = @rows_del_frm_es;
      COMMIT TRANSACTION;
      END TRY
      BEGIN CATCH -- Error handling - in case the transaction has failed
        EXECUTE OOGetErrorInfo; -- Execute error retrieval routine.
        IF (XACT_STATE()) = -1 BEGIN
          SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' ERROR: OO_EXECUTION_SUMMARY table is busy, could not lock it for deletion.';
          RAISERROR(@msg, 0,1) WITH NOWAIT;
          ROLLBACK TRANSACTION;
          END;
        IF (XACT_STATE()) = 1
          COMMIT TRANSACTION;
      END CATCH;

      SET @n_o_w = GETDATE();
      IF (@n_o_w > @end_purge_by) BEGIN
        RAISERROR('------ Time limit reached ---------' , 0,1) WITH NOWAIT;
        GOTO Summary;
      END;

    END; -- End of child-less OO_EXECUTION_SUMMARY records purge WHILE loop

    IF (@Verbose >= 4 AND @deleted_rows_in_current_batch > 0) BEGIN
      SET @elapsed_seconds = DATEPART(SECOND,(GETDATE()-@temp_timestamp))+(DATEPART(MINUTE,(GETDATE()-@temp_timestamp))*60);
      IF (@elapsed_seconds > 0) BEGIN
        SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + CAST(@deleted_rows_in_current_batch AS VARCHAR) + ' child-less OO_EXECUTION_SUMMARY records in ' + CAST(@elapsed_seconds AS VARCHAR) + ' seconds (' + CAST(ROUND(@deleted_rows_in_current_batch/@elapsed_seconds,0) AS VARCHAR) + ' row/sec)';
        RAISERROR(@msg, 0,1) WITH NOWAIT;
      END;
      ELSE BEGIN
        SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + CAST(@deleted_rows_in_current_batch AS VARCHAR) + ' child-less OO_EXECUTION_SUMMARY records';
        RAISERROR(@msg, 0,1) WITH NOWAIT;
      END;
    END;
  END;

  --=======================================================================================================--
  -- Main loop                                                                                             --
  --=======================================================================================================--
  WHILE ( @curr_batch_min_id <= @max_exec_summary_id_to_delete )
  BEGIN

    SET @n_o_w = GETDATE();
    IF (@n_o_w > @end_purge_by) BEGIN
      RAISERROR('------ Time limit reached ---------' , 0,1) WITH NOWAIT;
      GOTO Summary;
    END;

    ---------------------------------------------------------------------------------------------------------
    -- Get the next batch of flows and corresponding step count into the @sl_strtd_stps collection    --
    -- The number of elements is between 0 and vFlowBatchSize                                              --
    ---------------------------------------------------------------------------------------------------------
    SET @temp_timestamp = GETDATE(); -- Start measuring time for OO_EXECUTION_SUMMARY child-less records purge

    DELETE FROM @sl_strtd_stps;
    SET @query_parameters = N'@flow_batch_sizeIN INTEGER, @curr_batch_min_idIN BIGINT, @max_exec_summary_id_to_deleteIN BIGINT, @last_start_timeIN DATETIME';
    INSERT INTO @sl_strtd_stps EXECUTE sp_executesql @query3, @query_parameters, @flow_batch_sizeIN = @flow_batch_size, @curr_batch_min_idIN = @curr_batch_min_id, @max_exec_summary_id_to_deleteIN = @max_exec_summary_id_to_delete, @last_start_timeIN = @last_start_time;

    IF NOT EXISTS (SELECT COUNT(1) FROM @sl_strtd_stps) BEGIN
      SET @curr_set_size = 0;
    END;
    ELSE BEGIN
      SELECT @curr_set_size = COUNT(1) FROM @sl_strtd_stps;
    END;

    ---------------------------------------------------------------------------------------------------------
    -- If there are no more flows to handle - exit main WHILE loop                                         --
    ---------------------------------------------------------------------------------------------------------
    IF (@curr_set_size IS NULL) OR (@curr_set_size < 1) BEGIN
      SET @curr_batch_min_id = @max_exec_summary_id_to_delete + 1; -- Qualify exit condition just in case
      BREAK
    END;

    IF (@Verbose >= 4) BEGIN
      SET @elapsed_seconds = DATEPART(SECOND,(GETDATE()-@temp_timestamp))+(DATEPART(MINUTE,(GETDATE()-@temp_timestamp))*60);
      IF (@elapsed_seconds > 0) BEGIN
        SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Done populating @sl_strtd_stps. ' + CAST(@curr_set_size AS VARCHAR) + ' records collected in ' + CAST(@elapsed_seconds AS VARCHAR) + ' seconds';
        RAISERROR(@msg, 0,1) WITH NOWAIT;
      END;
      ELSE BEGIN
        SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Done populating @sl_strtd_stps. ' + CAST(@curr_set_size AS VARCHAR) + ' records collected';
        RAISERROR(@msg, 0,1) WITH NOWAIT;
      END;
    END;

    ---------------------------------------------------------------------------------------------------------
    -- Set the current batch higher execution ID pointer                                                   --
    ---------------------------------------------------------------------------------------------------------
    IF (EXISTS (SELECT COUNT(1) FROM @sl_strtd_stps)) AND (@curr_set_size > 0) BEGIN
      SELECT @temp_idx = MAX(v_idx) FROM @sl_strtd_stps;
      SELECT @curr_batch_max_id = CAST(v_execution_id AS BIGINT) FROM @sl_strtd_stps WHERE v_idx = @temp_idx;
    END;

    ---------------------------------------------------------------------------------------------------------
    -- At this point @sl_strtd_stps contains a list of flows and their corresponding number of steps       --
    -- @sl_strtd_stps contains between 1 and @PurgeExecutionsInBatchesOf records                           --
    ---------------------------------------------------------------------------------------------------------
    IF (@curr_set_size > 0) BEGIN

      SET @curr_set_accum_step_count = 0;
      IF (@ShouldPurgeExecutionSummary = 0) BEGIN
        SET @total_num_of_flows_handled = @total_num_of_flows_handled + @curr_set_size;
      END;

      -- Disable indexes if applicable ----------------------------------------------------------------
      IF (@DisableIndexes = 1) AND (@are_constraints_disabled = 0)
      BEGIN
        SET @are_constraints_disabled = 1;
        ALTER INDEX STEP_LOG_ST_HASH ON [dbo].[OO_STEP_LOG_STARTED] DISABLE;
        ALTER INDEX STEP_LOG_ST_SQ ON [dbo].[OO_STEP_LOG_STARTED] DISABLE;
        IF (@Verbose >= 0) BEGIN
          SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' Disabling constraints in OO_STEP_LOG_STARTED table';
          RAISERROR( @msg, 0, 1) WITH NOWAIT;
        END;
      END;

      IF (@Verbose >= 2) BEGIN
        SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' INFO: Handling EXECUTION_ID range: ' +
          CAST (@curr_batch_min_id AS VARCHAR) + ' and ' +
          CAST (@curr_batch_max_id AS VARCHAR) + ' ...';
        RAISERROR( @msg, 0, 1) WITH NOWAIT;
      END;

      -- v_idx is an identity column - not re-seeding each time ...
      SELECT @curr_set_idx = MIN(v_idx) FROM @sl_strtd_stps;
      SET @idx_shift = @curr_set_idx;
      SET @max_idx_value = @curr_set_idx + @curr_set_size - 1;

      -------------------------------------------------------------------------------------------------------
      -- The following loop breaks down all flows in @sl_strtd_stps, into smaller groups,
      -- accumulating and handling @PurgeExecutionsInBatchesOf steps (or a bit more)
      -- to be handled together
      -------------------------------------------------------------------------------------------------------
      WHILE (@curr_set_idx <= @max_idx_value) BEGIN

        -- Accumulate step counts
        SET @curr_step_count = NULL;
        SET @curr_step_count = (SELECT v_step_count FROM @sl_strtd_stps WHERE v_idx = @curr_set_idx);
        SET @curr_set_accum_step_count =  @curr_set_accum_step_count + @curr_step_count;

        -- Get the minimal EXECUTION_ID index for this group of executions
        IF @curr_set_min_idx IS NULL
          SET @curr_set_min_idx = @curr_set_idx;

        -- Get the current maximal EXECUTION_ID index for this group of executions
        SET @curr_set_max_idx = @curr_set_idx;

        -- @PurgeExecutionsInBatchesOf (or a bit more) steps are accumulated, or this is the last round for this batch
        IF ((@curr_set_accum_step_count >= @PurgeExecutionsInBatchesOf) OR (@curr_set_idx - @idx_shift + 1 = @curr_set_size)) BEGIN

          -------------------------------------------------------------------------------------------------------
          -- All conditions are met - we can finally begin deleting records
          -------------------------------------------------------------------------------------------------------

          -- Alert that this flow has an unusually large number of steps and derivatives
          IF (@curr_step_count >= @lots_of_steps_threshold) AND (@Verbose >= 2) BEGIN
            SELECT @curr_set_max_id = CAST(v_execution_id AS BIGINT) FROM @sl_strtd_stps WHERE v_idx = @curr_set_max_idx;
            SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Note - step count for EXECUTION_ID ' + CAST(@curr_set_max_id AS VARCHAR) + ' is relatively high: ' + REPLACE(CONVERT(varchar(50),(CAST(@curr_step_count AS money)), 1), '.00', '');
            RAISERROR( @msg, 0, 1) WITH NOWAIT;
          END;

          -- Prepare/empty buffers for this batch of flows ...
          DELETE FROM @stps_hash_tab;
          DELETE FROM @stps_ended_tab;

          ----------------------------------------------------------------------------------------
          -- Deleting from OO_STEP_LOG_STARTED                                                  --
          ----------------------------------------------------------------------------------------
          SET @gen_prps_countr = @delete_batch_size; -- Qualify 'while' condition for the first round
          SET @temp_timestamp = GETDATE(); -- Start measuring time for deleting OO_STEP_LOG_STARTED records
          SET @deleted_rows_in_current_batch = 0;
          WHILE (@gen_prps_countr >= @delete_batch_size) BEGIN
            SET @gen_prps_countr = 0;
            BEGIN TRY
            BEGIN TRANSACTION;
              DELETE TOP (@delete_batch_size) [dbo].[OO_STEP_LOG_STARTED]
              WITH (ROWLOCK,READPAST)
              OUTPUT DELETED.STEP_HASH_ID INTO @stps_hash_tab
              WHERE EXECUTION_ID IN (
                SELECT v_execution_id FROM @sl_strtd_stps
                WHERE v_idx BETWEEN @curr_set_min_idx AND @curr_set_max_idx);
            SET @gen_prps_countr = @@ROWCOUNT;
            SET @rows_del_frm_slstr = @rows_del_frm_slstr + @gen_prps_countr;
            SET @deleted_rows_in_current_batch = @deleted_rows_in_current_batch + @gen_prps_countr;
            COMMIT TRANSACTION;
            END TRY
            BEGIN CATCH -- Error handling - in case the transaction has failed
            EXECUTE OOGetErrorInfo; -- Execute error retrieval routine.
            IF (XACT_STATE()) = -1 BEGIN
              SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' ERROR: OO_STEP_LOG_STARTED table is busy, could not lock it for deletion.';
              RAISERROR(@msg, 0,1) WITH NOWAIT;
              ROLLBACK TRANSACTION;
              END;
            IF (XACT_STATE()) = 1
              COMMIT TRANSACTION;
            END CATCH;

            IF (@Verbose >= 3) AND (@gen_prps_countr > 0) BEGIN
              SET @msg = ' - - ' + CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + REPLACE(CONVERT(varchar(50),(CAST(@gen_prps_countr AS money)), 1), '.00', '') + ' records from OO_STEP_LOG_STARTED';
              RAISERROR( @msg, 0, 1) WITH NOWAIT;
            END;

            SET @n_o_w = GETDATE();
            IF (@n_o_w > @end_purge_by) BEGIN
              RAISERROR('------ Time limit reached ---------' , 0,1) WITH NOWAIT;
              GOTO Summary;
            END;

          END; -- End of iteratively deleting records from OO_STEP_LOG_STARTED (while loop)

          IF (@Verbose >= 4 AND @deleted_rows_in_current_batch > 0) BEGIN
            SET @elapsed_seconds = DATEPART(SECOND,(GETDATE()-@temp_timestamp))+(DATEPART(MINUTE,(GETDATE()-@temp_timestamp))*60);
            IF (@elapsed_seconds > 0) BEGIN
              SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + CAST(@deleted_rows_in_current_batch AS VARCHAR) + ' records from OO_STEP_LOG_STARTED in ' + CAST(@elapsed_seconds AS VARCHAR) + ' seconds (' + CAST(ROUND(@deleted_rows_in_current_batch/@elapsed_seconds,0) AS VARCHAR) + ' row/sec)';
              RAISERROR(@msg, 0,1) WITH NOWAIT;
            END;
            ELSE BEGIN
              SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + CAST(@deleted_rows_in_current_batch AS VARCHAR) + ' records from OO_STEP_LOG_STARTED';
              RAISERROR(@msg, 0,1) WITH NOWAIT;
            END;
          END;

          ----------------------------------------------------------------------------------------
          -- Deleting from OO_STEP_LOG_ENDED                                                    --
          ----------------------------------------------------------------------------------------
          SET @gen_prps_countr = @delete_batch_size; -- Qualify 'while' condition for the first round
          SET @temp_timestamp = GETDATE(); -- Start measuring time for deleting OO_STEP_LOG_ENDED records
          SET @deleted_rows_in_current_batch = 0;
          WHILE (@gen_prps_countr >= @delete_batch_size) BEGIN
            SET @gen_prps_countr = 0;
            BEGIN TRY
            BEGIN TRANSACTION;
              DELETE TOP (@delete_batch_size) [dbo].[OO_STEP_LOG_ENDED]
              OUTPUT DELETED.ID INTO @stps_ended_tab
              WHERE STEP_ENDED_HASH_ID IN (SELECT v_step_hash_id FROM @stps_hash_tab);
            SET @gen_prps_countr = @@ROWCOUNT;
            SET @rows_del_frm_sle = @rows_del_frm_sle + @gen_prps_countr;
            SET @deleted_rows_in_current_batch = @deleted_rows_in_current_batch + @gen_prps_countr;
            COMMIT TRANSACTION; -- End of current transaction
            END TRY
            BEGIN CATCH -- Error handling - in case the transaction has failed
              EXECUTE OOGetErrorInfo; -- Execute error retrieval routine.
            IF (XACT_STATE()) = -1 BEGIN
              SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' ERROR: OO_STEP_LOG_ENDED table is busy, could not lock it for deletion.';
              RAISERROR(@msg, 0,1) WITH NOWAIT;
              ROLLBACK TRANSACTION;
            END;
            IF (XACT_STATE()) = 1
              COMMIT TRANSACTION;
            END CATCH;

            IF (@Verbose >= 3) AND (@gen_prps_countr > 0) BEGIN
              SET @msg = ' - - ' + CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + REPLACE(CONVERT(varchar(50),(CAST(@gen_prps_countr AS money)), 1), '.00', '') + ' records from OO_STEP_LOG_ENDED';
              RAISERROR( @msg, 0, 1) WITH NOWAIT;
            END;

            SET @n_o_w = GETDATE();
            IF (@n_o_w > @end_purge_by) BEGIN
              RAISERROR('------ Time limit reached ---------' , 0,1) WITH NOWAIT;
              GOTO Summary;
            END;

          END; -- End of iteratively deleting records from OO_STEP_LOG_ENDED (while loop)

          IF (@Verbose >= 4 AND @deleted_rows_in_current_batch > 0) BEGIN
            SET @elapsed_seconds = DATEPART(SECOND,(GETDATE()-@temp_timestamp))+(DATEPART(MINUTE,(GETDATE()-@temp_timestamp))*60);
            IF (@elapsed_seconds > 0) BEGIN
              SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + CAST(@deleted_rows_in_current_batch AS VARCHAR) + ' records from OO_STEP_LOG_ENDED in ' + CAST(@elapsed_seconds AS VARCHAR) + ' seconds (' + CAST(ROUND(@deleted_rows_in_current_batch/@elapsed_seconds,0) AS VARCHAR) + ' row/sec)';
              RAISERROR(@msg, 0,1) WITH NOWAIT;
            END;
            ELSE BEGIN
              SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + CAST(@deleted_rows_in_current_batch AS VARCHAR) + ' records from OO_STEP_LOG_ENDED';
              RAISERROR(@msg, 0,1) WITH NOWAIT;
            END;
          END;

          ----------------------------------------------------------------------------------------
          -- Deleting from OO_STEP_LOG_BINDINGS                                                 --
          ----------------------------------------------------------------------------------------
          SET @gen_prps_countr = @delete_batch_size; -- Qualify 'while' condition for the first round
          SET @temp_timestamp = GETDATE(); -- Start measuring time for deleting OO_STEP_LOG_BINDINGS records
          SET @deleted_rows_in_current_batch = 0;
          WHILE (@gen_prps_countr >= @delete_batch_size) BEGIN
            SET @gen_prps_countr = 0;
            BEGIN TRY
            BEGIN TRANSACTION;
              DELETE TOP (@delete_batch_size) [dbo].[OO_STEP_LOG_BINDINGS]
              WHERE STEP_LOG_ENDED_ID IN (SELECT v_ended_id FROM @stps_ended_tab);
              SET @gen_prps_countr = @@ROWCOUNT;
              SET @rows_del_frm_slb = @rows_del_frm_slb + @gen_prps_countr;
              SET @deleted_rows_in_current_batch = @deleted_rows_in_current_batch + @gen_prps_countr;
            COMMIT TRANSACTION; -- End of current transaction
            END TRY
            BEGIN CATCH -- Error handling - in case the transaction has failed
              EXECUTE OOGetErrorInfo; -- Execute error retrieval routine.
              IF (XACT_STATE()) = -1 BEGIN
                SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' ERROR: OO_STEP_LOG_BINDINGS table is busy, could not lock it for deletion.';
                RAISERROR(@msg, 0,1) WITH NOWAIT;
                ROLLBACK TRANSACTION;
              END;
              IF (XACT_STATE()) = 1
                COMMIT TRANSACTION;
              END CATCH;

            IF (@Verbose >= 3) AND (@gen_prps_countr > 0) BEGIN
              SET @msg = ' - - ' + CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + REPLACE(CONVERT(varchar(50),(CAST(@gen_prps_countr AS money)), 1), '.00', '') + ' records from OO_STEP_LOG_BINDINGS';
              RAISERROR( @msg, 0, 1) WITH NOWAIT;
            END;

            SET @n_o_w = GETDATE();
            IF (@n_o_w > @end_purge_by) BEGIN
              RAISERROR('------ Time limit reached ---------' , 0,1) WITH NOWAIT;
              GOTO Summary;
            END;

          END; -- End of iteratively deleting records from OO_STEP_LOG_BINDINGS (while loop)

          IF (@Verbose >= 4 AND @deleted_rows_in_current_batch > 0) BEGIN
            SET @elapsed_seconds = DATEPART(SECOND,(GETDATE()-@temp_timestamp))+(DATEPART(MINUTE,(GETDATE()-@temp_timestamp))*60);
            IF (@elapsed_seconds > 0) BEGIN
              SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + CAST(@deleted_rows_in_current_batch AS VARCHAR) + ' records from OO_STEP_LOG_BINDINGS in ' + CAST(@elapsed_seconds AS VARCHAR) + ' seconds (' + CAST(ROUND(@deleted_rows_in_current_batch/@elapsed_seconds,0) AS VARCHAR) + ' row/sec)';
              RAISERROR(@msg, 0,1) WITH NOWAIT;
            END;
            ELSE BEGIN
              SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + CAST(@deleted_rows_in_current_batch AS VARCHAR) + ' records from OO_STEP_LOG_BINDINGS';
              RAISERROR(@msg, 0,1) WITH NOWAIT;
            END;
          END;

          ----------------------------------------------------------------------------------------
          -- Deleting from OO_EXECUTION_BOUND_OUTPUTS                                           --
          ----------------------------------------------------------------------------------------
          SET @gen_prps_countr = @delete_batch_size; -- Qualify 'while' condition for the first round
          SET @temp_timestamp = GETDATE(); -- Start measuring time for deleting OO_EXECUTION_BOUND_OUTPUTS records
          SET @deleted_rows_in_current_batch = 0;
          WHILE (@gen_prps_countr >= @delete_batch_size) BEGIN
            SET @gen_prps_countr = 0;
            BEGIN TRY
            BEGIN TRANSACTION;
              DELETE TOP (@delete_batch_size) [dbo].[OO_EXECUTION_BOUND_OUTPUTS]
                WHERE EXECUTION_ID IN (SELECT v_execution_id FROM @sl_strtd_stps
                WHERE v_idx BETWEEN @curr_set_min_idx AND @curr_set_max_idx);
              SET @gen_prps_countr = @@ROWCOUNT;
              SET @rows_del_frm_bo = @rows_del_frm_bo + @gen_prps_countr;
              SET @deleted_rows_in_current_batch = @deleted_rows_in_current_batch + @gen_prps_countr;
            COMMIT TRANSACTION; -- End of current transaction
            END TRY
            BEGIN CATCH -- Error handling - in case the transaction has failed
            EXECUTE OOGetErrorInfo; -- Execute error retrieval routine.
            IF (XACT_STATE()) = -1 BEGIN
              SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' ERROR: OO_EXECUTION_BOUND_OUTPUTS table is busy, could not lock it for deletion.';
              RAISERROR(@msg, 0,1) WITH NOWAIT;
              ROLLBACK TRANSACTION;
            END;
            IF (XACT_STATE()) = 1
              COMMIT TRANSACTION;
            END CATCH;

            IF (@Verbose >= 3) AND (@gen_prps_countr > 0) BEGIN
              SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + REPLACE(CONVERT(varchar(50),(CAST(@gen_prps_countr AS money)), 1), '.00', '') + ' records from OO_EXECUTION_BOUND_OUTPUTS';
                RAISERROR( @msg, 0, 1) WITH NOWAIT;
              END;

            SET @n_o_w = GETDATE();
            IF (@n_o_w > @end_purge_by) BEGIN
              RAISERROR('------ Time limit reached ---------' , 0,1) WITH NOWAIT;
              GOTO Summary;
            END;

          END; -- End of iteratively deleting records from OO_EXECUTION_BOUND_OUTPUTS (while loop)

          IF (@Verbose >= 4 AND @deleted_rows_in_current_batch > 0) BEGIN
            SET @elapsed_seconds = DATEPART(SECOND,(GETDATE()-@temp_timestamp))+(DATEPART(MINUTE,(GETDATE()-@temp_timestamp))*60);
            IF (@elapsed_seconds > 0) BEGIN
              SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + CAST(@deleted_rows_in_current_batch AS VARCHAR) + ' records from OO_EXECUTION_BOUND_OUTPUTS in ' + CAST(@elapsed_seconds AS VARCHAR) + ' seconds (' + CAST(ROUND(@deleted_rows_in_current_batch/@elapsed_seconds,0) AS VARCHAR) + ' row/sec)';
              RAISERROR(@msg, 0,1) WITH NOWAIT;
            END;
            ELSE BEGIN
              SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + CAST(@deleted_rows_in_current_batch AS VARCHAR) + ' records from OO_EXECUTION_BOUND_OUTPUTS';
              RAISERROR(@msg, 0,1) WITH NOWAIT;
            END;
          END;

          ----------------------------------------------------------------------------------------
          -- Deleting from OO_EXECUTION_BOUND_INPUTS                                           --
          ----------------------------------------------------------------------------------------
          SET @gen_prps_countr = @delete_batch_size; -- Qualify 'while' condition for the first round
          SET @temp_timestamp = GETDATE(); -- Start measuring time for deleting OO_EXECUTION_BOUND_INPUTS records
          SET @deleted_rows_in_current_batch = 0;
          WHILE (@gen_prps_countr >= @delete_batch_size) BEGIN
            SET @gen_prps_countr = 0;
            BEGIN TRY
            BEGIN TRANSACTION;
              DELETE [dbo].[OO_EXECUTION_BOUND_INPUTS]
                WHERE EXECUTION_ID IN (SELECT v_execution_id FROM @sl_strtd_stps
                WHERE v_idx BETWEEN @curr_set_min_idx AND @curr_set_max_idx);
              SET @gen_prps_countr = @@ROWCOUNT;
              SET @rows_del_frm_bi = @rows_del_frm_bi + @gen_prps_countr;
              SET @deleted_rows_in_current_batch = @deleted_rows_in_current_batch + @gen_prps_countr;
            COMMIT TRANSACTION; -- End of current transaction
            END TRY
            BEGIN CATCH -- Error handling - in case the transaction has failed
            EXECUTE OOGetErrorInfo; -- Execute error retrieval routine.
            IF (XACT_STATE()) = -1
            BEGIN
              SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' ERROR: OO_EXECUTION_BOUND_INPUTS table is busy, could not lock it for deletion.';
              RAISERROR(@msg, 0,1) WITH NOWAIT;
              ROLLBACK TRANSACTION;
            END;
            IF (XACT_STATE()) = 1
              COMMIT TRANSACTION;
            END CATCH;

            IF (@Verbose >= 3) AND (@gen_prps_countr > 0) BEGIN
              SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + REPLACE(CONVERT(varchar(50),(CAST(@gen_prps_countr AS money)), 1), '.00', '') + ' records from OO_EXECUTION_BOUND_INPUTS';
              RAISERROR( @msg, 0, 1) WITH NOWAIT;
            END;

            SET @n_o_w = GETDATE();
            IF (@n_o_w > @end_purge_by) BEGIN
              RAISERROR('------ Time limit reached ---------' , 0,1) WITH NOWAIT;
              GOTO Summary;
            END;

          END; -- End of iteratively deleting records from OO_EXECUTION_BOUND_INPUTS (while loop)

          IF (@Verbose >= 4 AND @deleted_rows_in_current_batch > 0) BEGIN
            SET @elapsed_seconds = DATEPART(SECOND,(GETDATE()-@temp_timestamp))+(DATEPART(MINUTE,(GETDATE()-@temp_timestamp))*60);
            IF (@elapsed_seconds > 0) BEGIN
              SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + CAST(@deleted_rows_in_current_batch AS VARCHAR) + ' records from OO_EXECUTION_BOUND_INPUTS in ' + CAST(@elapsed_seconds AS VARCHAR) + ' seconds (' + CAST(ROUND(@deleted_rows_in_current_batch/@elapsed_seconds,0) AS VARCHAR) + ' row/sec)';
              RAISERROR(@msg, 0,1) WITH NOWAIT;
            END;
            ELSE BEGIN
              SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + CAST(@deleted_rows_in_current_batch AS VARCHAR) + ' records from OO_EXECUTION_BOUND_INPUTS';
              RAISERROR(@msg, 0,1) WITH NOWAIT;
            END;
          END;

          -- Prepare for the next set
          SET @curr_set_accum_step_count = 0;
          SET @curr_set_min_idx = NULL;
          SET @curr_set_max_idx = NULL;

        END; -- End of "IF" we have enough steps to process, or last round

        SET @curr_set_idx = @curr_set_idx + 1;

      END; -- End of "WHILE" - iterating through @sl_strtd_stps

    END; -- End of handling the current flow range

    ----------------------------------------------------------------------------------------
    -- Deleting from OO_EXECUTION_SUMMARY if applicable                                   --
    ----------------------------------------------------------------------------------------
    IF (@ShouldPurgeExecutionSummary = 1) BEGIN

      IF (@Verbose >= 4) BEGIN
        SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: About to delete from OO_EXECUTION_SUMMARY';
        RAISERROR( @msg, 0, 1) WITH NOWAIT;
      END;

      BEGIN TRY
      BEGIN TRANSACTION;
        SET @query_parameters = N'@curr_batch_min_idIN BIGINT, @curr_batch_max_idIN BIGINT, @last_start_timeIN DATETIME';
        EXECUTE sp_executesql @query4, @query_parameters, @curr_batch_min_idIN = @curr_batch_min_id, @curr_batch_max_idIN = @curr_batch_max_id, @last_start_timeIN = @last_start_time;
        SET @gen_prps_countr = @@ROWCOUNT;
        SET @rows_del_frm_es = @rows_del_frm_es + @gen_prps_countr;
        SET @total_num_of_flows_handled = @rows_del_frm_es;
      COMMIT TRANSACTION; -- End of current transaction
      END TRY
      BEGIN CATCH -- Error handling - in case the transaction has failed
        EXECUTE OOGetErrorInfo; -- Execute error retrieval routine.
        IF (XACT_STATE()) = -1 BEGIN
          SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' ERROR: OO_EXECUTION_SUMMARY table is busy, could not lock it for deletion.';
          RAISERROR(@msg, 0,1) WITH NOWAIT;
          ROLLBACK TRANSACTION;
        END;
        IF (XACT_STATE()) = 1
          COMMIT TRANSACTION;
      END CATCH;

      IF (@Verbose >= 3) AND (@gen_prps_countr > 0) BEGIN
        SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + REPLACE(CONVERT(varchar(50),(CAST(@gen_prps_countr AS money)), 1), '.00', '') + ' records from OO_EXECUTION_SUMMARY';
        RAISERROR( @msg, 0, 1) WITH NOWAIT;
      END; -- End of deleting the current range of OO_EXECUTION_SUMMARY

    END; -- Handling OO_EXECUTION_SUMMARY depending on @ShouldPurgeExecutionSummary

    -- Prepare for the next OO_EXECUTION_SUMMARY batch

    -- Advance pointers
    SET @curr_batch_min_id = @curr_batch_max_id + 1;
    SET @curr_batch_max_id = @curr_batch_min_id;

    -- Cutoff at upper limit
    IF @curr_batch_max_id >= @max_exec_summary_id_to_delete
      SET @curr_batch_max_id = @max_exec_summary_id_to_delete + 1; -- Qualify exit condition just in case...

    END -- End of main WHILE loop

  Check_For_Orphans:

  IF (@DeepClean <> 1)
    GOTO Summary;

  ----------------------------------------------------------------------------------------
  -- Deep clean operations, removing orphan records if such are detected                --
  ----------------------------------------------------------------------------------------

  IF (@Verbose >= 1) BEGIN
    SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' Beginning to search for (and remove) orphan records';
    RAISERROR( @msg, 0, 1) WITH NOWAIT;
  END;

  -- Handling orphan OO_STEP_LOG_BINDINGS records
  SET @gen_prps_countr = @delete_batch_size; -- Qualify 'while' condition for the first round
  SET @temp_timestamp = GETDATE(); -- Start measuring time for deleting orphan OO_STEP_LOG_BINDINGS records
  SET @deleted_rows_in_current_batch = 0;
  WHILE (@gen_prps_countr >= @delete_batch_size) BEGIN
    SET @gen_prps_countr = 0;
    BEGIN TRY
    BEGIN TRANSACTION;
      DELETE TOP (@delete_batch_size) [dbo].[OO_STEP_LOG_BINDINGS]
        WHERE STEP_LOG_ENDED_ID NOT IN (SELECT ID FROM [dbo].[OO_STEP_LOG_ENDED] WITH (NOLOCK));
      SET @gen_prps_countr = @@ROWCOUNT;
      SET @rows_del_frm_slb = @rows_del_frm_slb + @gen_prps_countr;
      SET @deleted_rows_in_current_batch = @deleted_rows_in_current_batch + @gen_prps_countr;
    COMMIT TRANSACTION; -- End of current transaction
    END TRY
    BEGIN CATCH -- Error handling - in case the transaction has failed
    EXECUTE OOGetErrorInfo; -- Execute error retrieval routine.
    IF (XACT_STATE()) = -1 BEGIN
      SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' ERROR: OO_STEP_LOG_BINDINGS table is busy, could not lock it for deletion.';
      RAISERROR(@msg, 0,1) WITH NOWAIT;
      ROLLBACK TRANSACTION;
    END;
    IF (XACT_STATE()) = 1
      COMMIT TRANSACTION;
    END CATCH;

    IF (@Verbose >= 3) AND (@gen_prps_countr > 0) BEGIN
      SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + REPLACE(CONVERT(varchar(50),(CAST(@gen_prps_countr AS money)), 1), '.00', '') + ' orphan records from OO_STEP_LOG_BINDINGS';
      RAISERROR( @msg, 0, 1) WITH NOWAIT;
    END;

    SET @n_o_w = GETDATE();
    IF (@n_o_w > @end_purge_by) BEGIN
      RAISERROR('------ Time limit reached ---------' , 0,1) WITH NOWAIT;
      GOTO Summary;
    END;

  END; -- End of handling orphan OO_STEP_LOG_BINDINGS records

  IF (@Verbose >= 4 AND @deleted_rows_in_current_batch > 0) BEGIN
    SET @elapsed_seconds = DATEPART(SECOND,(GETDATE()-@temp_timestamp))+(DATEPART(MINUTE,(GETDATE()-@temp_timestamp))*60);
    IF (@elapsed_seconds > 0) BEGIN
      SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + CAST(@deleted_rows_in_current_batch AS VARCHAR) + ' orphan records from OO_STEP_LOG_BINDINGS in ' + CAST(@elapsed_seconds AS VARCHAR) + ' seconds (' + CAST(ROUND(@deleted_rows_in_current_batch/@elapsed_seconds,0) AS VARCHAR) + ' row/sec)';
      RAISERROR(@msg, 0,1) WITH NOWAIT;
    END;
    ELSE BEGIN
      SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + CAST(@deleted_rows_in_current_batch AS VARCHAR) + ' orphan records from OO_STEP_LOG_BINDINGS';
      RAISERROR(@msg, 0,1) WITH NOWAIT;
    END;
  END;

  -- Handling orphan OO_STEP_LOG_ENDED records
  SET @gen_prps_countr = @delete_batch_size; -- Qualify 'while' condition for the first round
  SET @temp_timestamp = GETDATE(); -- Start measuring time for deleting orphan OO_STEP_LOG_ENDED records
  SET @deleted_rows_in_current_batch = 0;
  WHILE (@gen_prps_countr >= @delete_batch_size) BEGIN
    SET @gen_prps_countr = 0;
    BEGIN TRY
    BEGIN TRANSACTION;
      DELETE TOP (@delete_batch_size) [dbo].[OO_STEP_LOG_ENDED]
        WHERE STEP_ENDED_HASH_ID NOT IN (SELECT STEP_HASH_ID FROM [dbo].[OO_STEP_LOG_STARTED] WITH (NOLOCK));
      SET @gen_prps_countr = @@ROWCOUNT;
      SET @rows_del_frm_sle = @rows_del_frm_sle + @gen_prps_countr;
      SET @deleted_rows_in_current_batch = @deleted_rows_in_current_batch + @gen_prps_countr;
    COMMIT TRANSACTION; -- End of current transaction
    END TRY
    BEGIN CATCH -- Error handling - in case the transaction has failed
    EXECUTE OOGetErrorInfo; -- Execute error retrieval routine.
    IF (XACT_STATE()) = -1 BEGIN
      SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' ERROR: OO_STEP_LOG_ENDED table is busy, could not lock it for deletion.';
      RAISERROR(@msg, 0,1) WITH NOWAIT;
      ROLLBACK TRANSACTION;
    END;
    IF (XACT_STATE()) = 1
      COMMIT TRANSACTION;
    END CATCH;

    IF (@Verbose >= 3) AND (@gen_prps_countr > 0) BEGIN
      SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + REPLACE(CONVERT(varchar(50),(CAST(@gen_prps_countr AS money)), 1), '.00', '') + ' orphan records from OO_STEP_LOG_ENDED';
      RAISERROR( @msg, 0, 1) WITH NOWAIT;
    END;

    SET @n_o_w = GETDATE();
    IF (@n_o_w > @end_purge_by) BEGIN
      RAISERROR('------ Time limit reached ---------' , 0,1) WITH NOWAIT;
      GOTO Summary;
    END;

  END; -- End of handling orphan OO_STEP_LOG_ENDED records

  IF (@Verbose >= 4 AND @deleted_rows_in_current_batch > 0) BEGIN
    SET @elapsed_seconds = DATEPART(SECOND,(GETDATE()-@temp_timestamp))+(DATEPART(MINUTE,(GETDATE()-@temp_timestamp))*60);
    IF (@elapsed_seconds > 0) BEGIN
      SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + CAST(@deleted_rows_in_current_batch AS VARCHAR) + ' orphan records from OO_STEP_LOG_ENDED in ' + CAST(@elapsed_seconds AS VARCHAR) + ' seconds (' + CAST(ROUND(@deleted_rows_in_current_batch/@elapsed_seconds,0) AS VARCHAR) + ' row/sec)';
      RAISERROR(@msg, 0,1) WITH NOWAIT;
    END;
    ELSE BEGIN
      SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + CAST(@deleted_rows_in_current_batch AS VARCHAR) + ' orphan records from OO_STEP_LOG_ENDED';
      RAISERROR(@msg, 0,1) WITH NOWAIT;
    END;
  END;

  -- Handling orphan OO_STEP_LOG_STARTED records
  SET @gen_prps_countr = @delete_batch_size; -- Qualify 'while' condition for the first round
  SET @temp_timestamp = GETDATE(); -- Start measuring time for deleting orphan OO_STEP_LOG_STARTED records
  SET @deleted_rows_in_current_batch = 0;
  WHILE (@gen_prps_countr >= @delete_batch_size) BEGIN
    SET @gen_prps_countr = 0;
    BEGIN TRY
    BEGIN TRANSACTION;
      DELETE TOP (@delete_batch_size) [dbo].[OO_STEP_LOG_STARTED]
        WHERE EXECUTION_ID NOT IN (SELECT EXECUTION_ID FROM [dbo].[OO_EXECUTION_SUMMARY] WITH (NOLOCK));
      SET @gen_prps_countr = @@ROWCOUNT;
      SET @rows_del_frm_slstr = @rows_del_frm_slstr + @gen_prps_countr;
      SET @deleted_rows_in_current_batch = @deleted_rows_in_current_batch + @gen_prps_countr;
    COMMIT TRANSACTION; -- End of current transaction
    END TRY
    BEGIN CATCH -- Error handling - in case the transaction has failed
    EXECUTE OOGetErrorInfo; -- Execute error retrieval routine.
    IF (XACT_STATE()) = -1 BEGIN
      SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' ERROR: OO_STEP_LOG_STARTED table is busy, could not lock it for deletion.';
      RAISERROR(@msg, 0,1) WITH NOWAIT;
      ROLLBACK TRANSACTION;
    END;
    IF (XACT_STATE()) = 1
      COMMIT TRANSACTION;
    END CATCH;

    IF (@Verbose >= 3) AND (@gen_prps_countr > 0) BEGIN
      SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + REPLACE(CONVERT(varchar(50),(CAST(@gen_prps_countr AS money)), 1), '.00', '') + ' orphan records from OO_STEP_LOG_STARTED';
      RAISERROR( @msg, 0, 1) WITH NOWAIT;
    END;

    SET @n_o_w = GETDATE();
    IF (@n_o_w > @end_purge_by) BEGIN
      RAISERROR('------ Time limit reached ---------' , 0,1) WITH NOWAIT;
      GOTO Summary;
    END;

  END; -- End of handling orphan OO_STEP_LOG_STARTED records

  IF (@Verbose >= 4 AND @deleted_rows_in_current_batch > 0) BEGIN
    SET @elapsed_seconds = DATEPART(SECOND,(GETDATE()-@temp_timestamp))+(DATEPART(MINUTE,(GETDATE()-@temp_timestamp))*60);
    IF (@elapsed_seconds > 0) BEGIN
      SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + CAST(@deleted_rows_in_current_batch AS VARCHAR) + ' orphan records from OO_STEP_LOG_STARTED in ' + CAST(@elapsed_seconds AS VARCHAR) + ' seconds (' + CAST(ROUND(@deleted_rows_in_current_batch/@elapsed_seconds,0) AS VARCHAR) + ' row/sec)';
      RAISERROR(@msg, 0,1) WITH NOWAIT;
    END;
    ELSE BEGIN
      SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + CAST(@deleted_rows_in_current_batch AS VARCHAR) + ' orphan records from OO_STEP_LOG_STARTED';
      RAISERROR(@msg, 0,1) WITH NOWAIT;
    END;
  END;

  -- Handling orphan OO_EXECUTION_BOUND_INPUTS records
  SET @gen_prps_countr = @delete_batch_size; -- Qualify 'while' condition for the first round
  SET @temp_timestamp = GETDATE(); -- Start measuring time for deleting orphan OO_EXECUTION_BOUND_INPUTS records
  SET @deleted_rows_in_current_batch = 0;
  WHILE (@gen_prps_countr >= @delete_batch_size) BEGIN
    SET @gen_prps_countr = 0;
    BEGIN TRY
    BEGIN TRANSACTION;
      DELETE TOP (@delete_batch_size) [dbo].[OO_EXECUTION_BOUND_INPUTS]
        WHERE EXECUTION_ID NOT IN (SELECT EXECUTION_ID FROM [dbo].[OO_EXECUTION_SUMMARY] WITH (NOLOCK));
      SET @gen_prps_countr = @@ROWCOUNT;
      SET @rows_del_frm_bi = @rows_del_frm_bi + @gen_prps_countr;
      SET @deleted_rows_in_current_batch = @deleted_rows_in_current_batch + @gen_prps_countr;
      COMMIT TRANSACTION; -- End of current transaction
    END TRY
    BEGIN CATCH -- Error handling - in case the transaction has failed
      EXECUTE OOGetErrorInfo; -- Execute error retrieval routine.
      IF (XACT_STATE()) = -1
      BEGIN
          SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' ERROR: OO_EXECUTION_BOUND_INPUTS table is busy, could not lock it for deletion.';
          RAISERROR(@msg, 0,1) WITH NOWAIT;
          ROLLBACK TRANSACTION;
      END;
      IF (XACT_STATE()) = 1
          COMMIT TRANSACTION;
    END CATCH;

    IF (@Verbose >= 3) AND (@gen_prps_countr > 0) BEGIN
      SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + REPLACE(CONVERT(varchar(50),(CAST(@gen_prps_countr AS money)), 1), '.00', '') + ' orphan records from OO_EXECUTION_BOUND_INPUTS';
      RAISERROR( @msg, 0, 1) WITH NOWAIT;
    END;

    SET @n_o_w = GETDATE();
    IF (@n_o_w > @end_purge_by) BEGIN
      RAISERROR('------ Time limit reached ---------' , 0,1) WITH NOWAIT;
      GOTO Summary;
    END;

  END; -- End of handling orphan OO_EXECUTION_BOUND_INPUTS records

  IF (@Verbose >= 4 AND @deleted_rows_in_current_batch > 0) BEGIN
    SET @elapsed_seconds = DATEPART(SECOND,(GETDATE()-@temp_timestamp))+(DATEPART(MINUTE,(GETDATE()-@temp_timestamp))*60);
    IF (@elapsed_seconds > 0) BEGIN
      SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + CAST(@deleted_rows_in_current_batch AS VARCHAR) + ' orphan records from OO_EXECUTION_BOUND_INPUTS in ' + CAST(@elapsed_seconds AS VARCHAR) + ' seconds (' + CAST(ROUND(@deleted_rows_in_current_batch/@elapsed_seconds,0) AS VARCHAR) + ' row/sec)';
      RAISERROR(@msg, 0,1) WITH NOWAIT;
    END;
    ELSE BEGIN
      SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + CAST(@deleted_rows_in_current_batch AS VARCHAR) + ' orphan records from OO_EXECUTION_BOUND_INPUTS';
      RAISERROR(@msg, 0,1) WITH NOWAIT;
    END;
  END;

  -- Handling orphan OO_EXECUTION_BOUND_OUTPUTS records
  SET @gen_prps_countr = @delete_batch_size; -- Qualify 'while' condition for the first round
  SET @temp_timestamp = GETDATE(); -- Start measuring time for deleting orphan OO_EXECUTION_BOUND_OUTPUTS records
  SET @deleted_rows_in_current_batch = 0;
  WHILE (@gen_prps_countr >= @delete_batch_size) BEGIN
    SET @gen_prps_countr = 0;
    BEGIN TRY
    BEGIN TRANSACTION;
      DELETE TOP (@delete_batch_size) [dbo].[OO_EXECUTION_BOUND_OUTPUTS]
        WHERE EXECUTION_ID NOT IN (SELECT EXECUTION_ID FROM [dbo].[OO_EXECUTION_SUMMARY] WITH (NOLOCK));
      SET @gen_prps_countr = @@ROWCOUNT;
      SET @rows_del_frm_bo = @rows_del_frm_bo + @gen_prps_countr;
      SET @deleted_rows_in_current_batch = @deleted_rows_in_current_batch + @gen_prps_countr;
      COMMIT TRANSACTION; -- End of current transaction
    END TRY
    BEGIN CATCH -- Error handling - in case the transaction has failed
      EXECUTE OOGetErrorInfo; -- Execute error retrieval routine.
      IF (XACT_STATE()) = -1
      BEGIN
          SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' ERROR: OO_EXECUTION_BOUND_OUTPUTS table is busy, could not lock it for deletion.';
          RAISERROR(@msg, 0,1) WITH NOWAIT;
          ROLLBACK TRANSACTION;
      END;
      IF (XACT_STATE()) = 1
          COMMIT TRANSACTION;
    END CATCH;

    IF (@Verbose >= 3) AND (@gen_prps_countr > 0) BEGIN
      SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + REPLACE(CONVERT(varchar(50),(CAST(@gen_prps_countr AS money)), 1), '.00', '') + ' orphan records from OO_EXECUTION_BOUND_OUTPUTS';
      RAISERROR( @msg, 0, 1) WITH NOWAIT;
    END;

    SET @n_o_w = GETDATE();
    IF (@n_o_w > @end_purge_by) BEGIN
      RAISERROR('------ Time limit reached ---------' , 0,1) WITH NOWAIT;
      GOTO Summary;
    END;

  END; -- End of handling orphan OO_EXECUTION_BOUND_OUTPUTS records

  IF (@Verbose >= 4 AND @deleted_rows_in_current_batch > 0) BEGIN
    SET @elapsed_seconds = DATEPART(SECOND,(GETDATE()-@temp_timestamp))+(DATEPART(MINUTE,(GETDATE()-@temp_timestamp))*60);
    IF (@elapsed_seconds > 0) BEGIN
      SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + CAST(@deleted_rows_in_current_batch AS VARCHAR) + ' orphan records from OO_EXECUTION_BOUND_OUTPUTS in ' + CAST(@elapsed_seconds AS VARCHAR) + ' seconds (' + CAST(ROUND(@deleted_rows_in_current_batch/@elapsed_seconds,0) AS VARCHAR) + ' row/sec)';
      RAISERROR(@msg, 0,1) WITH NOWAIT;
    END;
    ELSE BEGIN
      SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Deleted ' + CAST(@deleted_rows_in_current_batch AS VARCHAR) + ' orphan records from OO_EXECUTION_BOUND_OUTPUTS';
      RAISERROR(@msg, 0,1) WITH NOWAIT;
    END;
  END;

  -- Summary
  Summary:

  IF (@DisableIndexes = 1) AND (@are_constraints_disabled = 1)
  BEGIN
    SET @temp_timestamp = GETDATE(); -- Start measuring time for index rebuild
    IF (@Verbose >= 0) BEGIN
      SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' Rebuilding constraints in OO_STEP_LOG_STARTED table';
      RAISERROR( @msg, 0, 1) WITH NOWAIT;
    END;
    ALTER INDEX ALL ON [dbo].[OO_STEP_LOG_STARTED] REBUILD;
  END;

  IF (@Verbose >= 4) BEGIN
    SET @elapsed_seconds = DATEPART(SECOND,(GETDATE()-@temp_timestamp))+(DATEPART(MINUTE,(GETDATE()-@temp_timestamp))*60);
    IF (@elapsed_seconds > 0) BEGIN
      SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' DEBUG: Indexes rebuilt in ' + CAST(@elapsed_seconds AS VARCHAR) + ' seconds';
      RAISERROR(@msg, 0,1) WITH NOWAIT;
    END;
  END;

  IF (@Verbose > 0) BEGIN
    SET @msg = CONVERT(VARCHAR,GETDATE(),108) + ' INFO: Total operation time: '
    + CONVERT(VARCHAR,GETDATE()-@purge_start_time,108);
    RAISERROR(@msg, 0,1) WITH NOWAIT;
  END;

  SET @total_rows_deleted = @rows_del_frm_es + @rows_del_frm_slb + @rows_del_frm_sle + @rows_del_frm_slstr + @rows_del_frm_bo + @rows_del_frm_bi;

  IF (@Verbose > 1 AND @total_rows_deleted > 0) BEGIN
    RAISERROR('' , 0,1) WITH NOWAIT;
    SET @msg = 'SUMMARY: Deleted ' + REPLACE(CONVERT(varchar(50),(CAST(@rows_del_frm_es AS money)), 1), '.00', '') + ' rows from OO_EXECUTION_SUMMARY';
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    SET @msg = 'SUMMARY: Deleted ' + REPLACE(CONVERT(varchar(50),(CAST(@rows_del_frm_slstr AS money)), 1), '.00', '') + ' rows from OO_STEP_LOG_STARTED';
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    SET @msg = 'SUMMARY: Deleted ' + REPLACE(CONVERT(varchar(50),(CAST(@rows_del_frm_sle AS money)), 1), '.00', '') + ' rows from OO_STEP_LOG_ENDED';
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    SET @msg = 'SUMMARY: Deleted ' + REPLACE(CONVERT(varchar(50),(CAST(@rows_del_frm_slb AS money)), 1), '.00', '') + ' rows from OO_STEP_LOG_BINDINGS';
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    SET @msg = 'SUMMARY: Deleted ' + REPLACE(CONVERT(varchar(50),(CAST(@rows_del_frm_bi AS money)), 1), '.00', '') + ' rows from OO_EXECUTION_BOUND_INPUTS';
    RAISERROR(@msg, 0,1) WITH NOWAIT;
    SET @msg = 'SUMMARY: Deleted ' + REPLACE(CONVERT(varchar(50),(CAST(@rows_del_frm_bo AS money)), 1), '.00', '') + ' rows from OO_EXECUTION_BOUND_OUTPUTS';
    RAISERROR(@msg, 0,1) WITH NOWAIT;
  END;

  IF (@Verbose >= 1) BEGIN
    IF (@total_rows_deleted > 0) BEGIN
      RAISERROR('' , 0,1) WITH NOWAIT;
      SET @msg = 'SUMMARY:  A total of ' + REPLACE(CONVERT(varchar(50),(CAST(@total_rows_deleted AS money)), 1), '.00', '') + ' rows deleted from database';
      RAISERROR(@msg, 0,1) WITH NOWAIT;
      SET @msg = '          A total of ' + REPLACE(CONVERT(varchar(50),(CAST(@total_num_of_flows_handled AS money)), 1), '.00', '') + ' flows were handled';
      RAISERROR(@msg, 0,1) WITH NOWAIT;
    END;
    ELSE BEGIN
      RAISERROR('--------------------------------------', 0,1) WITH NOWAIT;
      RAISERROR('SUMMARY: No rows deleted from database', 0,1) WITH NOWAIT;
      RAISERROR('--------------------------------------', 0,1) WITH NOWAIT;
    END;
  END;

  -- Release the multiple instances lock for this procedure
  SELECT @msg = APPLOCK_MODE('public','OOPurgeHistoryLock','Session');
  IF (@msg <> 'NoLock') BEGIN
    EXEC @lockRetVal = sp_releaseapplock 'OOPurgeHistoryLock','Session';
  END;

  -- Return value
  RETURN @total_num_of_flows_handled

END; -- End of OOPurgeHistory procedure
