CREATE OR REPLACE PACKAGE BODY hpe_oo_db_maint AS

/*
* Author: Harar Zafrir, zafrir@hpe.com
* Maintainer: Vlad Moldovan, moldovan@hpe.com
* Version: 4.2
* Last update: 2016-DEC-01
*
* This package provides tools for keeping HPE OO's database in good health
* Compatible with Oracle 11gR2 and 12c, and OO 10.22 to 10.51 (inclusive)
*
*/

FUNCTION CurrentEPOCHMS
RETURN NUMBER
IS

/*
* Author: Harar Zafrir, zafrir@hpe.com
* Maintainer: Vlad Moldovan, moldovan@hpe.com
* Version: 4.2
* Last update: 2016-DEC-01
*
* This function returns the current EPOCH UTC time in milliseconds.
* Compatible with Oracle 11gR2 and 12c
*
* Parameters: None.
* Return Values: Current EPOCH UTC time in milliseconds.
* -----------------------------------------------------------------------------
* Change List
* -----------------------------------------------------------------------------
* 2015-NOV-04 [3.7] HPE OO 10.60 adaptations
* 2016-MAR-16 [3.8] No changes in IndexMaintenance, version alignment
* 2016-MAY-16 [4.0] Assorted improvements, cross-DB-cross-release package version alignment
* 2016-AUG-09 [4.1] HPE OO 10.70 adaptations
* 2016-DEC-01 [4.2] DB Maintenance Package alignment - no changes
* -----------------------------------------------------------------------------
*/

CURREPOCHMS NUMBER;
BEGIN
  SELECT EXTRACT(DAY FROM(SYS_EXTRACT_UTC(SYSTIMESTAMP)-TO_TIMESTAMP('1970-01-01','YYYY-MM-DD')))*86400000
         + TO_NUMBER(TO_CHAR(SYS_EXTRACT_UTC(SYSTIMESTAMP),'SSSSSFF3')) INTO CURREPOCHMS FROM DUAL;
  RETURN CURREPOCHMS;
END CURRENTEPOCHMS;

PROCEDURE PurgeHistory ( pPurgeExecutionsOlderThan IN INTEGER,
    pPurgeExecutionsInBatchesOf IN INTEGER,
    pShouldPurgeExecutionSummary IN INTEGER,
    pVerbose IN INTEGER,
    pStopPurgingProcessAfter IN INTEGER,
    pDeepClean IN INTEGER,
    pDisableIndexes IN INTEGER,
    pReturnValue OUT INTEGER
) IS

/*
* Author: Harar Zafrir, zafrir@hpe.com
* Maintainer: Vlad Moldovan, moldovan@hpe.com
* Version: 4.2
* Last update: 2016-DEC-01
*
* This procedure purges data from OO database tables
* Compatible with Oracle 11gR2 and 12c, and OO 10.22 to 10.51 inclusive
*
* Parameters:
* -----------------------------------------------------------------------------
*
* pPurgeExecutionsOlderThan:
*       How many days of history to keep from "now". This value has no default and must be specified.
*
* pPurgeExecutionsInBatchesOf:
*       The maximum number of flows to handle at once. Default is 1000, minimum value is 500, maximum is 10000.
*
* pShouldPurgeExecutionSummary:
*       Should OO_EXECUTION_SUMMARY table be purged.
*
* pVerbose:
*       Verbosity level. 0 means "quiet" output, 1 provides medium level details (default)
*       and 2 provides extended details.
*
* pStopPurgingProcessAfter:
*       Maximum time for the procedure to run. Default is 4 hours.
*
* pDeepClean:
*       Perform deeper cleansing - search for orphan records.
*       This could potentially require additional time to complete.
*
* pDisableIndexes:
*       Disable OO_STEP_LOG_STARTED indexes for the duration of this procedure run
*       Using this option may result in slower step drill-down operations
*       When the procedure ends, indexes are rebuilt
*       This feature is disabled by default (0)
*
* Return Values:
* -----------------------------------------------------------------------------
* The actual number of processed flows.
*
* -----------------------------------------------------------------------------
* Change List
* -----------------------------------------------------------------------------
* 2014-JUN-19 [2.1] Adding version banner
*                   No default for pPurgeExecutionsOlderThan. Must be specified.
* 2014-AUG-13 [3.0] HP OO 10.20 modifications
*                   Adding execution bount inputs / outputs purge
* 2015-FEB-01 [3.1] No implications - MS-SQL fix
* 2015-APR-13 [3.2] Fix for million-bindings flow
* 2015-AUG-23 [3.5] No changes in this package, cross-db maintenance scripts version alignment
* 2015-OCT-15 [3.6] Adding ForceLobShrink procedure
* 2015-NOV-04 [3.7] No changes in this package, HPE OO 10.60 adaptations
* 2016-MAR-16 [3.8] Utilize "RETURNING" for better efficiency,
*                   force numeric EXECUTION_ID sort, adding "disable index" feature
* 2016-MAY-16 [4.0] Assorted improvements, cross-DB-cross-release package version alignment
* 2016-AUG-09 [4.1] HPE OO 10.70 adaptations
* 2016-DEC-01 [4.2] DB Maintenance Package alignment - no changes
* -----------------------------------------------------------------------------
*/

-- Resource busy exception definition
resrcBusyExcptn EXCEPTION;
PRAGMA EXCEPTION_INIT(resrcBusyExcptn, -54);

-- General properties ---------------------------
vPurgeHistoryVersion                         VARCHAR2(10)  := '4.2';
vPurgeHistoryCompatibility                   VARCHAR2(50)  := '10.22 to 10.51 inclusive';
vSupportedVersions                           VARCHAR2(255) := '''10.22'',''10.22.0001'',''10.22.0002'',''10.50'',''10.51''';
vPurgedFlowStatuses                          VARCHAR2(255) := '''COMPLETED'',''CANCELED'',''SYSTEM_FAILURE''';
vDetectedOOVersion                           VARCHAR2(30)  := '';
-- Thresholds, batch size -----------------------
vLotsOfStepsThreshold                        INTEGER := 100000; -- Flows with more than this number of steps are considered unusually big
vDeleteBatchSize                             INTEGER := 10000;  -- An optimal number of rows to be deleted together, database efficiency
vMinBatchSize                                INTEGER := 500;
vMaxBatchSize                                INTEGER := 10000;
-- Validations and general purpose --------------
vLockHandle                                  VARCHAR2(128);
vLockResponse                                INTEGER := 10;
-- Operations and flow control ------------------
vDeepClean                                   INTEGER := 0;
vDisableIndexes                              INTEGER := 0;
vAreIndexesDisabled                          INTEGER := 0;
vIsIndexExist                                INTEGER := 0;
vVerbose                                     INTEGER := 0;
vPurgeExecutionsInBatchesOf                  INTEGER := 0;
vFlowBatchSize                               INTEGER := 0;
-- Preformatted queries -------------------------
vQry                                         VARCHAR2(2000) := '';
vQry1                                        VARCHAR2(2000) := '';
vQry2                                        VARCHAR2(2000) := '';
vQry3                                        VARCHAR2(2000) := '';
vQry4                                        VARCHAR2(2000) := '';
vQry5                                        VARCHAR2(2000) := '';
-- Counters -------------------------------------
vGeneralCount                                INTEGER := 0;
vTotalOfFlowsHandled                         INTEGER := 0;
vESDeletedRows                               INTEGER := 0;
vSLSTRTDeletedRows                           INTEGER := 0;
vSLENDDeletedRows                            INTEGER := 0;
vSLBNDDeletedRows                            INTEGER := 0;
vDeletedRowsInCurrBatch                      INTEGER := 0;
vBODeletedRows                               INTEGER := 0;
vBIDeletedRows                               INTEGER := 0;
vTotalDelRows                                INTEGER := 0;
vCurBatchCollectionSize                      INTEGER := 0;
vSLStrtStpCount                              INTEGER := 0;
vCurStpCount                                 INTEGER := NULL;
vCurSetMinIdx                                INTEGER := NULL;
vCurSetMaxIdx                                INTEGER := NULL;
vMinExecIDToDelete                           NUMBER(38,0) := 0;
vMaxExecIdToDelete                           NUMBER(38,0) := 0;
vCurrEXCIDBatchMinId                         NUMBER(38,0) := 0;
vCurrEXCIDBatchMaxId                         NUMBER(38,0) := 0;
-- Time, duration -------------------------------
vSeconds                                     INTEGER := 0;
vPurgeStartTime                              TIMESTAMP;
vPurgeEndTime                                TIMESTAMP;
vLastStartTime                               TIMESTAMP;
vN_O_W                                       TIMESTAMP;
vTempTimestamp                               TIMESTAMP;

-- For iterating over OO_STEP_LOG_STARTED step counts
TYPE SL_STRTD_STPS_REC IS RECORD
(
    V_EXECTN_ID VARCHAR(64),
    V_STEP_COUNT INTEGER
);
TYPE SL_STRTD_STPS_TAB IS TABLE OF SL_STRTD_STPS_REC;
SLS_STEP_COLLECTION SL_STRTD_STPS_TAB;

-- A buffer for specific steps HASH IDs
TYPE STPS_HASH_TAB IS TABLE OF VARCHAR(42);
STPS_HASH_COLLECTION STPS_HASH_TAB;
STPS_HASH_TEMP_COLLECTION STPS_HASH_TAB;

-- A buffer for specific steps ENDED IDs
TYPE STPS_ENDED_TAB IS TABLE OF NUMBER(38,0);
STPS_ENDED_COLLECTION STPS_ENDED_TAB;
STPS_ENDED_TEMP_COLLECTION STPS_ENDED_TAB;

BEGIN

  dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' Beginning execution of PurgeHistory procedure version ' || vPurgeHistoryVersion);
  dbms_output.put_line('----------------------------------------------------------------------------------');
  dbms_output.put_line('This procedure is compatible with OO ' || vPurgeHistoryCompatibility);

  pReturnValue := 0;
  vDeepClean := pDeepClean;
  vDisableIndexes := pDisableIndexes;
  vVerbose := pVerbose;
  vPurgeExecutionsInBatchesOf := pPurgeExecutionsInBatchesOf;

  ---------------------------------------------------------------------------------------------------------
  -- Validate input parameters                                                                           --
  ---------------------------------------------------------------------------------------------------------
  IF pPurgeExecutionsInBatchesOf < vMinBatchSize THEN
    dbms_output.put_line('================================================================================');
    dbms_output.put_line('**** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR ***');
    dbms_output.put_line('**** Purging batch size is too small. (provided ' || TO_CHAR(pPurgeExecutionsInBatchesOf) || '). Minimum value is ' || TO_CHAR(vMinBatchSize) || '.');
    dbms_output.put_line('**** Defaulting to 1000');
    dbms_output.put_line('================================================================================');
    vPurgeExecutionsInBatchesOf := 1000;
  END IF;

  IF pPurgeExecutionsInBatchesOf > vMaxBatchSize THEN
    dbms_output.put_line('================================================================================');
    dbms_output.put_line('**** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR ***');
    dbms_output.put_line('**** Purging batch size is too big. (provided ' || TO_CHAR(pPurgeExecutionsInBatchesOf) || '). Maximum value is ' || TO_CHAR(vMaxBatchSize) || '.');
    dbms_output.put_line('**** Defaulting to 1000');
    dbms_output.put_line('================================================================================');
    vPurgeExecutionsInBatchesOf := 1000;
  END IF;

  IF pPurgeExecutionsOlderThan < 1 THEN
    dbms_output.put_line('================================================================================');
    dbms_output.put_line('**** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR ***');
    dbms_output.put_line('**** Cannot keep less than 1 day. (provided ' || TO_CHAR(pPurgeExecutionsOlderThan) || '). Set pPurgeExecutionsOlderThan parameter accordingly.');
    dbms_output.put_line('================================================================================');
    RETURN;
  END IF;

  IF pPurgeExecutionsOlderThan > 1000 THEN
    dbms_output.put_line('================================================================================');
    dbms_output.put_line('**** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR ***');
    dbms_output.put_line('**** Retention is more than 1000 days? (provided ' || TO_CHAR(pPurgeExecutionsOlderThan) || '). Please check the value provided for pPurgeExecutionsOlderThan.');
    dbms_output.put_line('================================================================================');
    RETURN;
  END IF;

  IF pStopPurgingProcessAfter > 1000 THEN
    dbms_output.put_line('================================================================================');
    dbms_output.put_line('**** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR ***');
    dbms_output.put_line('**** Unlimited run time? (provided pStopPurgingProcessAfter: ' || TO_CHAR(pStopPurgingProcessAfter) || '). Please check the value provided for pStopPurgingProcessAfter.');
    dbms_output.put_line('================================================================================');
    RETURN;
  END IF;

  IF pShouldPurgeExecutionSummary NOT IN (0,1) THEN
    dbms_output.put_line('================================================================================');
    dbms_output.put_line('**** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR **** ERROR ***');
    dbms_output.put_line('**** pShouldPurgeExecutionSummary must be either 0 or 1. (provided ' || TO_CHAR(pShouldPurgeExecutionSummary) || ').');
    dbms_output.put_line('================================================================================');
    RETURN;
  END IF;

  IF pDeepClean NOT IN (0,1) THEN
    dbms_output.put_line('**** ERROR: pDeepClean must be either 0 or 1. Deep clean is deactivated during this run.');
    vDeepClean := 0;
  END IF;

  IF pDisableIndexes NOT IN (0,1) THEN
    dbms_output.put_line('**** ERROR: pDisableIndexes must be either 0 or 1. Indexes will not be modified.');
    vDisableIndexes := 0;
  END IF;

  IF pVerbose NOT IN (0,1,2,3,4) THEN
    dbms_output.put_line('**** ERROR: pVerbose must be one of: 0,1 or 2. Defaulting to 1.');
    vVerbose := 1;
  END IF;

  ---------------------------------------------------------------------------------------------------------
  -- Verify that OO release is supported by this stored procedure                                        --
  ---------------------------------------------------------------------------------------------------------
  vQry := 'BEGIN SELECT VALUE, CASE WHEN VALUE IN (' || vSupportedVersions || ') THEN 1 ELSE 0 END "COMPATIBILITY" INTO :x, :y FROM OO_SYSINFO WHERE UPPER(KEY_)=''VERSION''; END;';
  EXECUTE IMMEDIATE vQry USING OUT vDetectedOOVersion, OUT vGeneralCount;
  IF vGeneralCount <> 1 THEN
    dbms_output.put_line('*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*');
    dbms_output.put_line('**** ERROR: Unsupported OO version: ' || vDetectedOOVersion || '. This procedure only supports: ');
    dbms_output.put_line('**** ERROR: ' || REPLACE(vSupportedVersions,'''''',''''));
    dbms_output.put_line('*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*');
    RETURN;
  END IF;

  vPurgeStartTime := CURRENT_TIMESTAMP;
  vPurgeEndTime := vPurgeStartTime + (pStopPurgingProcessAfter/24);

  -- Flow batch size is relative to the steps batch size, but at least 100
  vFlowBatchSize:= ROUND((vPurgeExecutionsInBatchesOf/10),0);
  IF vFlowBatchSize < 100 THEN
    vFlowBatchSize := 100;
  END IF;

  -- If OO_EXECUTION_SUMMARY table is empty - only need to check for orphan records
  SELECT COUNT(1) INTO vGeneralCount FROM OO_EXECUTION_SUMMARY;
  IF vGeneralCount = 0 THEN
    dbms_output.put_line('INFO: OO_EXECUTION_SUMMARY table is empty');
    IF vDeepClean = 1 THEN
      GOTO check_for_orphans;
    ELSE
      GOTO summary;
    END IF;
  END IF;

  -- Get a timestamp for the newest run start time we need to KEEP
  vLastStartTime := vPurgeStartTime - pPurgeExecutionsOlderThan;

  ---------------------------------------------------------------------------------------------------------
  -- Get the minimal and maximal EXECUTION_ID values to handle                                           --
  -- As EXECUTION_ID allocation is asynchroneous, this range might also include EXECUTION_IDs            --
  -- which have started after the threshold time and should not be handled                               --
  ---------------------------------------------------------------------------------------------------------
  vQry1 := 'BEGIN SELECT MIN(TO_NUMBER(EXECUTION_ID)), MAX(TO_NUMBER(EXECUTION_ID))
            INTO :x, :y FROM OO_EXECUTION_SUMMARY WHERE START_TIME < :1
            AND STATUS IN (' || vPurgedFlowStatuses || '); END;';

  vQry2 := 'BEGIN SELECT MIN(TO_NUMBER(ES.EXECUTION_ID)),MAX(TO_NUMBER(ES.EXECUTION_ID))
            INTO :x, :y FROM OO_EXECUTION_SUMMARY ES
			JOIN (SELECT DISTINCT(EXECUTION_ID) EXECUTION_ID FROM OO_STEP_LOG_STARTED) SLSTR
			ON ES.EXECUTION_ID = SLSTR.EXECUTION_ID WHERE ES.START_TIME < :1
			AND ES.STATUS IN (' || vPurgedFlowStatuses || '); END;';

  IF pShouldPurgeExecutionSummary = 1 THEN
    EXECUTE IMMEDIATE vQry1 USING OUT vMinExecIDToDelete, OUT vMaxExecIdToDelete, IN vLastStartTime;
  ELSE -- In case OO_EXECUTION_SUMMARY should not be purged, limit scope using OO_STEP_LOG_STARTED
      EXECUTE IMMEDIATE vQry2 USING OUT vMinExecIDToDelete, OUT vMaxExecIdToDelete, IN vLastStartTime;
  END IF;

  -- Beginning of operation ---------------------------------------------------------------------------------
  vQry := 'BEGIN dbms_output.put_line(TO_CHAR(CURRENT_DATE, ''dd/mm/yyyy hh24:mi:ss'') || '' This procedure will purge data for executions older than: '' || TO_CHAR(:1, ''dd/mm/yyyy hh24:mi:ss'')); END;';
  EXECUTE IMMEDIATE vQry USING vLastStartTime;
  vQry := 'BEGIN dbms_output.put_line(TO_CHAR(CURRENT_DATE, ''dd/mm/yyyy hh24:mi:ss'') || '' Time limit for this run is: '' || TO_CHAR(:1, ''dd/mm/yyyy hh24:mi:ss'')); dbms_output.new_line; END;';
  EXECUTE IMMEDIATE vQry USING vPurgeEndTime;
  vQry := 'BEGIN dbms_output.put_line(''----------------------------------------------------------------------------------''); END;';
  EXECUTE IMMEDIATE vQry;

  IF vMinExecIDToDelete IS NULL THEN
    dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' INFO: OO_EXECUTION_SUMMARY table contains no relevant records.');
    IF vDeepClean = 1 THEN
      GOTO check_for_orphans;
    ELSE
      GOTO summary;
    END IF;
  END IF;

  -- Prepare for the the first batch run --------------------------------------------------------------------
  vCurrEXCIDBatchMinId := vMinExecIDToDelete;
  vCurrEXCIDBatchMaxId := vMinExecIDToDelete;
  vTotalOfFlowsHandled := 0;

  -- Define queries -----------------------------------------------------------------------------------------
  vQry3 := 'SELECT SLS.EXECUTION_ID, SLS.STEP_COUNT
  FROM ( SELECT EXECUTION_ID, COUNT(EXECUTION_ID) STEP_COUNT FROM OO_STEP_LOG_STARTED
  WHERE TO_NUMBER(EXECUTION_ID) BETWEEN :1  AND :2 GROUP BY EXECUTION_ID) SLS
  JOIN OO_EXECUTION_SUMMARY ES ON SLS.EXECUTION_ID = ES.EXECUTION_ID
  WHERE ES.STATUS IN (' || vPurgedFlowStatuses || ')
  AND ES.START_TIME < :3 AND ROWNUM <= :4 ORDER BY SLS.EXECUTION_ID ASC';

  vQry4 := 'BEGIN DELETE OO_EXECUTION_SUMMARY WHERE EXECUTION_ID BETWEEN :1 AND :2
  AND STATUS IN (' || vPurgedFlowStatuses || ') AND START_TIME < :3; :x := SQL%ROWCOUNT; END;';

  vQry5 := 'BEGIN DELETE OO_EXECUTION_SUMMARY WHERE EXECUTION_ID NOT IN
  (SELECT DISTINCT(EXECUTION_ID) FROM OO_STEP_LOG_STARTED)
  AND STATUS IN (' || vPurgedFlowStatuses || ') AND START_TIME < :1 AND ROWNUM <= :2;
  :x := SQL%ROWCOUNT; END;';
 
  ---------------------------------------------------------------------------------------------------------
  --  Remove child-less OO_EXECUTION_SUMMARY records if applicable                                       --
  --  -------------------------------------------------------------                                      --
  --  If OO_EXECUTION_SUMMARY contains records that do not have corresponding OO_STEP_LOG_STARTED data   --
  --  repeatedly purge these records                                                                     --
  ---------------------------------------------------------------------------------------------------------
  IF pShouldPurgeExecutionSummary = 1 THEN
    vTempTimestamp := CURRENT_TIMESTAMP; -- Start measuring time for OO_EXECUTION_SUMMARY child-less records purge
    vDeletedRowsInCurrBatch := 0;
    
    vGeneralCount := vDeleteBatchSize; -- Qualify 'while' condition for the first round
    WHILE vGeneralCount >= vDeleteBatchSize
    LOOP
    
      BEGIN
		EXECUTE IMMEDIATE vQry5 USING IN vLastStartTime, IN vDeleteBatchSize, OUT vGeneralCount;
        vDeletedRowsInCurrBatch := vDeletedRowsInCurrBatch + vGeneralCount;
		vESDeletedRows := vESDeletedRows + vGeneralCount;
        vTotalOfFlowsHandled := vESDeletedRows;
      EXCEPTION
      WHEN resrcBusyExcptn THEN
        dbms_output.put_line('ERROR: OO_EXECUTION_SUMMARY table is busy, could not lock it for deletion');
      WHEN OTHERS THEN
        dbms_output.put_line(SUBSTR(SQLERRM, 1, 200));
      END;
    
      COMMIT;
    
      vN_O_W := CURRENT_TIMESTAMP;
      IF vN_O_W > vPurgeEndTime THEN
        dbms_output.put_line('------ Time limit reached ---------');
        GOTO summary;
      END IF;
    
    END LOOP; -- End of child-less OO_EXECUTION_SUMMARY records purge WHILE loop
    
    IF vVerbose >= 4 AND vDeletedRowsInCurrBatch > 0 THEN
      vSeconds := (TO_NUMBER(TO_CHAR(CURRENT_TIMESTAMP ,'J'))*86400 + TO_NUMBER(TO_CHAR(CURRENT_TIMESTAMP ,'sssss.FF9')))
                   - (TO_NUMBER(TO_CHAR(vTempTimestamp ,'J'))*86400 + TO_NUMBER(TO_CHAR(vTempTimestamp ,'sssss.FF9')));
      IF vSeconds > 0 THEN
        dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' DEBUG: Deleted ' || TO_CHAR(vDeletedRowsInCurrBatch) || ' child-less OO_EXECUTION_SUMMARY records in ' || TO_CHAR(vSeconds) || ' seconds (' || TO_CHAR(ROUND(vDeletedRowsInCurrBatch/vSeconds,0)) || ' row/sec)');
      ELSE
        dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' DEBUG: Deleted ' || TO_CHAR(vDeletedRowsInCurrBatch) || ' child-less OO_EXECUTION_SUMMARY records');
      END IF;
    END IF;
  END IF;

  --=======================================================================================================--
  -- Main loop                                                                                             --
  --=======================================================================================================--
  WHILE vCurrEXCIDBatchMinId <= vMaxExecIdToDelete
  LOOP

    vN_O_W := CURRENT_TIMESTAMP;
    IF vN_O_W > vPurgeEndTime THEN
      dbms_output.put_line('------ Time limit reached ---------');
      GOTO summary;
    END IF;

    ---------------------------------------------------------------------------------------------------------
    -- Get the next batch of flows and corresponding step count into the SLS_STEP_COLLECTION collection    --
    -- The number of elements is between 0 and vFlowBatchSize                                              --
    ---------------------------------------------------------------------------------------------------------
    vTempTimestamp := CURRENT_TIMESTAMP;

    EXECUTE IMMEDIATE vQry3 BULK COLLECT INTO SLS_STEP_COLLECTION
    USING IN vCurrEXCIDBatchMinId, IN vMaxExecIdToDelete, IN vLastStartTime, IN vFlowBatchSize;

    IF SLS_STEP_COLLECTION IS NULL THEN
      vCurBatchCollectionSize := 0;
    ELSE
      vCurBatchCollectionSize := SLS_STEP_COLLECTION.COUNT;
    END IF;

    ---------------------------------------------------------------------------------------------------------
    -- If there are no more flows to handle - exit main WHILE loop                                         --
    ---------------------------------------------------------------------------------------------------------
    IF vCurBatchCollectionSize IS NULL OR vCurBatchCollectionSize < 1 THEN
      vCurrEXCIDBatchMinId := vMaxExecIdToDelete + 1; -- Qualify exit condition just in case
      EXIT;
    END IF;

    IF vVerbose >= 4 THEN
      vSeconds := (TO_NUMBER(TO_CHAR(CURRENT_TIMESTAMP,'J'))*86400 + TO_NUMBER(TO_CHAR(CURRENT_TIMESTAMP,'sssss.FF9')))
                   - (TO_NUMBER(TO_CHAR(vTempTimestamp,'J'))*86400 + TO_NUMBER(TO_CHAR(vTempTimestamp,'sssss.FF9')));
      IF vSeconds > 0 THEN
        dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' DEBUG: Done populating SLS_STEP_COLLECTION. ' || TO_CHAR(vCurBatchCollectionSize) || ' records collected in ' || TO_CHAR(vSeconds) || ' seconds');
      END IF;
    END IF;

    ---------------------------------------------------------------------------------------------------------
    -- Set the current batch higher execution ID pointer                                                   --
    ---------------------------------------------------------------------------------------------------------
    IF SLS_STEP_COLLECTION IS NOT NULL AND SLS_STEP_COLLECTION.LAST IS NOT NULL THEN
      vCurrEXCIDBatchMaxId := TO_NUMBER(SLS_STEP_COLLECTION(SLS_STEP_COLLECTION.LAST).V_EXECTN_ID);
    END IF;

    ---------------------------------------------------------------------------------------------------------
    -- At this point SLS_STEP_COLLECTION contains a list of flows and their corresponding number of steps  --
    -- SLS_STEP_COLLECTION contains between 1 and vPurgeExecutionsInBatchesOf records                      --
    ---------------------------------------------------------------------------------------------------------
    IF vCurBatchCollectionSize > 0 THEN

      IF pShouldPurgeExecutionSummary = 0 THEN
        vTotalOfFlowsHandled := vTotalOfFlowsHandled + vCurBatchCollectionSize;
      END IF;

      -- Disable indexes if applicable ----------------------------------------------------------------------
      IF vDisableIndexes = 1 AND vAreIndexesDisabled = 0 THEN

        -- Disable OO_STEP_LOG_STARTED.STEP_LOG_ST_HASH index if applicable ---------------------------------
        vIsIndexExist := 0;
        SELECT COUNT(1) INTO vIsIndexExist FROM USER_INDEXES
        WHERE INDEX_NAME='STEP_LOG_ST_HASH' AND STATUS='VALID';
        IF vIsIndexExist = 1 THEN
          IF vVerbose >= 0 THEN
            dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' Disabling index STEP_LOG_ST_HASH in OO_STEP_LOG_STARTED table');
          END IF;
          BEGIN -- Wrapping in a block for exception handling
            EXECUTE IMMEDIATE 'ALTER INDEX STEP_LOG_ST_HASH UNUSABLE';
            EXCEPTION
            WHEN resrcBusyExcptn THEN
              dbms_output.put_line('ERROR: Failed to disable OO_STEP_LOG_STARTED.STEP_LOG_ST_HASH index');
            WHEN OTHERS THEN
              dbms_output.put_line(SUBSTR(SQLERRM, 1, 200));
          END;
          vAreIndexesDisabled := 1;
        END IF;

        -- Disable OO_STEP_LOG_STARTED.STEP_LOG_ST_SQ index if applicable -----------------------------------
        vIsIndexExist := 0;
        SELECT COUNT(1) INTO vIsIndexExist FROM USER_INDEXES
        WHERE INDEX_NAME='STEP_LOG_ST_SQ' AND STATUS='VALID';
        IF vIsIndexExist = 1 THEN
          IF vVerbose >= 0 THEN
            dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' Disabling index STEP_LOG_ST_SQ in OO_STEP_LOG_STARTED table');
          END IF;
          BEGIN -- Wrapping in a block for exception handling
            EXECUTE IMMEDIATE 'ALTER INDEX STEP_LOG_ST_SQ UNUSABLE';
            EXCEPTION
            WHEN resrcBusyExcptn THEN
              dbms_output.put_line('ERROR: Failed to disable OO_STEP_LOG_STARTED.STEP_LOG_ST_SQ index');
            WHEN OTHERS THEN
              dbms_output.put_line(SUBSTR(SQLERRM, 1, 200));
          END;
          vAreIndexesDisabled := 1;
        END IF;

      END IF;

      IF vVerbose >= 2 THEN
        dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' INFO: Handling EXECUTION_ID range: ' || TO_CHAR(vCurrEXCIDBatchMinId) || ' ' || TO_CHAR(vCurrEXCIDBatchMaxId));
      END IF;

      -------------------------------------------------------------------------------------------------------
      -- The following loop breaks down all flows in SLS_STEP_COLLECTION, into smaller groups,
      -- accumulating and handling vPurgeExecutionsInBatchesOf steps (or a bit more)
      -- to be handled together
      -------------------------------------------------------------------------------------------------------
      FOR i IN SLS_STEP_COLLECTION.FIRST .. SLS_STEP_COLLECTION.LAST
      LOOP

        -- Accumulate step counts
        vCurStpCount := SLS_STEP_COLLECTION(i).V_STEP_COUNT;
        vSLStrtStpCount := vSLStrtStpCount + vCurStpCount;

        -- Get the minimal EXECUTION_ID collection index for this group of executions
        IF vCurSetMinIdx IS NULL THEN
          vCurSetMinIdx := i;
        END IF;

        -- Get the current maximal EXECUTION_ID collection index for this group of executions
        vCurSetMaxIdx := i;

        -- vPurgeExecutionsInBatchesOf (or a bit more) steps are accumulated, or this is the last round for this batch
        IF vSLStrtStpCount >= vPurgeExecutionsInBatchesOf OR vCurBatchCollectionSize = vCurSetMaxIdx THEN

          /* Notify that this is a case in which flow has an unusually large number of steps and derivatives */
          IF vCurStpCount >= vLotsOfStepsThreshold AND vVerbose >= 2 THEN
            dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' DEBUG: Note - discovered a flow with an unusually high number of steps: ' || TO_CHAR(vCurStpCount));
          END IF;

          -- Initilize STPS_HASH_COLLECTION if necessary
          IF STPS_HASH_COLLECTION IS NULL THEN
            STPS_HASH_COLLECTION := STPS_HASH_TAB();
          END IF;

          -- Initilize STPS_HASH_TEMP_COLLECTION if necessary
          IF STPS_HASH_TEMP_COLLECTION IS NULL THEN
            STPS_HASH_TEMP_COLLECTION := STPS_HASH_TAB();
          END IF;

          -- Initilize STPS_ENDED_COLLECTION if necessary
          IF STPS_ENDED_COLLECTION IS NULL THEN
            STPS_ENDED_COLLECTION := STPS_ENDED_TAB();
          END IF;

          -- Initilize STPS_ENDED_TEMP_COLLECTION if necessary
          IF STPS_ENDED_TEMP_COLLECTION IS NULL THEN
            STPS_ENDED_TEMP_COLLECTION := STPS_ENDED_TAB();
          END IF;

          vTempTimestamp := CURRENT_TIMESTAMP; -- Start measuring time for OO_STEP_LOG_STARTED purge

          -------------------------------------------------------------------------------------------
          -- Deleting from OO_STEP_LOG_STARTED                                                     --
          -------------------------------------------------------------------------------------------
          vGeneralCount := vDeleteBatchSize; -- Qualify 'while' condition for the first round
          WHILE vGeneralCount >= vDeleteBatchSize
          LOOP

            vGeneralCount := 0;

            -- Repetitively delete the amount of vDeleteBatchSize rows from OO_STEP_LOG_STARTED
            BEGIN
              FORALL j IN vCurSetMinIdx .. vCurSetMaxIdx
              EXECUTE IMMEDIATE
              'DELETE OO_STEP_LOG_STARTED WHERE EXECUTION_ID = :1 AND ROWNUM <= :2 RETURNING STEP_HASH_ID INTO :3'
              USING SLS_STEP_COLLECTION(j).V_EXECTN_ID, vDeleteBatchSize RETURNING BULK COLLECT INTO STPS_HASH_TEMP_COLLECTION;
              vGeneralCount := SQL%ROWCOUNT;
              vSLSTRTDeletedRows := vSLSTRTDeletedRows + vGeneralCount;
            EXCEPTION
            WHEN resrcBusyExcptn THEN
              dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' ERROR: OO_STEP_LOG_STARTED table is busy, could not lock it for deletion');
            WHEN OTHERS THEN
              dbms_output.put_line(SUBSTR(SQLERRM, 1, 200));
            END;

            COMMIT;

            IF STPS_HASH_TEMP_COLLECTION IS NOT NULL THEN
              STPS_HASH_COLLECTION := STPS_HASH_TEMP_COLLECTION MULTISET UNION STPS_HASH_COLLECTION;
              STPS_HASH_TEMP_COLLECTION.DELETE;
            END IF;

            IF vVerbose >= 3 AND vGeneralCount > 0 THEN
              dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' DEBUG: Deleted ' || TO_CHAR(vGeneralCount) || ' records from OO_STEP_LOG_STARTED' );
            END IF;

            vN_O_W := CURRENT_TIMESTAMP;
            IF vN_O_W > vPurgeEndTime THEN
              dbms_output.put_line('------ Time limit reached ---------');
                 GOTO summary;
            END IF;

          END LOOP; -- End of deleting from OO_STEP_LOG_STARTED -------------------------------------

          IF vVerbose >= 4 AND STPS_HASH_COLLECTION.COUNT > 0 THEN
            vSeconds := (TO_NUMBER(TO_CHAR(CURRENT_TIMESTAMP,'J'))*86400 + TO_NUMBER(TO_CHAR(CURRENT_TIMESTAMP,'sssss.FF9')))
                         - (TO_NUMBER(TO_CHAR(vTempTimestamp,'J'))*86400 + TO_NUMBER(TO_CHAR(vTempTimestamp,'sssss.FF9')));
            IF vSeconds > 0 THEN
              dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' DEBUG: Deleted  ' || TO_CHAR(STPS_HASH_COLLECTION.COUNT) || ' OO_STEP_LOG_STARTED records in ' || TO_CHAR(vSeconds) || ' seconds (' || TO_CHAR(ROUND(STPS_HASH_COLLECTION.COUNT/vSeconds,0)) || ' row/sec)');
            END IF;
          END IF;

          IF STPS_HASH_TEMP_COLLECTION IS NOT NULL THEN
            STPS_HASH_TEMP_COLLECTION.DELETE;
          END IF;

          vTempTimestamp := CURRENT_TIMESTAMP; -- Start measuring time for OO_STEP_LOG_ENDED purge

          -------------------------------------------------------------------------------------------
          -- Deleting from OO_STEP_LOG_ENDED                                                     --
          -------------------------------------------------------------------------------------------
          vGeneralCount := vDeleteBatchSize; -- Qualify 'while' condition for the first round
          WHILE vGeneralCount >= vDeleteBatchSize
          LOOP

            vGeneralCount := 0;

            -- Initilize STPS_ENDED_COLLECTION if necessary
            IF STPS_ENDED_COLLECTION IS NULL THEN
              STPS_ENDED_COLLECTION := STPS_ENDED_TAB();
            END IF;

            -- Repetitively delete the amount of vDeleteBatchSize rows from OO_STEP_LOG_ENDED
            BEGIN
            FORALL j IN STPS_HASH_COLLECTION.FIRST .. STPS_HASH_COLLECTION.LAST
            EXECUTE IMMEDIATE
              'DELETE OO_STEP_LOG_ENDED WHERE STEP_ENDED_HASH_ID = :1 AND ROWNUM <= :2 RETURNING ID INTO :3'
              USING STPS_HASH_COLLECTION(j), vDeleteBatchSize RETURNING BULK COLLECT INTO STPS_ENDED_TEMP_COLLECTION;
              vGeneralCount := SQL%ROWCOUNT;
              vSLENDDeletedRows := vSLENDDeletedRows + vGeneralCount;
            EXCEPTION
              WHEN resrcBusyExcptn THEN
                dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' ERROR: OO_STEP_LOG_ENDED table is busy, could not lock it for deletion');
              WHEN OTHERS THEN
                dbms_output.put_line(SUBSTR(SQLERRM, 1, 200));
              END;

            COMMIT;

            IF STPS_ENDED_TEMP_COLLECTION IS NOT NULL THEN
              STPS_ENDED_COLLECTION := STPS_ENDED_TEMP_COLLECTION MULTISET UNION STPS_ENDED_COLLECTION;
              STPS_ENDED_TEMP_COLLECTION.DELETE;
            END IF;

            IF vVerbose >= 3 AND vGeneralCount > 0 THEN
              dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' DEBUG: Deleted ' || TO_CHAR(vGeneralCount) || ' records from OO_STEP_LOG_ENDED' );
            END IF;

            vN_O_W := CURRENT_TIMESTAMP;
            IF vN_O_W > vPurgeEndTime THEN
              dbms_output.put_line('------ Time limit reached ---------');
                 GOTO summary;
            END IF;

          END LOOP; -- End of deleting from OO_STEP_LOG_ENDED----------------------------------------

          IF vVerbose >= 4 AND STPS_ENDED_COLLECTION.COUNT > 0 THEN
            vSeconds := (TO_NUMBER(TO_CHAR(CURRENT_TIMESTAMP,'J'))*86400 + TO_NUMBER(TO_CHAR(CURRENT_TIMESTAMP,'sssss.FF9')))
                         - (TO_NUMBER(TO_CHAR(vTempTimestamp,'J'))*86400 + TO_NUMBER(TO_CHAR(vTempTimestamp,'sssss.FF9')));
            IF vSeconds > 0 THEN
              dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' DEBUG: Deleted  ' || TO_CHAR(STPS_ENDED_COLLECTION.COUNT) || ' OO_STEP_LOG_ENDED records in ' || TO_CHAR(vSeconds) || ' seconds (' || TO_CHAR(ROUND(STPS_ENDED_COLLECTION.COUNT/vSeconds,0)) || ' row/sec)');
            END IF;
          END IF;

          IF STPS_HASH_COLLECTION IS NOT NULL THEN
            STPS_HASH_COLLECTION.DELETE;
          END IF;

          IF STPS_ENDED_TEMP_COLLECTION IS NOT NULL THEN
            STPS_ENDED_TEMP_COLLECTION.DELETE;
          END IF;

          vTempTimestamp := CURRENT_TIMESTAMP; -- Start measuring time for OO_STEP_LOG_BINDINGS purge
          vDeletedRowsInCurrBatch := 0;

          -------------------------------------------------------------------------------------------
          -- Deleting from OO_STEP_LOG_BINDINGS                                                    --
          -------------------------------------------------------------------------------------------
          vGeneralCount := vDeleteBatchSize; -- Qualify 'while' condition for the first round
          WHILE vGeneralCount >= vDeleteBatchSize
          LOOP

            vGeneralCount := 0;

            -- Repetitively delete the amount of vDeleteBatchSize rows from OO_STEP_LOG_BINDINGS
            BEGIN
            FORALL j IN STPS_ENDED_COLLECTION.FIRST .. STPS_ENDED_COLLECTION.LAST
            EXECUTE IMMEDIATE
              'DELETE OO_STEP_LOG_BINDINGS WHERE STEP_LOG_ENDED_ID = :1 AND ROWNUM <= :2'
              USING STPS_ENDED_COLLECTION(j), vDeleteBatchSize;
              vGeneralCount := SQL%ROWCOUNT;
              vSLBNDDeletedRows := vSLBNDDeletedRows + vGeneralCount;
              vDeletedRowsInCurrBatch := vDeletedRowsInCurrBatch + vGeneralCount;
            EXCEPTION
            WHEN resrcBusyExcptn THEN
              dbms_output.put_line('ERROR: OO_STEP_LOG_BINDINGS table is busy, could not lock it for deletion');
            WHEN OTHERS THEN
              dbms_output.put_line(SUBSTR(SQLERRM, 1, 200));
            END;

            COMMIT;

            IF vVerbose >= 3 AND vGeneralCount > 0 THEN
              dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' DEBUG: Deleted ' || TO_CHAR(vGeneralCount) || ' records from OO_STEP_LOG_BINDINGS' );
            END IF;

            vN_O_W := CURRENT_TIMESTAMP;
            IF vN_O_W > vPurgeEndTime THEN
              dbms_output.put_line('------ Time limit reached ---------');
                 GOTO summary;
            END IF;

          END LOOP; -- End of deleting from OO_STEP_LOG_BINDINGS ------------------------------------

          IF STPS_ENDED_COLLECTION IS NOT NULL THEN
            STPS_ENDED_COLLECTION.DELETE;
          END IF;

          IF vVerbose >= 4 AND vDeletedRowsInCurrBatch > 0 THEN
            vSeconds := (TO_NUMBER(TO_CHAR(CURRENT_TIMESTAMP,'J'))*86400 + TO_NUMBER(TO_CHAR(CURRENT_TIMESTAMP,'sssss.FF9')))
                         - (TO_NUMBER(TO_CHAR(vTempTimestamp,'J'))*86400 + TO_NUMBER(TO_CHAR(vTempTimestamp,'sssss.FF9')));
            IF vSeconds > 0 THEN
              dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' DEBUG: Deleted ' || TO_CHAR(vDeletedRowsInCurrBatch) || ' OO_STEP_LOG_BINDINGS records in ' || TO_CHAR(vSeconds) || ' seconds (' || TO_CHAR(ROUND(vDeletedRowsInCurrBatch/vSeconds,0)) || ' row/sec)');
            END IF;
          END IF;

          -------------------------------------------------------------------------------------------
          -- Deleting from OO_EXECUTION_BOUND_OUTPUTS                                              --
          -------------------------------------------------------------------------------------------
          vGeneralCount := vDeleteBatchSize; -- Qualify 'while' condition for the first round
          WHILE vGeneralCount >= vDeleteBatchSize
          LOOP

            vGeneralCount := 0;

            -- Repetitively delete the amount of vDeleteBatchSize rows from OO_EXECUTION_BOUND_OUTPUTS
            BEGIN
              FORALL j IN vCurSetMinIdx .. vCurSetMaxIdx
              EXECUTE IMMEDIATE
                'DELETE OO_EXECUTION_BOUND_OUTPUTS WHERE EXECUTION_ID = :1 AND ROWNUM <= :2'
              USING SLS_STEP_COLLECTION(j).V_EXECTN_ID, vDeleteBatchSize;
              vGeneralCount := SQL%ROWCOUNT;
              vBODeletedRows := vBODeletedRows + vGeneralCount;
            EXCEPTION
            WHEN resrcBusyExcptn THEN
              dbms_output.put_line('ERROR: OO_EXECUTION_BOUND_OUTPUTS table is busy, could not lock it for deletion');
            WHEN OTHERS THEN
              dbms_output.put_line(SUBSTR(SQLERRM, 1, 200));
            END;

            COMMIT;

            IF vVerbose >= 3 AND vGeneralCount > 0 THEN
              dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' DEBUG: Deleted ' || TO_CHAR(vGeneralCount) || ' records from OO_EXECUTION_BOUND_OUTPUTS');
            END IF;

            vN_O_W := CURRENT_TIMESTAMP;
            IF vN_O_W > vPurgeEndTime THEN
              dbms_output.put_line('------ Time limit reached ---------');
              GOTO summary;
            END IF;

          END LOOP; -- End of deleting from OO_EXECUTION_BOUND_OUTPUTS ------------------------------

          -------------------------------------------------------------------------------------------
          -- Deleting from OO_EXECUTION_BOUND_INPUTS                                               --
          -------------------------------------------------------------------------------------------
          vGeneralCount := vDeleteBatchSize; -- Qualify 'while' condition for the first round
          WHILE vGeneralCount >= vDeleteBatchSize
          LOOP

            vGeneralCount := 0;

            BEGIN
              FORALL j IN vCurSetMinIdx .. vCurSetMaxIdx
              EXECUTE IMMEDIATE
                'DELETE OO_EXECUTION_BOUND_INPUTS WHERE EXECUTION_ID = :1 AND ROWNUM <= :2'
              USING SLS_STEP_COLLECTION(j).V_EXECTN_ID, vDeleteBatchSize;
              vGeneralCount := SQL%ROWCOUNT;
              vBIDeletedRows := vBIDeletedRows + vGeneralCount;
            EXCEPTION
            WHEN resrcBusyExcptn THEN
              dbms_output.put_line('ERROR: OO_EXECUTION_BOUND_INPUTS table is busy, could not lock it for deletion');
            WHEN OTHERS THEN
              dbms_output.put_line(SUBSTR(SQLERRM, 1, 200));
            END;

            COMMIT;

            IF vVerbose >= 3 AND vGeneralCount > 0 THEN
              dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' DEBUG: Deleted ' || TO_CHAR(vGeneralCount) || ' records from OO_EXECUTION_BOUND_INPUTS');
            END IF;

            vN_O_W := CURRENT_TIMESTAMP;
            IF vN_O_W > vPurgeEndTime THEN
              dbms_output.put_line('------ Time limit reached ---------');
              GOTO summary;
            END IF;

          END LOOP; -- End of deleting from OO_EXECUTION_BOUND_INPUTS -------------------------------

          -- Prepare for the next set
          vSLStrtStpCount := 0;
          vCurSetMinIdx := NULL;
          vCurSetMaxIdx := NULL;

        END IF; -- End of "IF" we have enough steps to process, or last round

      END LOOP; -- End of "WHILE" iterating over the current collection

    END IF; -- End of handling the current flow range

    -- Delete from OO_EXECUTION_SUMMARY if applicable
    IF pShouldPurgeExecutionSummary = 1 THEN

      IF vVerbose >= 4 THEN
        dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' DEBUG: About to delete from OO_EXECUTION_SUMMARY');
      END IF;

      vGeneralCount := 0;

      BEGIN -- Delete current OO_EXECUTION_SUMMARY batch
	    EXECUTE IMMEDIATE vQry4 USING IN vCurrEXCIDBatchMinId, IN vCurrEXCIDBatchMaxId, IN vLastStartTime, OUT vGeneralCount;
        vESDeletedRows := vESDeletedRows + vGeneralCount;
        vTotalOfFlowsHandled := vESDeletedRows;
      EXCEPTION
        WHEN resrcBusyExcptn THEN
          dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' ERROR: OO_EXECUTION_SUMMARY table is busy, could not lock it for deletion');
        WHEN OTHERS THEN
          dbms_output.put_line(SUBSTR(SQLERRM, 1, 200));
        END;

        COMMIT; -- OO_EXECUTION_SUMMARY rows deleted

        IF vVerbose >= 3 AND vGeneralCount > 0 THEN
          dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' DEBUG: Deleted ' || TO_CHAR(vGeneralCount) || ' records from OO_EXECUTION_SUMMARY');
        END IF;

    END IF; -- End of deleting from OO_EXECUTION_SUMMARY

    -- Prepare for the next OO_EXECUTION_SUMMARY batch

    -- Update return value
    pReturnValue := vTotalOfFlowsHandled;

    -- Advance counters
    vCurrEXCIDBatchMinId := vCurrEXCIDBatchMaxId + 1;
	vCurrEXCIDBatchMaxId := vCurrEXCIDBatchMinId;

    -- Cutoff at upper limit
    IF vCurrEXCIDBatchMinId >= vMaxExecIdToDelete THEN
      vCurrEXCIDBatchMinId := vMaxExecIdToDelete + 1; -- Qualify exit condition just in case...
      EXIT;
    END IF;

	-- Empty the collection, prepare for the next round
     IF SLS_STEP_COLLECTION IS NOT NULL THEN
       SLS_STEP_COLLECTION.DELETE;
     END IF;
     vCurBatchCollectionSize := 0;

  END LOOP; -- End of main WHILE loop

  <<check_for_orphans>>
  IF vDeepClean <> 1 THEN
    GOTO summary;
  END IF;

  ----------------------------------------------------------------------------------------
  -- Deep clean operations, removing orphan records if such are detected                --
  ----------------------------------------------------------------------------------------

  IF vVerbose >= 1 THEN
    dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' INFO: Beginning to search for (and remove) orphan records');
  END IF;

  -- Handling orphan OO_STEP_LOG_BINDINGS records if applicable
  vTempTimestamp := CURRENT_TIMESTAMP; -- Start measuring time for OO_STEP_LOG_BINDINGS orphan records purge
  vDeletedRowsInCurrBatch := 0;

  vGeneralCount := vDeleteBatchSize; -- Qualify 'while' condition for the first round
  WHILE vGeneralCount >= vDeleteBatchSize
  LOOP

    BEGIN
      EXECUTE IMMEDIATE
      'DELETE OO_STEP_LOG_BINDINGS WHERE STEP_LOG_ENDED_ID NOT IN (SELECT ID FROM OO_STEP_LOG_ENDED) AND ROWNUM <= :1'
      USING vDeleteBatchSize;
      vGeneralCount := SQL%ROWCOUNT;
      vDeletedRowsInCurrBatch := vDeletedRowsInCurrBatch + vGeneralCount;
      vSLBNDDeletedRows := vSLBNDDeletedRows + vGeneralCount;
    EXCEPTION
    WHEN resrcBusyExcptn THEN
      dbms_output.put_line('ERROR: OO_STEP_LOG_BINDINGS table is busy, could not lock it for deletion');
    WHEN OTHERS THEN
      dbms_output.put_line(SUBSTR(SQLERRM, 1, 200));
    END;

    COMMIT;

    vN_O_W := CURRENT_TIMESTAMP;
    IF vN_O_W > vPurgeEndTime THEN
      dbms_output.put_line('------ Time limit reached ---------');
      GOTO summary;
    END IF;

  END LOOP; -- End of OO_STEP_LOG_BINDINGS orphan records WHILE loop

  IF vVerbose >= 4 AND vDeletedRowsInCurrBatch > 0 THEN
    vSeconds := (TO_NUMBER(TO_CHAR(CURRENT_TIMESTAMP ,'J'))*86400 + TO_NUMBER(TO_CHAR(CURRENT_TIMESTAMP ,'sssss.FF9')))
                 - (TO_NUMBER(TO_CHAR(vTempTimestamp ,'J'))*86400 + TO_NUMBER(TO_CHAR(vTempTimestamp ,'sssss.FF9')));
    IF vSeconds > 0 THEN
      dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' DEBUG: Deleted ' || TO_CHAR(vDeletedRowsInCurrBatch) || ' orphan OO_STEP_LOG_BINDINGS records in ' || TO_CHAR(vSeconds) || ' seconds (' || TO_CHAR(ROUND(vDeletedRowsInCurrBatch/vSeconds,0)) || ' row/sec)');
    ELSE
      dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' DEBUG: Deleted ' || TO_CHAR(vDeletedRowsInCurrBatch) || ' orphan OO_STEP_LOG_BINDINGS records');
    END IF;
  END IF;

  -- Handling orphan OO_STEP_LOG_ENDED records if applicable
  vTempTimestamp := CURRENT_TIMESTAMP; -- Start measuring time for OO_STEP_LOG_ENDED orphan records purge
  vDeletedRowsInCurrBatch := 0;

  vGeneralCount := vDeleteBatchSize; -- Qualify 'while' condition for the first round
  WHILE vGeneralCount >= vDeleteBatchSize
  LOOP

    BEGIN
      EXECUTE IMMEDIATE
      'DELETE OO_STEP_LOG_ENDED WHERE STEP_ENDED_HASH_ID NOT IN (SELECT STEP_HASH_ID FROM OO_STEP_LOG_STARTED) AND ROWNUM <= :1'
      USING vDeleteBatchSize;
      vGeneralCount := SQL%ROWCOUNT;
      vDeletedRowsInCurrBatch := vDeletedRowsInCurrBatch + vGeneralCount;
      vSLENDDeletedRows := vSLENDDeletedRows + vGeneralCount;
    EXCEPTION
    WHEN resrcBusyExcptn THEN
      dbms_output.put_line('ERROR: OO_STEP_LOG_ENDED table is busy, could not lock it for deletion');
    WHEN OTHERS THEN
      dbms_output.put_line(SUBSTR(SQLERRM, 1, 200));
    END;

    COMMIT;

    vN_O_W := CURRENT_TIMESTAMP;
    IF vN_O_W > vPurgeEndTime THEN
      dbms_output.put_line('------ Time limit reached ---------');
      GOTO summary;
    END IF;

  END LOOP; -- End of OO_STEP_LOG_ENDED orphan records WHILE loop

  IF vVerbose >= 4 AND vDeletedRowsInCurrBatch > 0 THEN
    vSeconds := (TO_NUMBER(TO_CHAR(CURRENT_TIMESTAMP ,'J'))*86400 + TO_NUMBER(TO_CHAR(CURRENT_TIMESTAMP ,'sssss.FF9')))
                 - (TO_NUMBER(TO_CHAR(vTempTimestamp ,'J'))*86400 + TO_NUMBER(TO_CHAR(vTempTimestamp ,'sssss.FF9')));
    IF vSeconds > 0 THEN
      dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' DEBUG: Deleted ' || TO_CHAR(vDeletedRowsInCurrBatch) || ' orphan OO_STEP_LOG_ENDED records in ' || TO_CHAR(vSeconds) || ' seconds (' || TO_CHAR(ROUND(vDeletedRowsInCurrBatch/vSeconds,0)) || ' row/sec)');
    ELSE
      dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' DEBUG: Deleted ' || TO_CHAR(vDeletedRowsInCurrBatch) || ' orphan OO_STEP_LOG_ENDED records');
    END IF;
  END IF;

  -- Handling orphan OO_STEP_LOG_STARTED records if applicable
  vTempTimestamp := CURRENT_TIMESTAMP; -- Start measuring time for OO_STEP_LOG_STARTED orphan records purge
  vDeletedRowsInCurrBatch := 0;

  vGeneralCount := vDeleteBatchSize; -- Qualify 'while' condition for the first round
  WHILE vGeneralCount >= vDeleteBatchSize
  LOOP

    BEGIN
      EXECUTE IMMEDIATE
      'DELETE OO_STEP_LOG_STARTED WHERE EXECUTION_ID NOT IN (SELECT EXECUTION_ID FROM OO_EXECUTION_SUMMARY) AND ROWNUM <= :1'
      USING vDeleteBatchSize;
      vGeneralCount := SQL%ROWCOUNT;
      vDeletedRowsInCurrBatch := vDeletedRowsInCurrBatch + vGeneralCount;
      vSLSTRTDeletedRows := vSLSTRTDeletedRows + vGeneralCount;
    EXCEPTION
    WHEN resrcBusyExcptn THEN
      dbms_output.put_line('ERROR: OO_STEP_LOG_STARTED table is busy, could not lock it for deletion');
    WHEN OTHERS THEN
      dbms_output.put_line(SUBSTR(SQLERRM, 1, 200));
    END;

    COMMIT;

    vN_O_W := CURRENT_TIMESTAMP;
    IF vN_O_W > vPurgeEndTime THEN
      dbms_output.put_line('------ Time limit reached ---------');
      GOTO summary;
    END IF;

  END LOOP; -- End of OO_STEP_LOG_STARTED orphan records WHILE loop

  IF vVerbose >= 4 AND vDeletedRowsInCurrBatch > 0 THEN
    vSeconds := (TO_NUMBER(TO_CHAR(CURRENT_TIMESTAMP ,'J'))*86400 + TO_NUMBER(TO_CHAR(CURRENT_TIMESTAMP ,'sssss.FF9')))
                 - (TO_NUMBER(TO_CHAR(vTempTimestamp ,'J'))*86400 + TO_NUMBER(TO_CHAR(vTempTimestamp ,'sssss.FF9')));
    IF vSeconds > 0 THEN
      dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' DEBUG: Deleted ' || TO_CHAR(vDeletedRowsInCurrBatch) || ' orphan OO_STEP_LOG_STARTED records in ' || TO_CHAR(vSeconds) || ' seconds (' || TO_CHAR(ROUND(vDeletedRowsInCurrBatch/vSeconds,0)) || ' row/sec)');
    ELSE
      dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' DEBUG: Deleted ' || TO_CHAR(vDeletedRowsInCurrBatch) || ' orphan OO_STEP_LOG_STARTED records');
    END IF;
  END IF;

  -- Handling orphan OO_EXECUTION_BOUND_INPUTS records if applicable
  vTempTimestamp := CURRENT_TIMESTAMP; -- Start measuring time for OO_EXECUTION_BOUND_INPUTS orphan records purge
  vDeletedRowsInCurrBatch := 0;

  vGeneralCount := vDeleteBatchSize; -- Qualify 'while' condition for the first round
  WHILE vGeneralCount >= vDeleteBatchSize
  LOOP

    BEGIN
      EXECUTE IMMEDIATE
      'DELETE OO_EXECUTION_BOUND_INPUTS WHERE EXECUTION_ID NOT IN (SELECT EXECUTION_ID FROM OO_EXECUTION_SUMMARY) AND ROWNUM <= :1'
      USING vDeleteBatchSize;
      vGeneralCount := SQL%ROWCOUNT;
      vDeletedRowsInCurrBatch := vDeletedRowsInCurrBatch + vGeneralCount;
      vBIDeletedRows := vBIDeletedRows + vGeneralCount;
    EXCEPTION
    WHEN resrcBusyExcptn THEN
      dbms_output.put_line('ERROR: OO_EXECUTION_BOUND_INPUTS table is busy, could not lock it for deletion');
    WHEN OTHERS THEN
      dbms_output.put_line(SUBSTR(SQLERRM, 1, 200));
    END;

    COMMIT;

    vN_O_W := CURRENT_TIMESTAMP;
    IF vN_O_W > vPurgeEndTime THEN
      dbms_output.put_line('------ Time limit reached ---------');
      GOTO summary;
    END IF;

  END LOOP; -- End of orphan OO_EXECUTION_BOUND_INPUTS records WHILE loop

  IF vVerbose >= 4 AND vDeletedRowsInCurrBatch > 0 THEN
    vSeconds := (TO_NUMBER(TO_CHAR(CURRENT_TIMESTAMP ,'J'))*86400 + TO_NUMBER(TO_CHAR(CURRENT_TIMESTAMP ,'sssss.FF9')))
                 - (TO_NUMBER(TO_CHAR(vTempTimestamp ,'J'))*86400 + TO_NUMBER(TO_CHAR(vTempTimestamp ,'sssss.FF9')));
    IF vSeconds > 0 THEN
      dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' DEBUG: Deleted ' || TO_CHAR(vDeletedRowsInCurrBatch) || ' orphan OO_EXECUTION_BOUND_INPUTS records in ' || TO_CHAR(vSeconds) || ' seconds (' || TO_CHAR(ROUND(vDeletedRowsInCurrBatch/vSeconds,0)) || ' row/sec)');
    ELSE
      dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' DEBUG: Deleted ' || TO_CHAR(vDeletedRowsInCurrBatch) || ' orphan OO_EXECUTION_BOUND_INPUTS records');
    END IF;
  END IF;

  -- Handling orphan OO_EXECUTION_BOUND_OUTPUTS records if applicable
  vTempTimestamp := CURRENT_TIMESTAMP; -- Start measuring time for OO_EXECUTION_BOUND_OUTPUTS orphan records purge
  vDeletedRowsInCurrBatch := 0;

  vGeneralCount := vDeleteBatchSize; -- Qualify 'while' condition for the first round
  WHILE vGeneralCount >= vDeleteBatchSize
  LOOP

    BEGIN
      EXECUTE IMMEDIATE
      'DELETE OO_EXECUTION_BOUND_OUTPUTS WHERE EXECUTION_ID NOT IN (SELECT EXECUTION_ID FROM OO_EXECUTION_SUMMARY) AND ROWNUM <= :1'
      USING vGeneralCount;
      vGeneralCount := SQL%ROWCOUNT;
      vDeletedRowsInCurrBatch := vDeletedRowsInCurrBatch + vGeneralCount;
      vBODeletedRows := vBODeletedRows + vGeneralCount;
    EXCEPTION
    WHEN resrcBusyExcptn THEN
      dbms_output.put_line('ERROR: OO_EXECUTION_BOUND_OUTPUTS table is busy, could not lock it for deletion');
    WHEN OTHERS THEN
      dbms_output.put_line(SUBSTR(SQLERRM, 1, 200));
    END;

    COMMIT;

    vN_O_W := CURRENT_TIMESTAMP;
    IF vN_O_W > vPurgeEndTime THEN
      dbms_output.put_line('------ Time limit reached ---------');
      GOTO summary;
    END IF;

  END LOOP; -- End of orphan bound inputs records WHILE loop

  IF vVerbose >= 4 AND vDeletedRowsInCurrBatch > 0 THEN
    vSeconds := (TO_NUMBER(TO_CHAR(CURRENT_TIMESTAMP ,'J'))*86400 + TO_NUMBER(TO_CHAR(CURRENT_TIMESTAMP ,'sssss.FF9')))
                 - (TO_NUMBER(TO_CHAR(vTempTimestamp ,'J'))*86400 + TO_NUMBER(TO_CHAR(vTempTimestamp ,'sssss.FF9')));
    IF vSeconds > 0 THEN
      dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' DEBUG: Deleted ' || TO_CHAR(vDeletedRowsInCurrBatch) || ' orphan OO_EXECUTION_BOUND_OUTPUTS records in ' || TO_CHAR(vSeconds) || ' seconds (' || TO_CHAR(ROUND(vDeletedRowsInCurrBatch/vSeconds,0)) || ' row/sec)');
    ELSE
      dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' DEBUG: Deleted ' || TO_CHAR(vDeletedRowsInCurrBatch) || ' orphan OO_EXECUTION_BOUND_OUTPUTS records');
    END IF;
  END IF;

  -- Summary
  <<summary>>

  vTempTimestamp := CURRENT_TIMESTAMP; -- Start measuring time for optional index rebuild

  -- Rebuilding OO_STEP_LOG_STARTED.STEP_LOG_ST_HASH index if applicable ------------------------------------
  -- Relevant only for 10.50 or greater
  vIsIndexExist := 0;
  SELECT COUNT(1) INTO vIsIndexExist FROM USER_INDEXES
  WHERE INDEX_NAME='STEP_LOG_ST_HASH' AND STATUS='UNUSABLE';
  IF vIsIndexExist = 1 THEN
    IF vVerbose >= 0 THEN
      dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' Rebuilding index STEP_LOG_ST_HASH in OO_STEP_LOG_STARTED table');
    END IF;
    BEGIN -- Wrapping in a block for exception handling
      EXECUTE IMMEDIATE 'ALTER INDEX STEP_LOG_ST_HASH REBUILD';
      EXCEPTION
      WHEN resrcBusyExcptn THEN
        dbms_output.put_line('ERROR: Failed to disable OO_STEP_LOG_STARTED.STEP_LOG_ST_HASH index');
      WHEN OTHERS THEN
        dbms_output.put_line(SUBSTR(SQLERRM, 1, 200));
    END;
  END IF;

  -- Rebuilding OO_STEP_LOG_STARTED.STEP_LOG_ST_SQ index if applicable --------------------------------------
  vIsIndexExist := 0;
  SELECT COUNT(1) INTO vIsIndexExist FROM USER_INDEXES
  WHERE INDEX_NAME='STEP_LOG_ST_SQ' AND STATUS='UNUSABLE';
  IF vIsIndexExist = 1 THEN
    IF vVerbose >= 0 THEN
      dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' Rebuilding index STEP_LOG_ST_SQ in OO_STEP_LOG_STARTED table');
    END IF;
    BEGIN -- Wrapping in a block for exception handling
      EXECUTE IMMEDIATE 'ALTER INDEX STEP_LOG_ST_SQ REBUILD';
      EXCEPTION
      WHEN resrcBusyExcptn THEN
        dbms_output.put_line('ERROR: Failed to disable OO_STEP_LOG_STARTED.STEP_LOG_ST_SQ index');
      WHEN OTHERS THEN
        dbms_output.put_line(SUBSTR(SQLERRM, 1, 200));
    END;
  END IF;

  IF vVerbose >= 4 THEN
    vSeconds := (TO_NUMBER(TO_CHAR(CURRENT_TIMESTAMP ,'J'))*86400 + TO_NUMBER(TO_CHAR(CURRENT_TIMESTAMP ,'sssss.FF9')))
                 - (TO_NUMBER(TO_CHAR(vTempTimestamp ,'J'))*86400 + TO_NUMBER(TO_CHAR(vTempTimestamp ,'sssss.FF9')));
    IF vSeconds > 0 THEN
      dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' DEBUG: Indexes rebuilt in ' || TO_CHAR(vSeconds) || ' seconds');
    END IF;
  END IF;

  dbms_output.new_line;

  IF vVerbose > 0 THEN
    dbms_output.put_line('----------------------------------------------------------------------------------');
    dbms_output.put_line('INFO: Total operation time : ' || LPAD(EXTRACT(HOUR FROM (CURRENT_TIMESTAMP - vPurgeStartTime)),2,'0') || ':' || LPAD(EXTRACT(MINUTE FROM (CURRENT_TIMESTAMP - vPurgeStartTime)),2,'0') || ':' || LPAD(TO_CHAR(CEIL(TO_NUMBER(EXTRACT(SECOND FROM (CURRENT_TIMESTAMP - vPurgeStartTime))))),2,'0'));
  END IF;

  vTotalDelRows := vESDeletedRows + vSLSTRTDeletedRows + vSLENDDeletedRows + vSLBNDDeletedRows + vBODeletedRows + vBIDeletedRows;
  IF vVerbose > 1 AND vTotalDelRows > 0 THEN
    dbms_output.put_line('SUMMARY: Deleted ' || TO_CHAR(vESDeletedRows) || ' rows from OO_EXECUTION_SUMMARY');
    dbms_output.put_line('SUMMARY: Deleted ' || TO_CHAR(vSLSTRTDeletedRows) || ' rows from OO_STEP_LOG_STARTED');
    dbms_output.put_line('SUMMARY: Deleted ' || TO_CHAR(vSLENDDeletedRows) || ' rows from OO_STEP_LOG_ENDED');
    dbms_output.put_line('SUMMARY: Deleted ' || TO_CHAR(vSLBNDDeletedRows) || ' rows from OO_STEP_LOG_BINDINGS');
    dbms_output.put_line('SUMMARY: Deleted ' || TO_CHAR(vBIDeletedRows) || ' rows from OO_EXECUTION_BOUND_INPUTS');
    dbms_output.put_line('SUMMARY: Deleted ' || TO_CHAR(vBODeletedRows) || ' rows from OO_EXECUTION_BOUND_OUTPUTS');
  END IF;

  IF vVerbose >= 1 THEN
    IF vTotalDelRows > 0 THEN
      dbms_output.put_line('SUMMARY:  Total: ' || TO_CHAR(vTotalDelRows) || ' rows deleted from database');
    ELSE
      dbms_output.put_line('--------------------------------------');
      dbms_output.put_line('SUMMARY: No rows deleted from database');
      dbms_output.put_line('--------------------------------------');
    END IF;
  END IF;

  EXCEPTION
    WHEN OTHERS THEN
      dbms_output.put_line(SUBSTR(SQLERRM, 1, 200));

  RETURN; -- END PROCEDURE
END; -- Procedure PurgeHistory

PROCEDURE IndexMaintenance
(   pMaxHeight IN INTEGER,
    pMaxLeafsDeleted IN INTEGER,
    pRebuild IN INTEGER,
    pReturnValue OUT INTEGER
) IS

/*
* Author: Harar Zafrir, zafrir@hpe.com
* Maintainer: Vlad Moldovan, moldovan@hpe.com
* Version: 4.2
* Last update: 2016-DEC-01
*
* This procedure performs index maintenance for OO database.
*
* Parameters:
* -----------------------------------------------------------------------------
* pMaxHeight:           Minimal index height threshold for index rebuild
* pMaxLeafsDeleted:     Minimal deleted leaves threshold for index rebuild
* pRebuild:             Should indexes be rebuilt (1) or only perform a dry-run (0)
*
* Return Value:
* -----------------------------------------------------------------------------
* pReturnValue:         Number of rebuilt indexes
*
* -----------------------------------------------------------------------------
* Change list
* -----------------------------------------------------------------------------
* 2014-AUG-13 [3.0] No changes
* 2015-FEB-01 [3.1] No implications - MS-SQL fix
* 2015-APR-13 [3.2] No changes - fix for million-bindings flow
* 2015-AUG-23 [3.5] No changes in IndexMaintenance, cross-db maintenance scripts version alignment
* 2015-OCT-15 [3.6] Adding ForceLobShrink procedure
* 2015-NOV-04 [3.7] HPE OO 10.60 adaptations
* 2016-MAR-16 [3.8] No changes in IndexMaintenance, maintenance package version alignment
* 2016-MAY-16 [4.0] No changes in IndexMaintenance, maintenance package version alignment
* 2016-AUG-09 [4.1] HPE OO 10.70 adaptations
* 2016-DEC-01 [4.2] DB Maintenance Package alignment - no changes
* -----------------------------------------------------------------------------
*/

-- Resource busy exception definition
resrcBusyExcptn EXCEPTION;
PRAGMA EXCEPTION_INIT(resrcBusyExcptn, -54);

vIndexMaintenanceVersion VARCHAR2(10) := '4.2';

-- A Cursor for iterating through the index_stats data dictionary view
CURSOR csrIndexStats IS
    SELECT
        name
        ,height
        ,lf_rows AS leafRows
        ,del_lf_rows AS leafRowsDeleted
    FROM index_stats;

vIndexStats csrIndexStats%rowtype;

-- A Cursor for iterating through the user_indexes data dictionary view
CURSOR csrOOIndexes IS
    SELECT
        index_name
        ,tablespace_name
    FROM user_indexes
    WHERE partitioned = 'NO'
    AND index_type NOT IN ('LOB')
    AND status = 'VALID';

vCount      INTEGER := 0;   -- Number of rebuilt indexes
vErrors     INTEGER := 0;   -- Number of operations ending with an unhandled exception
vIsEnt      INTEGER := 0;   -- Is this an Enterprise edition (allowing online index rebuild)
vRebuildStr VARCHAR2(25) := ' REBUILD'; -- Rebuild option - regular or online

BEGIN

  dbms_output.put_line('Beginning execution of IndexMaintenance procedure version ' || vIndexMaintenanceVersion || ' at ' || TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss'));

  pReturnValue := 0;

  -- Check if this Oracle is an enterprise version or not
  SELECT COUNT(*) INTO vIsEnt FROM product_component_version WHERE product LIKE '%Oracle%Database%Enterprise%';
  IF vIsEnt = 1 THEN
    vRebuildStr := ' REBUILD ONLINE';
    dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' [INFO] An Enterprise Oracle version was detected.');
  END IF;

  -- Looping through all OO's indexes
  FOR vIndexRec IN csrOOIndexes
  LOOP -- Main loop

    -- Analyzing the current index. Results are saved in index_stats
    BEGIN -- Wrapping in a block for exception handling
        EXECUTE IMMEDIATE 'ANALYZE INDEX ' || vIndexRec.index_name || ' VALIDATE STRUCTURE';
        EXCEPTION
        WHEN resrcBusyExcptn THEN
            dbms_output.put_line('ERROR: Index ' || vIndexRec.index_name || ' is busy, could not lock it');
        WHEN OTHERS THEN
            dbms_output.put_line(SUBSTR(SQLERRM, 1, 200));
    END;

    OPEN csrIndexStats;
    FETCH csrIndexStats INTO vIndexStats;
    IF csrIndexStats%found THEN
      dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' [INFO] Checking index ' || vIndexStats.name);
        IF (vIndexStats.height > pMaxHeight)
            OR
            (vIndexStats.leafRows > 0 AND
             vIndexStats.leafRowsDeleted > 0 AND
            (vIndexStats.leafRowsDeleted * 100 / vIndexStats.leafRows) > pMaxLeafsDeleted)
        THEN
            vCount := vCount + 1;
            pReturnValue := pReturnValue + 1;

            IF pRebuild = 1 THEN
                BEGIN -- Wrapping in a block for exception handling
            dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || 'Rebuilding index ' || vIndexRec.index_name || '...');
                    EXECUTE IMMEDIATE 'ALTER INDEX ' || vIndexRec.index_name || vRebuildStr;
                    EXCEPTION
                    WHEN resrcBusyExcptn THEN
                        dbms_output.put_line('ERROR: Index ' || vIndexRec.index_name || ' is busy, could not lock it');
                    WHEN OTHERS THEN
                        vErrors := vErrors + 1;
                        dbms_output.put_line(SUBSTR(SQLERRM, 1, 200));
                END;
            ELSE
          dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || 'Dry run, index ' || vIndexRec.index_name || ' should be rebuilt.');
        END IF; -- End of dry-run condition
      END IF; -- End of rebuild criteria condition
    END IF; -- End of index list

    CLOSE csrIndexStats;

  END LOOP; -- End of main loop

  -- Summary
  IF vCount > 0 THEN
    IF pRebuild = 1 THEN
      dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' Total number of indexes rebuilt: ' || TO_CHAR(vCount));
    ELSE
      dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' Total number of indexes recommended for rebuild (not actually rebuilt): ' || TO_CHAR(vCount));
    END IF;
  ELSE
    dbms_output.put_line('No indexes were processed.');
  END IF;

END; -- IndexMaintenance

PROCEDURE ForceLobShrink
( pReturnValue OUT INTEGER ) IS

/*
* Author: Harar Zafrir, zafrir@hpe.com
* Maintainer: Vlad Moldovan, moldovan@hpe.com
* Version: 4.2
* Last update: 2016-DEC-01
*
* This procedure forcefully shrinks LOBs in OO's database.
* Remember that LOB shrinking requires a considerable amount of time
* and database resources. It is highly recommended to execute this
* procedure when OO and the database server are least active.
*
* Also make sure that your database settings allow shrinking:
* 1. TABLESPACE is not compressed
* 2. TABLESPACE uses AUTO segment space management
* 3. Tables are not compresses
* 4. Row movement is enabled for each table
*
* Parameters:
* -----------------------------------------------------------------------------
* none
*
* Return Value:
* -----------------------------------------------------------------------------
* pReturnValue: 0 for success
*               1 for one or more shrink failures
*
* -----------------------------------------------------------------------------
* Change list
* -----------------------------------------------------------------------------
* 2015-OCT-15 [3.6] First version
* 2015-NOV-04 [3.7] No changes in ForceLobShrink, maintenance package version alignment
* 2016-MAR-16 [3.8] No changes in ForceLobShrink, maintenance package version alignment
* 2016-MAY-16 [4.0] No changes in ForceLobShrink, maintenance package version alignment
* 2016-AUG-09 [4.1] HPE OO 10.70 adaptations
* -----------------------------------------------------------------------------
*/

-- Resource busy exception definition
resrcBusyExcptn EXCEPTION;
PRAGMA EXCEPTION_INIT(resrcBusyExcptn, -54);

-- Resource busy exception definition
invldsgmntExcptn EXCEPTION;
PRAGMA EXCEPTION_INIT(invldsgmntExcptn, -10635);

vForceLobShrinkVersion VARCHAR2(10) := '4.2';

-- A Cursor for iterating through the user_indexes data dictionary view
CURSOR crsOOLOBs IS
    SELECT L.TABLE_NAME,
      L.COLUMN_NAME,
      S.SEGMENT_NAME,
      S.BYTES
    FROM USER_SEGMENTS S
    LEFT OUTER JOIN USER_LOBS L
    ON S.SEGMENT_NAME = L.SEGMENT_NAME
    JOIN USER_TABLES T
    ON L.TABLE_NAME = T.TABLE_NAME
    AND S.SEGMENT_TYPE = 'LOBSEGMENT'
    ORDER BY S.BYTES DESC;

BEGIN

  dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' Beginning execution of ForceLobShrink procedure version ' || vForceLobShrinkVersion);
  dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') || ' LOB shrink operations may take some time. Please wait until the results for all operations are displayed together');

  pReturnValue := 0;

  -- Looping through all OO's LOBs
  FOR vLOBrec IN crsOOLOBs
  LOOP -- Main loop

    BEGIN -- Wrapping in a block for exception handling
      dbms_output.put_line(TO_CHAR(CURRENT_DATE, 'dd/mm/yyyy hh24:mi:ss') ||  ' [INFO] Shrinking LOB ' || vLOBrec.SEGMENT_NAME || ' in table ' || vLOBrec.TABLE_NAME || ' and column ' || vLOBrec.COLUMN_NAME);
      EXECUTE IMMEDIATE 'ALTER TABLE ' || vLOBrec.TABLE_NAME || ' MODIFY LOB ( ' || vLOBrec.COLUMN_NAME || ' ) (shrink space)';
      EXCEPTION
      WHEN resrcBusyExcptn THEN
        dbms_output.put_line('ERROR: LOB ' || vLOBrec.SEGMENT_NAME || ' is busy, could not shrink it');
        pReturnValue := 1;
      WHEN invldsgmntExcptn THEN
        dbms_output.put_line('[ERROR]: "ORA-10635: Invalid segment or tablespace type" has been detected');
        dbms_output.put_line('.  In order for "MODIFY LOB shrink space" to work, the following conditions must be met:');
        dbms_output.put_line('.    1. TABLESPACE must not be compressed');
        dbms_output.put_line('.    2. TABLESPACE must use AUTO segment space management');
        dbms_output.put_line('.    3. Tables must not be compresses');
        dbms_output.put_line('.    4. Row movement must be enabled for each table');
        dbms_output.put_line('.    5. LOBs storage must not be SECUREFILE');
        dbms_output.put_line('Exiting');
        pReturnValue := 1;
        RETURN;
      WHEN OTHERS THEN
        dbms_output.put_line(SUBSTR(SQLERRM, 1, 200));
        pReturnValue := 1;
    END;
  END LOOP; -- End of main loop

END; -- End of ForceLobShrink

END hpe_oo_db_maint;
/
