﻿CREATE OR REPLACE FUNCTION OOPurgeHistory(pPurgeExecutionsOlderThan    INT DEFAULT -1,
                                         pPurgeExecutionsInBatchesOf   INT DEFAULT 10000,
                                         pShouldPurgeExecutionSummary  INT DEFAULT 0,
                                         pVerbose                      INT DEFAULT 1,
                                         pStopPurgingProcessAfter      INT DEFAULT 4,
                                         pDeepClean                    INT DEFAULT 0,
                                         pSleepInterval            NUMERIC DEFAULT 0)
RETURNS INT AS $$
  /*
  * Author: Harar Zafrir, zafrir@hpe.com
  * Maintainer: Vlad Moldovan, moldovan@hpe.com
  * Version: 4.2
  * Last update: 2016-DEC-01
  *
  * This procedure purges flow and step data
  * from HPE OO's 10.6x to 10.70 database tables
  * Compatible with Posgres 9.2, 9.3, 9.4 and 9.5.
  *
  * Parameters:
  * -----------------------------------------------------------------------------
  *
  * pPurgeExecutionsOlderThan:
  *       How many days of history to keep. No default is provided, this parameter must be specified.
  *
  * pPurgeExecutionsInBatchesOf:
  *       The maximum number of flows to handled at once. Default is 10000, minimum value is 500, maximum is 100000.
  *
  * pShouldPurgeExecutionSummary:
  *       Should OO_EXECUTION_SUMMARY table be purged.
  *
  * pVerbose:
  *       Verbosity level. 0 means "quiet" output. 1 provides medium level details
  *       and 2 provides extended details.
  *
  * pStopPurgingProcessAfter:
  *       Maximum time for the procedure to run. Default is 4 hours.
  *
  * pDeepClean:
  *       Perform deeper cleansing - search for orphan records.
  *       Using this option usually extends procedure run time.
  *
  * pSleepInterval:
  *       Sleep this amount of seconds to sleep between batch iterations
  *       Can be set as small as 0.01
  *
  * Additional information:
  * -----------------------------------------------------------------------------
  * Change List
  * -----------------------------------------------------------------------------
  * 2014-JUN-19 [2.1] No default for pPurgeExecutionsOlderThan. Must be specified.
  * 2014-AUG-13 [3.0] HP OO 10.20 modifications
  *                   Adding execution bount inputs / outputs purge
  * 2015-FEB-01 [3.1] No implications - MS-SQL fix
  * 2015-APR-13 [3.2] Fix for million-bindings flow
  * 2015-JUL-01 [3.3] Performance improvements
  * 2015-AUG-23 [3.5] No changes in this script, cross-db maintenance scripts version alignment
  * 2015-NOV-04 [3.7] HPE OO 10.60 adaptations
  * 2016-AUG-04 [4.2] HPE OO 10.70 adaptations and alignment of package version
  * -----------------------------------------------------------------------------
  */

  DECLARE
  -- General properties ---------------------------
  vPurgeHistoryVersion        CHARACTER VARYING(10) DEFAULT '';
  vPurgeHistoryCompatibility  CHARACTER VARYING(15) DEFAULT '';
  vSupportedVersions          CHARACTER VARYING(255) DEFAULT '';
  vPurgedFlowStatuses         CHARACTER VARYING(255) DEFAULT '';
  vDetectedOOVersion          CHARACTER VARYING(30) DEFAULT '';
  vPgVersion                  CHARACTER VARYING(10) DEFAULT '';

  -- Thresholds, batch size -----------------------
  vLotsOfStepsThreshold       BIGINT DEFAULT 100000;  -- Flows with more than this number of steps are considered unusually big
  vDeleteBatchSize            BIGINT DEFAULT 10000;  -- An optimal number of rows to be deleted together, database efficiency
  vMinBatchSize               INT DEFAULT 100;   -- Batch size lower limit
  vMaxBatchSize               INT DEFAULT 10000; -- Batch size upper limit
  vFlowBatchSize              INT DEFAULT 0;
  vBigFlowCurrSteps           INT DEFAULT 0;
  -- Validations and general purpose --------------
  vGeneralCounter             INT DEFAULT 0;
  vPreviousCount              INT DEFAULT 0;
  vOutputRowLength            INT DEFAULT 80;
  -- Operations and flow control ------------------
  vAdLckID                    INT DEFAULT 1023;
  vMaxExecIdToDelete          BIGINT DEFAULT 0;
  vMinExecIdToDelete          BIGINT DEFAULT 0;
  vCurrEXCIDBatchMinId        BIGINT DEFAULT 0;
  vCurrEXCIDBatchMaxId        BIGINT DEFAULT 0;
  vEXCIDBatchStart            CHARACTER VARYING(64) DEFAULT '';
  vOnlyHandleOrphans          BOOLEAN DEFAULT FALSE;
  vSkipToSummary              BOOLEAN DEFAULT FALSE;
  vCurSetMinIdx               INT DEFAULT 0;
  vCurSetMaxIdx               INT DEFAULT 0;
  vCurSetSize                 INT DEFAULT 0;
  vCurSetAccumStepCount       INT DEFAULT 0;
  vDeletedRowsInCurrentBatch  INT DEFAULT 0;
  vAreConstraintsDisabled     BOOLEAN DEFAULT 0;
  -- Preformatted queries -------------------------
  query                       text DEFAULT '';
  query1                      text DEFAULT '';
  query2                      text DEFAULT '';
  query3                      text DEFAULT '';
  query4                      text DEFAULT '';
  query5                      text DEFAULT '';
  -- Counters -------------------------------------
  vCurStpCount                INT DEFAULT 0;
  vLastRowCount               INT DEFAULT 0;
  vSLSTRTDeletedRows          INT DEFAULT 0;
  vSLENDDeletedRows           INT DEFAULT 0;
  vSLBNDDeletedRows           INT DEFAULT 0;
  vExSumDeletedRows           INT DEFAULT 0;
  vBODeletedRows              INT DEFAULT 0;
  vBIDeletedRows              INT DEFAULT 0;
  vTotalDelRows               INT DEFAULT 0;
  vTotalFlowsHandled          INT DEFAULT 0;
  vSLStrtStpCount             INT DEFAULT 0;
  vCurBatchFlowCount          INT DEFAULT 0;
  -- Time, duration -------------------------------
  vPurgeStartTime             TIMESTAMP;
  vPurgeEndTime               TIMESTAMP;
  vEXCIDBatchStartTime        TIMESTAMP;
  vTempTimestamp              TIMESTAMP;
  vLastStartTime              BIGINT;
  -- Internal structures --------------------------
  vSlStpCntRec                RECORD;
  vSD_state                   TEXT DEFAULT '';
  vSD_msg                     TEXT DEFAULT '';
  vSD_detail                  TEXT DEFAULT '';
  vSD_hint                    TEXT DEFAULT '';
  vSD_context                 TEXT DEFAULT '';

  BEGIN

    -- Verifying this procedure is the single instance running
    IF NOT pg_try_advisory_lock(vAdLckID)
    THEN
      RAISE NOTICE 'ERROR: Could not aquire lock. There may be another instance of this procedure already running. Exiting';
      RETURN 0;
    END IF;

    -- Validate input parameters
    IF pPurgeExecutionsInBatchesOf < vMinBatchSize
    THEN
      RAISE NOTICE '%', repeat('=',vOutputRowLength);
      RAISE NOTICE '%', repeat('**** ERROR ',5);
      RAISE NOTICE '**** Purging batch size is too small (provided %).  Minimum value is %', pPurgeExecutionsInBatchesOf, vMinBatchSize;
      RAISE NOTICE '%', repeat('=',vOutputRowLength);
      PERFORM pg_advisory_unlock_all();
      RETURN 0;
    END IF;

    IF pPurgeExecutionsInBatchesOf > vMaxBatchSize
    THEN
      RAISE NOTICE '%', repeat('=',vOutputRowLength);
      RAISE NOTICE '%', repeat('**** ERROR ',5);
      RAISE NOTICE '**** Purging batch size is too big (provided %). Maximum value is %', pPurgeExecutionsInBatchesOf, vMaxBatchSize;
      RAISE NOTICE '%', repeat('=',vOutputRowLength);
      PERFORM pg_advisory_unlock_all();
      RETURN 0;
    END IF;

    IF pPurgeExecutionsOlderThan < 1
    THEN
      RAISE NOTICE '%', repeat('=',vOutputRowLength);
      RAISE NOTICE '%', repeat('**** ERROR ',5);
      RAISE NOTICE '**** Cannot keep less than 1 day (provided %). Set pPurgeExecutionsOlderThan parameter accordingly.', pPurgeExecutionsOlderThan;
      RAISE NOTICE '%', repeat('=',vOutputRowLength);
      PERFORM pg_advisory_unlock_all();
      RETURN 0;
    END IF;

    IF pPurgeExecutionsOlderThan > 1000
    THEN
      RAISE NOTICE '%', repeat('=',vOutputRowLength);
      RAISE NOTICE '%', repeat('**** ERROR ',5);
      RAISE NOTICE '**** Retention is more than 1000 days? (provided %). Please check the value provided for PurgeExecutionsOlderThan.', pPurgeExecutionsOlderThan;
      RAISE NOTICE '%', repeat('=',vOutputRowLength);
      PERFORM pg_advisory_unlock_all();
      RETURN 0;
    END IF;

    IF pShouldPurgeExecutionSummary NOT IN (0,1)
    THEN
      RAISE NOTICE '%', repeat('=',vOutputRowLength);
      RAISE NOTICE '%', repeat('**** ERROR ',5);
      RAISE NOTICE 'INFO: pShouldPurgeExecutionSummary must be either 0 or 1 (provided %)', pShouldPurgeExecutionSummary;
      RAISE NOTICE '%', repeat('=',vOutputRowLength);
      PERFORM pg_advisory_unlock_all();
      RETURN 0;
    END IF;

    IF pDeepClean NOT IN (0,1)
    THEN
      RAISE NOTICE 'INFO: pDeepClean must be either 0 or 1. Defaulting to 0.';
      pDeepClean := 0;
    END IF;

    IF pVerbose NOT IN (0,1,2,3,4)
    THEN
      RAISE NOTICE 'INFO: pVerbose must be one of: 0,1 or 2. Defaulting to 1.';
      pVerbose := 1;
    END IF;

    IF pSleepInterval < 0.01
    THEN
      RAISE NOTICE 'INFO: pSleepInterval must fall in the range 0.01 - 5.0. Setting to 0.';
      pSleepInterval := 0;
    END IF;

    IF pSleepInterval > 5
    THEN
      RAISE NOTICE 'INFO: pVerbose must fall in the range 0.01 - 5.0. Setting to top limit.';
      pSleepInterval := 5;
    END IF;

    vPurgeHistoryVersion := '4.2';
    vPurgeHistoryCompatibility := '10.6x to 10.70';
    vSupportedVersions := '''10.60'',''10.70''';
    vPurgedFlowStatuses := '''COMPLETED'',''CANCELED'',''SYSTEM_FAILURE''';
    vPurgeStartTime := clock_timestamp();
    vPurgeEndTime := vPurgeStartTime + INTERVAL '1 HOUR' * pStopPurgingProcessAfter;

    ---------------------------------------------------------------------------------------------------------
    -- Verify that OO release is supported by this stored procedure                                        --
    ---------------------------------------------------------------------------------------------------------
    query := 'SELECT VALUE, CASE WHEN VALUE IN (' || vSupportedVersions || ') THEN 1 ELSE 0 END FROM OO_SYSINFO WHERE UPPER(KEY_)=''VERSION''';
    EXECUTE query INTO vDetectedOOVersion,vGeneralCounter;
    IF vGeneralCounter <> 1
    THEN
      RAISE NOTICE '%', repeat('=',vOutputRowLength);
      RAISE NOTICE '%', repeat('**** ERROR ',5);
      RAISE NOTICE '**** Unsupported OO version: %. This procedure only supports:', vDetectedOOVersion;
      RAISE NOTICE '%', REPLACE(vSupportedVersions,'''''','''');
      RAISE NOTICE '%', repeat('=',vOutputRowLength);
      RETURN 0;
    END IF;

    -- Output header --------------------------------------------------------------------------------------------------------------------
    RAISE NOTICE '%', repeat('-',vOutputRowLength);
    RAISE NOTICE 'Beginning execution of OOPurgeHistory procedure version % at %', vPurgeHistoryVersion, TO_CHAR(clock_timestamp(),'DD/MM/YYYY HH24:MI:SS');
    RAISE NOTICE 'This procedure is compatible with OO %', vPurgeHistoryCompatibility;
    RAISE NOTICE '%', repeat('-',vOutputRowLength);
    -- End of output header -------------------------------------------------------------------------------------------------------------

    -- Count rows in OO_EXECUTION_SUMMARY table. This could take a while, but it is worth it
    SELECT COUNT(1) INTO vGeneralCounter FROM OO_EXECUTION_SUMMARY;

    IF vGeneralCounter = 0
    THEN
      RAISE NOTICE 'INFO: OO_EXECUTION_SUMMARY table is empty';
      vOnlyHandleOrphans := TRUE;
    ELSE
      vSkipToSummary := TRUE;
    END IF;

    -- Retention period calculation. Get the newest run start time we need to KEEP
    vLastStartTime := (extract(epoch from date_trunc('milliseconds', now()))*1000) - (86400000*pPurgeExecutionsOlderThan);

    -- Flow batch size is relative to the steps batch size, but at least 100
    vFlowBatchSize = ROUND((pPurgeExecutionsInBatchesOf/10),0);
    IF vFlowBatchSize < 100
    THEN
      vFlowBatchSize := 100;
    END IF;

    ---------------------------------------------------------------------------------------------------------
    -- Get the minimal and maximal EXECUTION_ID values to handle                                           --
    -- As EXECUTION_ID allocation is asynchroneous, this range might also include EXECUTION_IDs            --
    -- which have started after the threshold time and should not be handled                               --
    ---------------------------------------------------------------------------------------------------------
    query1 := 'SELECT' ||
        ' MIN(CAST(EXECUTION_ID AS BIGINT)), MAX(CAST(EXECUTION_ID AS BIGINT))' ||
        ' FROM OO_EXECUTION_SUMMARY' ||
        ' WHERE START_TIME_LONG < $1' ||
        ' AND STATUS IN (' || vPurgedFlowStatuses || ')';

    query2 := 'SELECT' ||
        ' MIN(CAST(ES.EXECUTION_ID AS BIGINT)), MAX(CAST(ES.EXECUTION_ID AS BIGINT))' ||
        ' FROM OO_EXECUTION_SUMMARY ES' ||
        ' JOIN (SELECT MIN(CAST(EXECUTION_ID AS BIGINT)) MIN_EXEC_ID, MAX(CAST(EXECUTION_ID AS BIGINT)) MAX_EXEC_ID FROM OO_STEP_LOG_STARTED) SLSTR' ||
        ' ON ( CAST(ES.EXECUTION_ID AS BIGINT) BETWEEN SLSTR.MIN_EXEC_ID AND SLSTR.MAX_EXEC_ID )' ||
        ' WHERE ES.START_TIME_LONG < $1' ||
        ' AND ES.STATUS IN (' || vPurgedFlowStatuses || ')';

    IF pShouldPurgeExecutionSummary = 1
    THEN
      EXECUTE query1 USING vLastStartTime INTO vMinExecIdToDelete, vMaxExecIdToDelete;
    ELSE
      -- In case OO_EXECUTION_SUMMARY should not be purged, limit scpoe using OO_STEP_LOG_STARTED
      EXECUTE query2 USING vLastStartTime INTO vMinExecIdToDelete, vMaxExecIdToDelete;
    END IF;

    IF vMinExecIdToDelete IS NULL OR vMaxExecIdToDelete IS NULL
    THEN
      RAISE NOTICE 'Could not find purge-eligible flows.';
      vOnlyHandleOrphans := TRUE;
      vMinExecIdToDelete := 0;
      vMaxExecIdToDelete := 0;
      IF pDeepClean = 1
      THEN
        vOnlyHandleOrphans := TRUE;
      ELSE
        vSkipToSummary := TRUE;
      END IF;
    END IF;

    SELECT substring(substring(version() from 'PostgreSQL [9][.][1-9]') from '[9][.][1-9]') "Version"
    INTO vPgVersion;

    -- Increase lock wait timeout for the function run if possible - 9.3 and above
    IF vPgVersion >= '9.3'
    THEN
      SET LOCAL lock_timeout = '5s';
    END IF;

    -- Beginning of operation ---------------------------------------------------------------------------------
    RAISE NOTICE 'This procedure will handle flows records older than: %',TO_CHAR((vPurgeStartTime - INTERVAL '1 DAY' * pPurgeExecutionsOlderThan),'DD/MM/YYYY HH24:MI:SS');
    RAISE NOTICE 'This procedure will automatically quit around: %',TO_CHAR(vPurgeEndTime,'DD/MM/YYYY HH24:MI:SS');

    -- Prepare for the the first batch run
    vCurrEXCIDBatchMinId := vMinExecIdToDelete; -- Starting at the oldest record we need to delete
    vCurrEXCIDBatchMaxId := vMinExecIdToDelete;

    -- Define queries -----------------------------------------------------------------------------------------
    query3 := 'INSERT INTO SL_STRTD_STPS' ||
              ' SELECT row_number() OVER (ORDER BY NULL), q.* FROM (' ||
              ' SELECT CAST(SLS.EXECUTION_ID AS BIGINT), SLS.STEP_COUNT' ||
              ' FROM ( SELECT EXECUTION_ID, COUNT(EXECUTION_ID) STEP_COUNT FROM OO_STEP_LOG_STARTED' ||
              ' WHERE CAST(EXECUTION_ID AS BIGINT) BETWEEN $1 AND $2 GROUP BY EXECUTION_ID) SLS' ||
              ' JOIN OO_EXECUTION_SUMMARY ES ON SLS.EXECUTION_ID = ES.EXECUTION_ID' ||
              ' WHERE ES.STATUS IN (' || vPurgedFlowStatuses || ')' ||
              ' AND ES.START_TIME_LONG < $3 ORDER BY SLS.EXECUTION_ID ASC LIMIT $4) q';

    query4 := 'DELETE FROM OO_EXECUTION_SUMMARY' ||
              ' WHERE CAST(EXECUTION_ID AS BIGINT) BETWEEN $1 AND $2' ||
              ' AND STATUS IN (' || vPurgedFlowStatuses || ') AND START_TIME_LONG < $3';

    query5 := 'DELETE FROM OO_EXECUTION_SUMMARY WHERE ID IN' ||
              ' (SELECT ID FROM OO_EXECUTION_SUMMARY WHERE EXECUTION_ID NOT IN' ||
              ' (SELECT DISTINCT(EXECUTION_ID) FROM OO_STEP_LOG_STARTED)' ||
              ' AND STATUS IN (' || vPurgedFlowStatuses || ') AND START_TIME_LONG < $1 LIMIT $2)';

    ---------------------------------------------------------------------------------------------------------
    --  Remove child-less OO_EXECUTION_SUMMARY records if applicable                                       --
    --  -------------------------------------------------------------                                      --
    --  If OO_EXECUTION_SUMMARY contains records that do not have corresponding OO_STEP_LOG_STARTED data   --
    --  repeatedly purge these records                                                                     --
    ---------------------------------------------------------------------------------------------------------
    IF pShouldPurgeExecutionSummary = 1
    THEN
      vTempTimestamp := clock_timestamp(); -- Start measuring time for OO_EXECUTION_SUMMARY child-less records purge
      vDeletedRowsInCurrentBatch := 0;

      vGeneralCounter := vDeleteBatchSize; -- Qualify 'while' condition for the first round
      WHILE vGeneralCounter >= vDeleteBatchSize
      LOOP

        vGeneralCounter := 0;
        EXECUTE query5 USING vLastStartTime, vDeleteBatchSize;
        GET DIAGNOSTICS vGeneralCounter = ROW_COUNT;
        vDeletedRowsInCurrentBatch := vDeletedRowsInCurrentBatch + vGeneralCounter;
        vExSumDeletedRows := vExSumDeletedRows + vGeneralCounter;
        vTotalFlowsHandled := vExSumDeletedRows;

        IF clock_timestamp() > vPurgeEndTime
        THEN
          RAISE NOTICE '------ Time limit reached ---------';
          vSkipToSummary := TRUE;
        END IF;

      END LOOP; -- End of child-less OO_EXECUTION_SUMMARY records purge WHILE loop

    END IF;

    --=======================================================================================================--
    -- Main loop                                                                                             --
    --=======================================================================================================--

    vCurrEXCIDBatchMaxId := CAST(CAST(vCurrEXCIDBatchMinId AS BIGINT) + pPurgeExecutionsInBatchesOf - 1 AS CHARACTER VARYING);

    -- Prepare a temporary table containing EXECUTION_ID and their associated step count
    DISCARD TEMP;
    CREATE TEMPORARY TABLE IF NOT EXISTS SL_STRTD_STPS
    (id INT,
     EXECUTION_ID BIGINT,
     STEP_COUNT INT
    ) ON COMMIT PRESERVE ROWS;

    -- Prepare a temporary table containing step HASH IDs
    CREATE TEMPORARY TABLE IF NOT EXISTS ST_HASH_IDS
    (STEP_HASH_ID CHARACTER VARYING(42)) ON COMMIT PRESERVE ROWS;

    -- Prepare a temporary table containing ended step IDs
    CREATE TEMPORARY TABLE IF NOT EXISTS ST_ENDED_IDS
    (ENDED_ID BIGINT) ON COMMIT PRESERVE ROWS;

     -- Prepare a temporary table containing IDs
    CREATE TEMPORARY TABLE IF NOT EXISTS ARBITRARY_IDS
    (ID BIGINT) ON COMMIT PRESERVE ROWS;

     -- Prepare a temporary table containing unique IDs
    CREATE TEMPORARY TABLE IF NOT EXISTS ARBITRARY_IDS_COMPACT
    (ID BIGINT) ON COMMIT PRESERVE ROWS;

    -- Main loop
    <<main_loop_label>>
    WHILE vCurrEXCIDBatchMinId <= vMaxExecIdToDelete AND vOnlyHandleOrphans = FALSE
    LOOP

      IF clock_timestamp() > vPurgeEndTime
      THEN
        RAISE NOTICE 'INFO: Time limit reached, exiting.';
        EXIT main_loop_label;
      END IF;

      ---------------------------------------------------------------------------------------------------------
      -- Get the next batch of flows and corresponding step count into the SL_STRTD_STPS collection          --
      -- The number of elements is between 0 and vFlowBatchSize                                              --
      ---------------------------------------------------------------------------------------------------------
      TRUNCATE TABLE SL_STRTD_STPS;
      EXECUTE query3 USING vCurrEXCIDBatchMinId, vMaxExecIdToDelete
      , vLastStartTime, vFlowBatchSize;
      SELECT COUNT(1) INTO vCurSetSize FROM SL_STRTD_STPS;

      ---------------------------------------------------------------------------------------------------------
      -- If there are no more flows to handle - exit main WHILE loop                                         --
      ---------------------------------------------------------------------------------------------------------
      IF vCurSetSize IS NULL OR vCurSetSize < 1
      THEN
        vCurrEXCIDBatchMinId := vMaxExecIdToDelete + 1; -- Qualify exit condition just in case
        EXIT main_loop_label;
      END IF;

      ---------------------------------------------------------------------------------------------------------
      -- Set the current batch higher execution ID pointer                                                   --
      ---------------------------------------------------------------------------------------------------------
      SELECT CAST(EXECUTION_ID AS BIGINT) INTO vCurrEXCIDBatchMaxId FROM SL_STRTD_STPS WHERE id = (SELECT MAX(id) FROM SL_STRTD_STPS);

      ---------------------------------------------------------------------------------------------------------
      -- At this point SL_STRTD_STPS contains a list of flows and their corresponding number of steps        --
      -- SL_STRTD_STPS contains between 1 and vFlowBatchSize records                                         --
      ---------------------------------------------------------------------------------------------------------
      IF pVerbose >= 2
      THEN
        RAISE NOTICE '% INFO: Handling EXECUTION_ID range % %',TO_CHAR(clock_timestamp(),'HH24:MI:SS') , vCurrEXCIDBatchMinId, vCurrEXCIDBatchMaxId;
      END IF;

      vCurSetAccumStepCount := 0;
      IF pShouldPurgeExecutionSummary = 0
      THEN
        vTotalFlowsHandled := vTotalFlowsHandled + vCurSetSize;
      END IF;

      -------------------------------------------------------------------------------------------------------
      -- The following loop breaks down all flows in SL_STRTD_STPS, into smaller groups,
      -- accumulating and handling pPurgeExecutionsInBatchesOf steps (or a bit more)
      -- to be handled together
      -------------------------------------------------------------------------------------------------------

      -- Reset variables
      vSLStrtStpCount := 0;
      vCurSetMinIdx := NULL;
      vCurSetMaxIdx := NULL;

      -- Iterate over the list of OO_STEP_LOG_STARTED EXECUTION_IDs and step counts
      FOR vSlStpCntRec IN SELECT * FROM SL_STRTD_STPS ORDER BY id ASC
      LOOP

        -- Accumulate step counts
        vCurStpCount := vSlStpCntRec.STEP_COUNT;
        vSLStrtStpCount := vSLStrtStpCount + vCurStpCount;

        -- Get the minimal EXECUTION_ID index for this group of executions
        IF vCurSetMinIdx IS NULL
        THEN
          vCurSetMinIdx := vSlStpCntRec.id;
        END IF;

        -- Get the current maximal EXECUTION_ID index for this group of executions
        vCurSetMaxIdx := vSlStpCntRec.id;

        -- PurgeExecutionsInBatchesOf (or a bit more) steps are accumulated, or this is the last round for this batch
        IF vSLStrtStpCount >= pPurgeExecutionsInBatchesOf OR vCurSetMaxIdx = vCurSetSize
        THEN

          -------------------------------------------------------------------------------------------------------
          -- All conditions are met - we can finally begin deleting records
          -------------------------------------------------------------------------------------------------------

          -- Alert that this flow has an unusually large number of steps and derivatives
          IF vCurStpCount >= vLotsOfStepsThreshold AND pVerbose >= 2
          THEN
            RAISE NOTICE '% DEBUG: Note - step count for EXECUTION_ID %  is relatively high: %',TO_CHAR(clock_timestamp(),'HH24:MI:SS'), TO_CHAR(vSlStpCntRec.EXECUTION_ID), TO_CHAR(vSlStpCntRec.STEP_COUNT);
          END IF;

          -- Prepare/empty buffers for this batch of flows ...
          TRUNCATE TABLE ST_HASH_IDS;
          TRUNCATE TABLE ST_ENDED_IDS;
          TRUNCATE TABLE ARBITRARY_IDS;
          TRUNCATE TABLE ARBITRARY_IDS_COMPACT;

          ----------------------------------------------------------------------------------------
          -- Deleting from OO_STEP_LOG_STARTED                                                  --
          ----------------------------------------------------------------------------------------
          vGeneralCounter := vDeleteBatchSize; -- Qualify 'while' condition for the first round
          vDeletedRowsInCurrentBatch := 0;
          WHILE vGeneralCounter >= vDeleteBatchSize
		  LOOP

            -- Get the current number of records
            SELECT COUNT(1) INTO vGeneralCounter FROM ST_HASH_IDS;

			WITH delres AS (DELETE FROM OO_STEP_LOG_STARTED
			WHERE ID IN ( SELECT ID FROM OO_STEP_LOG_STARTED
            WHERE EXECUTION_ID IN (
              SELECT CAST(EXECUTION_ID AS CHARACTER VARYING(64)) FROM SL_STRTD_STPS
              WHERE id BETWEEN vCurSetMinIdx AND vCurSetMaxIdx
            ) LIMIT vDeleteBatchSize) RETURNING STEP_HASH_ID)
			INSERT INTO ST_HASH_IDS SELECT STEP_HASH_ID FROM delres;

            -- vGeneralCounter is set to the number of records that were added this round...
            SELECT (COUNT(*) - vGeneralCounter) INTO vGeneralCounter FROM ST_HASH_IDS;
            vSLSTRTDeletedRows := vSLSTRTDeletedRows + vGeneralCounter;
            vDeletedRowsInCurrentBatch := vDeletedRowsInCurrentBatch + vGeneralCounter;

            -- Delay if necessary
            IF pSleepInterval > 0
            THEN
              PERFORM pg_sleep(pSleepInterval);
            END IF;

            IF clock_timestamp() > vPurgeEndTime
            THEN
              RAISE NOTICE 'INFO: Time limit reached, exiting.';
              EXIT main_loop_label;
            END IF;

          END LOOP; -- End of iteratively deleting records from OO_STEP_LOG_STARTED (while loop)

          ----------------------------------------------------------------------------------------
          -- Deleting from OO_STEP_LOG_ENDED                                                    --
          ----------------------------------------------------------------------------------------
          vGeneralCounter := vDeleteBatchSize; -- Qualify 'while' condition for the first round
          vDeletedRowsInCurrentBatch := 0;
          WHILE vGeneralCounter >= vDeleteBatchSize
		  LOOP

            -- Get the current number of records
            SELECT COUNT(1) INTO vGeneralCounter FROM ST_ENDED_IDS;

			WITH delres AS (DELETE FROM OO_STEP_LOG_ENDED
			WHERE ID IN (SELECT ID FROM OO_STEP_LOG_ENDED
			WHERE STEP_ENDED_HASH_ID IN (
              SELECT STEP_HASH_ID FROM ST_HASH_IDS
            ) LIMIT vDeleteBatchSize) RETURNING ID)
			INSERT INTO ST_ENDED_IDS SELECT ID FROM delres;

            -- vGeneralCounter is set to the number of records that were added this round...
            SELECT (COUNT(*) - vGeneralCounter) INTO vGeneralCounter FROM ST_ENDED_IDS;
            vSLENDDeletedRows := vSLENDDeletedRows + vGeneralCounter;
            vDeletedRowsInCurrentBatch := vDeletedRowsInCurrentBatch + vGeneralCounter;

            -- Delay if necessary
            IF pSleepInterval > 0
            THEN
              PERFORM pg_sleep(pSleepInterval);
            END IF;

            IF clock_timestamp() > vPurgeEndTime
            THEN
              RAISE NOTICE 'INFO: Time limit reached, exiting.';
              EXIT main_loop_label;
            END IF;

          END LOOP; -- End of iteratively deleting records from OO_STEP_LOG_ENDED (while loop)

          ----------------------------------------------------------------------------------------
          -- Deleting from OO_STEP_LOG_BINDINGS                                                  --
          ----------------------------------------------------------------------------------------
          vGeneralCounter := vDeleteBatchSize; -- Qualify 'while' condition for the first round
          vDeletedRowsInCurrentBatch := 0;
          WHILE vGeneralCounter >= vDeleteBatchSize
		  LOOP

            -- Get the current number of records
            SELECT COUNT(1) INTO vGeneralCounter FROM ARBITRARY_IDS;

			WITH delres AS (DELETE FROM OO_STEP_LOG_BINDINGS
			WHERE ID IN ( SELECT ID FROM OO_STEP_LOG_BINDINGS
            WHERE STEP_LOG_ENDED_ID IN (
              SELECT ENDED_ID FROM ST_ENDED_IDS
            ) LIMIT vDeleteBatchSize) RETURNING ID)
			INSERT INTO ARBITRARY_IDS SELECT ID FROM delres;

            -- vGeneralCounter is set to the number of records that were added this round...
            SELECT (COUNT(*) - vGeneralCounter) INTO vGeneralCounter FROM ARBITRARY_IDS;
            vSLBNDDeletedRows := vSLBNDDeletedRows + vGeneralCounter;
            vDeletedRowsInCurrentBatch := vDeletedRowsInCurrentBatch + vGeneralCounter;

            -- Delay if necessary
            IF pSleepInterval > 0
            THEN
              PERFORM pg_sleep(pSleepInterval);
            END IF;

            IF clock_timestamp() > vPurgeEndTime
            THEN
              RAISE NOTICE 'INFO: Time limit reached, exiting.';
              EXIT main_loop_label;
            END IF;

          END LOOP; -- End of iteratively deleting records from OO_STEP_LOG_BINDINGS (while loop)

          ----------------------------------------------------------------------------------------
          -- Deleting from OO_EXECUTION_BOUND_OUTPUTS                                           --
          ----------------------------------------------------------------------------------------
          vGeneralCounter := vDeleteBatchSize; -- Qualify 'while' condition for the first round
          vDeletedRowsInCurrentBatch := 0;
          WHILE vGeneralCounter >= vDeleteBatchSize
		  LOOP

			DELETE FROM OO_EXECUTION_BOUND_OUTPUTS
			WHERE ID IN ( SELECT ID FROM OO_EXECUTION_BOUND_OUTPUTS
            WHERE EXECUTION_ID IN (
              SELECT CAST(EXECUTION_ID AS CHARACTER VARYING(64)) FROM SL_STRTD_STPS
              WHERE id BETWEEN vCurSetMinIdx AND vCurSetMaxIdx
            ) LIMIT vDeleteBatchSize);
			GET DIAGNOSTICS vGeneralCounter = ROW_COUNT;

            vBODeletedRows := vBODeletedRows + vGeneralCounter;
            vDeletedRowsInCurrentBatch := vDeletedRowsInCurrentBatch + vGeneralCounter;

            -- Delay if necessary
            IF pSleepInterval > 0
            THEN
              PERFORM pg_sleep(pSleepInterval);
            END IF;

            IF clock_timestamp() > vPurgeEndTime
            THEN
              RAISE NOTICE 'INFO: Time limit reached, exiting.';
              EXIT main_loop_label;
            END IF;

          END LOOP; -- End of iteratively deleting records from OO_EXECUTION_BOUND_OUTPUTS (while loop)

          ----------------------------------------------------------------------------------------
          -- Deleting from OO_EXECUTION_BOUND_INPUTS                                            --
          ----------------------------------------------------------------------------------------
          vGeneralCounter := vDeleteBatchSize; -- Qualify 'while' condition for the first round
          vDeletedRowsInCurrentBatch := 0;
          WHILE vGeneralCounter >= vDeleteBatchSize
		  LOOP

			DELETE FROM OO_EXECUTION_BOUND_INPUTS
			WHERE ID IN ( SELECT ID FROM OO_EXECUTION_BOUND_INPUTS
            WHERE EXECUTION_ID IN (
              SELECT CAST(EXECUTION_ID AS CHARACTER VARYING(64)) FROM SL_STRTD_STPS
              WHERE id BETWEEN vCurSetMinIdx AND vCurSetMaxIdx
            ) LIMIT vDeleteBatchSize);
            GET DIAGNOSTICS vGeneralCounter = ROW_COUNT;

            vBIDeletedRows := vBIDeletedRows + vGeneralCounter;
            vDeletedRowsInCurrentBatch := vDeletedRowsInCurrentBatch + vGeneralCounter;

            -- Delay if necessary
            IF pSleepInterval > 0
            THEN
              PERFORM pg_sleep(pSleepInterval);
            END IF;

            IF clock_timestamp() > vPurgeEndTime
            THEN
              RAISE NOTICE 'INFO: Time limit reached, exiting.';
              EXIT main_loop_label;
            END IF;

          END LOOP; -- End of iteratively deleting records from OO_EXECUTION_BOUND_INPUTS (while loop)

          -- Prepare for the next set
          vCurSetAccumStepCount := 0;
          vCurSetMinIdx := NULL;
          vCurSetMaxIdx := NULL;

        END IF; -- End of "IF" we have enough steps to process, or last round

      END LOOP; -- End of "WHILE" - iterating through SL_STRTD_STPS

      ----------------------------------------------------------------------------------------
      -- Deleting from OO_EXECUTION_SUMMARY if applicable                                   --
      ----------------------------------------------------------------------------------------
      IF pShouldPurgeExecutionSummary = 1
      THEN

        IF pVerbose >= 3
        THEN
          RAISE NOTICE '% DEBUG: Deleting from OO_EXECUTION_SUMMARY',TO_CHAR(clock_timestamp(),'HH24:MI:SS');
        END IF;

        EXECUTE query4 USING vCurrEXCIDBatchMinId, vCurrEXCIDBatchMaxId, vLastStartTime;
        GET DIAGNOSTICS vGeneralCounter = ROW_COUNT;
		vExSumDeletedRows := vExSumDeletedRows + vGeneralCounter;
        vTotalFlowsHandled := vTotalFlowsHandled + vGeneralCounter;

      END IF; -- Handling OO_EXECUTION_SUMMARY depending on pShouldPurgeExecutionSummary

      IF pVerbose >= 2
      THEN
        vTotalDelRows := vExSumDeletedRows + vSLSTRTDeletedRows + vSLENDDeletedRows + vSLBNDDeletedRows + vBODeletedRows + vBIDeletedRows;
        RAISE NOTICE '% INFO: Time elapsed so far: %, rows deleted so far: %',TO_CHAR(clock_timestamp(),'HH24:MI:SS'), age(clock_timestamp(),vPurgeStartTime),CAST(vTotalDelRows AS CHARACTER VARYING);
      END IF;

      -- Prepare for the next OO_EXECUTION_SUMMARY batch

      -- Advance pointers
      vCurrEXCIDBatchMinId := vCurrEXCIDBatchMaxId + 1;
      vCurrEXCIDBatchMaxId := vCurrEXCIDBatchMinId;

      -- Cutoff at upper limit
      IF vCurrEXCIDBatchMaxId >= vMaxExecIdToDelete
      THEN
        vCurrEXCIDBatchMaxId := vMaxExecIdToDelete + 1; -- Qualify exit condition just in case...
      END IF;

    END LOOP main_loop_label; -- End of main WHILE loop

    ----------------------------------------------------------------------------------------
    -- Deep clean operations, removing orphan records if such are detected                --
    ----------------------------------------------------------------------------------------
    <<orphan_records_label>>
    BEGIN

      IF pDeepClean <> 1
      THEN
        EXIT orphan_records_label;
      END IF;

      ----------------------------------------------------------------------------------------
      -- Deep cleaning OO_STEP_LOG_BINDINGS                                                 --
      ----------------------------------------------------------------------------------------

	  -- Get a list of orphan OO_STEP_LOG_BINDINGS records
      TRUNCATE TABLE ARBITRARY_IDS;
      INSERT INTO ARBITRARY_IDS
	  SELECT DISTINCT STEP_LOG_ENDED_ID FROM OO_STEP_LOG_BINDINGS
      WHERE STEP_LOG_ENDED_ID NOT IN (SELECT ID FROM OO_STEP_LOG_ENDED)
      ORDER BY STEP_LOG_ENDED_ID ASC LIMIT vDeleteBatchSize;
      GET DIAGNOSTICS vGeneralCounter = ROW_COUNT;

      -- Get the current number of records, and qualify WHILE loop condition if applicable
	  IF vGeneralCounter > 0
	  THEN
        vGeneralCounter := vDeleteBatchSize; -- Qualify 'while' condition for the first round
        IF pVerbose >= 2
        THEN
          RAISE NOTICE '% INFO: Handling orphan records in step log bindings table',TO_CHAR(clock_timestamp(),'HH24:MI:SS');
        END IF;
	  END IF;

      vDeletedRowsInCurrentBatch := 0;
      WHILE vGeneralCounter >= vDeleteBatchSize
      LOOP

        IF clock_timestamp() > vPurgeEndTime
        THEN
          RAISE NOTICE 'INFO: Time limit reached, exiting.';
          EXIT orphan_records_label;
        END IF;

        DELETE FROM OO_STEP_LOG_BINDINGS
        WHERE ID IN (
          SELECT ID FROM OO_STEP_LOG_BINDINGS
          WHERE STEP_LOG_ENDED_ID IN (SELECT ID FROM ARBITRARY_IDS)
          ORDER BY STEP_LOG_ENDED_ID ASC LIMIT vDeleteBatchSize);
        GET DIAGNOSTICS vGeneralCounter = ROW_COUNT;
        vSLBNDDeletedRows := vSLBNDDeletedRows + vGeneralCounter;

        -- Delay if necessary
        IF pSleepInterval > 0
        THEN
          PERFORM pg_sleep(pSleepInterval);
        END IF;

        IF clock_timestamp() > vPurgeEndTime
        THEN
          RAISE NOTICE 'INFO: Time limit reached, exiting.';
          EXIT orphan_records_label;
        END IF;

      END LOOP; -- Deleting from OO_STEP_LOG_BINDINGS

      ----------------------------------------------------------------------------------------
      -- Deep cleaning OO_STEP_LOG_ENDED                                                    --
      ----------------------------------------------------------------------------------------

      -- Handling step log ended orphan records

	  -- Get a list of orphan OO_STEP_LOG_ENDED records
      TRUNCATE TABLE ARBITRARY_IDS;
      INSERT INTO ARBITRARY_IDS
	  SELECT ID FROM OO_STEP_LOG_ENDED
      WHERE STEP_ENDED_HASH_ID NOT IN (SELECT STEP_HASH_ID FROM OO_STEP_LOG_STARTED)
	  ORDER BY ID ASC LIMIT vDeleteBatchSize;
      GET DIAGNOSTICS vGeneralCounter = ROW_COUNT;

      -- Get the current number of records, and qualify WHILE loop condition if applicable
	  IF vGeneralCounter > 0
	  THEN
        vGeneralCounter := vDeleteBatchSize; -- Qualify 'while' condition for the first round
        IF pVerbose >= 2
        THEN
          RAISE NOTICE '% INFO: Handling orphan records in step log ended table',TO_CHAR(clock_timestamp(),'HH24:MI:SS');
        END IF;
	  END IF;

      vDeletedRowsInCurrentBatch := 0;
      WHILE vGeneralCounter >= vDeleteBatchSize
      LOOP

        IF clock_timestamp() > vPurgeEndTime
        THEN
          RAISE NOTICE 'INFO: Time limit reached, exiting.';
          EXIT orphan_records_label;
        END IF;

        DELETE FROM OO_STEP_LOG_ENDED
        WHERE ID IN (SELECT ID FROM ARBITRARY_IDS
          ORDER BY ID ASC LIMIT vDeleteBatchSize);
        GET DIAGNOSTICS vGeneralCounter = ROW_COUNT;
        vSLENDDeletedRows := vSLENDDeletedRows + vGeneralCounter;

        DELETE FROM ARBITRARY_IDS
        WHERE ID IN (SELECT ID FROM ARBITRARY_IDS
          ORDER BY ID ASC LIMIT vDeleteBatchSize);

        -- Delay if necessary
        IF pSleepInterval > 0
        THEN
          PERFORM pg_sleep(pSleepInterval);
        END IF;

        IF clock_timestamp() > vPurgeEndTime
        THEN
          RAISE NOTICE 'INFO: Time limit reached, exiting.';
          EXIT orphan_records_label;
        END IF;

      END LOOP; -- Deleting from OO_STEP_LOG_ENDED

      ----------------------------------------------------------------------------------------
      -- Deep cleaning OO_STEP_LOG_STARTED                                                  --
      ----------------------------------------------------------------------------------------

	  -- Get a list of orphan OO_STEP_LOG_STARTED records
      TRUNCATE TABLE ARBITRARY_IDS;
      INSERT INTO ARBITRARY_IDS
	  SELECT ID FROM OO_STEP_LOG_STARTED
      WHERE EXECUTION_ID NOT IN (SELECT EXECUTION_ID FROM OO_EXECUTION_SUMMARY)
	  ORDER BY ID ASC LIMIT vDeleteBatchSize;
      GET DIAGNOSTICS vGeneralCounter = ROW_COUNT;

      -- Get the current number of records, and qualify WHILE loop condition if applicable
	  IF vGeneralCounter > 0
	  THEN
        vGeneralCounter := vDeleteBatchSize; -- Qualify 'while' condition for the first round
        IF pVerbose >= 2
        THEN
          RAISE NOTICE '% INFO: Handling orphan records in step log started table',TO_CHAR(clock_timestamp(),'HH24:MI:SS');
        END IF;
	  END IF;

      vDeletedRowsInCurrentBatch := 0;
      WHILE vGeneralCounter >= vDeleteBatchSize
      LOOP

        IF clock_timestamp() > vPurgeEndTime
        THEN
          RAISE NOTICE 'INFO: Time limit reached, exiting.';
          EXIT orphan_records_label;
        END IF;

        DELETE FROM OO_STEP_LOG_STARTED
        WHERE ID IN (SELECT ID FROM ARBITRARY_IDS
          ORDER BY ID ASC LIMIT vDeleteBatchSize);
        GET DIAGNOSTICS vGeneralCounter = ROW_COUNT;
        vSLSTRTDeletedRows := vSLSTRTDeletedRows + vLastRowCount;

        DELETE FROM ARBITRARY_IDS
        WHERE ID IN (SELECT ID FROM ARBITRARY_IDS
          ORDER BY ID ASC LIMIT vDeleteBatchSize);

        -- Delay if necessary
        IF pSleepInterval > 0
        THEN
          PERFORM pg_sleep(pSleepInterval);
        END IF;

        IF clock_timestamp() > vPurgeEndTime
        THEN
          RAISE NOTICE 'INFO: Time limit reached, exiting.';
          EXIT orphan_records_label;
        END IF;

      END LOOP; -- Deleting from OO_STEP_LOG_STARTED

      ----------------------------------------------------------------------------------------
      -- Deep cleaning OO_EXECUTION_BOUND_INPUTS                                            --
      ----------------------------------------------------------------------------------------

	  -- Get a list of orphan OO_EXECUTION_BOUND_INPUTS records
      TRUNCATE TABLE ARBITRARY_IDS;
      INSERT INTO ARBITRARY_IDS
	  SELECT ID FROM OO_EXECUTION_BOUND_INPUTS
      WHERE EXECUTION_ID NOT IN (SELECT EXECUTION_ID FROM OO_EXECUTION_SUMMARY)
	  ORDER BY ID ASC LIMIT vDeleteBatchSize;
      GET DIAGNOSTICS vGeneralCounter = ROW_COUNT;

      -- Get the current number of records, and qualify WHILE loop condition if applicable
	  IF vGeneralCounter > 0
	  THEN
        vGeneralCounter := vDeleteBatchSize; -- Qualify 'while' condition for the first round
        IF pVerbose >= 2
        THEN
          RAISE NOTICE '% INFO: Handling orphan records in execution bound inputs table',TO_CHAR(clock_timestamp(),'HH24:MI:SS');
        END IF;
	  END IF;

      vDeletedRowsInCurrentBatch := 0;
      WHILE vGeneralCounter >= vDeleteBatchSize
      LOOP

        IF clock_timestamp() > vPurgeEndTime
        THEN
          RAISE NOTICE 'INFO: Time limit reached, exiting.';
          EXIT orphan_records_label;
        END IF;

        DELETE FROM OO_EXECUTION_BOUND_INPUTS
        WHERE ID IN (SELECT ID FROM ARBITRARY_IDS
          ORDER BY ID ASC LIMIT vDeleteBatchSize);
        GET DIAGNOSTICS vGeneralCounter = ROW_COUNT;
        vBIDeletedRows := vBIDeletedRows + vLastRowCount;

        DELETE FROM ARBITRARY_IDS
        WHERE ID IN (SELECT ID FROM ARBITRARY_IDS
          ORDER BY ID ASC LIMIT vDeleteBatchSize);

      END LOOP; -- Deleting from OO_EXECUTION_BOUND_INPUTS

      ----------------------------------------------------------------------------------------
      -- Deep cleaning OO_EXECUTION_BOUND_OUTPUTS                                           --
      ----------------------------------------------------------------------------------------

	  -- Get a list of orphan OO_EXECUTION_BOUND_OUTPUTS records
      TRUNCATE TABLE ARBITRARY_IDS;
      INSERT INTO ARBITRARY_IDS
	  SELECT ID FROM OO_EXECUTION_BOUND_OUTPUTS
      WHERE EXECUTION_ID NOT IN (SELECT EXECUTION_ID FROM OO_EXECUTION_SUMMARY)
	  ORDER BY ID ASC LIMIT vDeleteBatchSize;
      GET DIAGNOSTICS vGeneralCounter = ROW_COUNT;

      -- Get the current number of records, and qualify WHILE loop condition if applicable
	  IF vGeneralCounter > 0
	  THEN
        vGeneralCounter := vDeleteBatchSize; -- Qualify 'while' condition for the first round
        IF pVerbose >= 2
        THEN
          RAISE NOTICE '% INFO: Handling orphan records in execution bound inputs table',TO_CHAR(clock_timestamp(),'HH24:MI:SS');
        END IF;
	  END IF;

      vDeletedRowsInCurrentBatch := 0;
      WHILE vGeneralCounter >= vDeleteBatchSize
      LOOP

        IF clock_timestamp() > vPurgeEndTime
        THEN
          RAISE NOTICE 'INFO: Time limit reached, exiting.';
          EXIT orphan_records_label;
        END IF;

        DELETE FROM OO_EXECUTION_BOUND_OUTPUTS
        WHERE ID IN (SELECT ID FROM ARBITRARY_IDS
          ORDER BY ID ASC LIMIT vDeleteBatchSize);
        GET DIAGNOSTICS vGeneralCounter = ROW_COUNT;
        vBODeletedRows := vBODeletedRows + vLastRowCount;

        DELETE FROM ARBITRARY_IDS
        WHERE ID IN (SELECT ID FROM ARBITRARY_IDS
          ORDER BY ID ASC LIMIT vDeleteBatchSize);

        -- Delay if necessary
        IF pSleepInterval > 0
        THEN
          PERFORM pg_sleep(pSleepInterval);
        END IF;

        IF clock_timestamp() > vPurgeEndTime
        THEN
          RAISE NOTICE 'INFO: Time limit reached, exiting.';
          EXIT orphan_records_label;
        END IF;

      END LOOP; -- Deleting from OO_EXECUTION_BOUND_OUTPUTS

    END; -- End of handling orphan records

    -- Summary
    <<summary_label>>
    BEGIN

      IF pVerbose > 0
      THEN
        RAISE NOTICE 'Total purging time was: %', age(clock_timestamp(),vPurgeStartTime);
      END IF;

      IF pVerbose > 1
      THEN
        RAISE NOTICE 'SUMMARY: Deleted % rows from OO_EXECUTION_SUMMARY table',vExSumDeletedRows;
        RAISE NOTICE 'SUMMARY: Deleted % rows from OO_STEP_LOG_STARTED table',vSLSTRTDeletedRows;
        RAISE NOTICE 'SUMMARY: Deleted % rows from OO_STEP_LOG_ENDED table',vSLENDDeletedRows;
        RAISE NOTICE 'SUMMARY: Deleted % rows from OO_STEP_LOG_BINDINGS table',vSLBNDDeletedRows;
        RAISE NOTICE 'SUMMARY: Deleted % rows from OO_EXECUTION_BOUND_OUTPUTS table',vBODeletedRows;
        RAISE NOTICE 'SUMMARY: Deleted % rows from OO_EXECUTION_BOUND_INPUTS table',vBIDeletedRows;
      END IF;

      IF pVerbose >= 1
      THEN
        vTotalDelRows := vExSumDeletedRows + vSLSTRTDeletedRows + vSLENDDeletedRows + vSLBNDDeletedRows + vBODeletedRows + vBIDeletedRows;
        RAISE NOTICE 'SUMMARY:  A total of % rows deleted from database',CAST(vTotalDelRows AS CHARACTER VARYING);
        IF vTotalDelRows > vLotsOfStepsThreshold
        THEN
          RAISE NOTICE 'SUMMARY:  It is advisable to execute "vacuum full" in the next quiet period in order to reclaim LOB storage space';
        END IF;
      END IF;

      -- Release mutual exclusion lock
      PERFORM pg_advisory_unlock_all();

      RETURN vTotalFlowsHandled;

      EXCEPTION WHEN OTHERS THEN
      BEGIN
        IF vPgVersion >= '9.2'
        THEN
          GET STACKED DIAGNOSTICS
                vSD_state   = RETURNED_SQLSTATE,
                vSD_msg     = MESSAGE_TEXT,
                vSD_detail  = PG_EXCEPTION_DETAIL,
                vSD_hint    = PG_EXCEPTION_HINT,
                vSD_context = PG_EXCEPTION_CONTEXT;
          RAISE NOTICE E'Got exception:
                state  : %
                message: %
                detail : %
                hint   : %
                context: %', vSD_state, vSD_msg, vSD_detail, vSD_hint, vSD_context;
        END IF;
        PERFORM pg_advisory_unlock_all();
        RETURN vTotalFlowsHandled;
      END;
    END;
  END;
$$ LANGUAGE plpgsql;